Package: tabnet
Title: Fit 'TabNet' Models for Classification and Regression
Version: 0.9.1
Authors@R: c(
    person("Daniel", "Falbel", , "daniel@rstudio.com", role = "aut"),
    person(, "RStudio", role = "cph"),
    person("Christophe", "Regouby", , "christophe.regouby@free.fr", role = c("cre", "ctb")),
    person("Egill", "Fridgeirsson", role = "ctb"),
    person("Philipp", "Haarmeyer", role = "ctb"),
    person("Sven", "Verweij", role = "ctb",
           comment = c(ORCID = "0000-0002-5573-3952"))
  )
Description: Implements the 'TabNet' model by Sercan O. Arik et al. (2019)
    <doi:10.48550/arXiv.1908.07442> with 'Coherent Hierarchical
    Multi-label Classification Networks' by Giunchiglia et al.
    <doi:10.48550/arXiv.2010.10151> and provides a consistent interface
    for fitting and creating predictions.  It's also fully compatible with
    the 'tidymodels' ecosystem.
License: MIT + file LICENSE
URL: https://mlverse.github.io/tabnet/,
        https://github.com/mlverse/tabnet
BugReports: https://github.com/mlverse/tabnet/issues
Depends: R (>= 3.6)
Imports: coro, data.tree, dials, dplyr, generics, ggplot2, hardhat (>=
        1.3.0), magrittr, methods, parsnip, progress, purrr, rlang,
        stats, stringr, tibble, tidyr, torch (>= 0.4.0), tune, utils,
        vctrs, withr, zeallot
Suggests: cli, knitr, modeldata, patchwork, quarto, recipes, rmarkdown,
        rsample, spelling, testthat (>= 3.0.0), tidymodels, tidyverse,
        vip, visdat, workflows, xgboost, yardstick
VignetteBuilder: knitr
Additional_repositories: https://bgreenwell.r-universe.dev
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Config/testthat/parallel: false
Config/testthat/start-first: interface, explain, params
Encoding: UTF-8
Language: en-US
NeedsCompilation: no
Packaged: 2026-07-22 06:47:08 UTC; creg
Author: Daniel Falbel [aut],
  RStudio [cph],
  Christophe Regouby [cre, ctb],
  Egill Fridgeirsson [ctb],
  Philipp Haarmeyer [ctb],
  Sven Verweij [ctb] (ORCID: <https://orcid.org/0000-0002-5573-3952>)
Maintainer: Christophe Regouby <christophe.regouby@free.fr>
Repository: CRAN
Date/Publication: 2026-07-22 07:10:08 UTC
Built: R 4.5.2; ; 2026-07-22 11:49:53 UTC; unix

Package: tabnet
Title: Fit 'TabNet' Models for Classification and Regression
Version: 0.9.0
Authors@R: c(
    person("Daniel", "Falbel", , "daniel@rstudio.com", role = "aut"),
    person(, "RStudio", role = "cph"),
    person("Christophe", "Regouby", , "christophe.regouby@free.fr", role = c("cre", "ctb")),
    person("Egill", "Fridgeirsson", role = "ctb"),
    person("Philipp", "Haarmeyer", role = "ctb"),
    person("Sven", "Verweij", role = "ctb",
           comment = c(ORCID = "0000-0002-5573-3952"))
  )
Description: Implements the 'TabNet' model by Sercan O. Arik et al. (2019)
    <doi:10.48550/arXiv.1908.07442> with 'Coherent Hierarchical Multi-label
    Classification Networks' by Giunchiglia et al.  <doi:10.48550/arXiv.2010.10151> and
    provides a consistent interface for fitting and creating predictions.
    It's also fully compatible with the 'tidymodels' ecosystem.
License: MIT + file LICENSE
URL: https://mlverse.github.io/tabnet/,
        https://github.com/mlverse/tabnet
BugReports: https://github.com/mlverse/tabnet/issues
Depends: R (>= 3.6)
Imports: coro, data.tree, dials, dplyr, generics, ggplot2, hardhat (>=
        1.3.0), magrittr, methods, parsnip, progress, purrr, rlang,
        stats, stringr, tibble, tidyr, torch (>= 0.4.0), tune, utils,
        vctrs, withr, zeallot
Suggests: cli, knitr, modeldata, patchwork, quarto, recipes, rmarkdown,
        rsample, spelling, testthat (>= 3.0.0), tidymodels, tidyverse,
        vip, visdat, workflows, xgboost, yardstick
VignetteBuilder: knitr
Config/testthat/edition: 3
Config/testthat/parallel: false
Config/testthat/start-first: interface, explain, params
Encoding: UTF-8
Language: en-US
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-06-12 15:07:39 UTC; creg
Author: Daniel Falbel [aut],
  RStudio [cph],
  Christophe Regouby [cre, ctb],
  Egill Fridgeirsson [ctb],
  Philipp Haarmeyer [ctb],
  Sven Verweij [ctb] (ORCID: <https://orcid.org/0000-0002-5573-3952>)
Maintainer: Christophe Regouby <christophe.regouby@free.fr>
Repository: CRAN
Date/Publication: 2026-06-12 16:00:02 UTC
Built: R 4.6.0; ; 2026-06-12 17:43:34 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(augment,tabnet_fit)
S3method(min_grid,tabnet)
S3method(multi_predict,"_tabnet_fit")
S3method(nn_prune_head,tabnet_fit)
S3method(nn_prune_head,tabnet_pretrain)
S3method(predict,tabnet_fit)
S3method(print,tabnet_fit)
S3method(print,tabnet_pretrain)
S3method(tabnet_explain,default)
S3method(tabnet_explain,model_fit)
S3method(tabnet_explain,tabnet_fit)
S3method(tabnet_explain,tabnet_pretrain)
S3method(tabnet_fit,Node)
S3method(tabnet_fit,data.frame)
S3method(tabnet_fit,default)
S3method(tabnet_fit,formula)
S3method(tabnet_fit,recipe)
S3method(tabnet_pretrain,Node)
S3method(tabnet_pretrain,data.frame)
S3method(tabnet_pretrain,default)
S3method(tabnet_pretrain,formula)
S3method(tabnet_pretrain,recipe)
S3method(update,tabnet)
export("%>%")
export(attention_width)
export(augment)
export(build_ancestor_matrix_from_outcomes)
export(cat_emb_dim)
export(check_compliant_node)
export(checkpoint_epochs)
export(decision_width)
export(drop_last)
export(encoder_activation)
export(entmax)
export(entmax15)
export(feature_reusage)
export(lr_scheduler)
export(mask_type)
export(mlp_activation)
export(mlp_hidden_multiplier)
export(momentum)
export(nn_aum_loss)
export(nn_mc_loss)
export(nnf_mc_loss)
export(nnf_multilabel_one_hot)
export(node_to_df)
export(num_independent)
export(num_independent_decoder)
export(num_shared)
export(num_shared_decoder)
export(num_steps)
export(optimizer)
export(penalty)
export(sparsemax)
export(sparsemax15)
export(tabnet)
export(tabnet_config)
export(tabnet_explain)
export(tabnet_fit)
export(tabnet_nn)
export(tabnet_pretrain)
export(verbose)
export(virtual_batch_size)
importFrom(dplyr,filter)
importFrom(dplyr,last_col)
importFrom(dplyr,mutate)
importFrom(dplyr,mutate_all)
importFrom(dplyr,mutate_if)
importFrom(dplyr,select)
importFrom(dplyr,starts_with)
importFrom(dplyr,where)
importFrom(generics,augment)
importFrom(magrittr,"%>%")
importFrom(parsnip,multi_predict)
importFrom(rlang,.data)
importFrom(stats,predict)
importFrom(stats,setNames)
importFrom(stats,update)
importFrom(tidyr,replace_na)
importFrom(torch,as_array)
importFrom(torch,nn_module)
importFrom(torch,nn_prune_head)
importFrom(torch,nnf_binary_cross_entropy_with_logits)
importFrom(torch,torch_argsort)
importFrom(torch,torch_cat)
importFrom(torch,torch_long)
importFrom(torch,torch_matmul)
importFrom(torch,torch_mean)
importFrom(torch,torch_minimum)
importFrom(torch,torch_mul)
importFrom(torch,torch_std)
importFrom(torch,torch_sum)
importFrom(torch,torch_tensor)
importFrom(tune,min_grid)
importFrom(zeallot,"%<-%")

# README

# TabNet : Attentive Interpretable Tabular Learning

This is a pyTorch implementation of Tabnet (Arik, S. O., & Pfister, T. (2019). TabNet: Attentive Interpretable Tabular Learning. arXiv preprint arXiv:1908.07442.) https://arxiv.org/pdf/1908.07442.pdf. Please note that some different choices have been made overtime to improve the library which can differ from the orginal paper.

<!--- BADGES: START --->
[![CircleCI](https://circleci.com/gh/dreamquark-ai/tabnet.svg?style=svg)](https://circleci.com/gh/dreamquark-ai/tabnet)

[![PyPI version](https://badge.fury.io/py/pytorch-tabnet.svg)](https://badge.fury.io/py/pytorch-tabnet)

![PyPI - Downloads](https://img.shields.io/pypi/dm/pytorch-tabnet)

[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/pytorch-tabnet?logo=pypi&style=flat&color=blue)][#pypi-package]

[![Conda - Platform](https://img.shields.io/conda/pn/conda-forge/pytorch-tabnet?logo=anaconda&style=flat)][#conda-forge-package]

[![Conda (channel only)](https://img.shields.io/conda/vn/conda-forge/pytorch-tabnet?logo=anaconda&style=flat&color=orange)][#conda-forge-package]

[![GitHub - License](https://img.shields.io/github/license/dreamquark-ai/tabnet?logo=github&style=flat&color=green)][#github-license]

[#github-license]: https://github.com/dreamquark-ai/tabnet/blob/main/LICENSE
[#pypi-package]: https://pypi.org/project/pytorch-tabnet/
[#conda-forge-package]: https://anaconda.org/conda-forge/pytorch-tabnet
<!--- BADGES: END --->

Any questions ? Want to contribute ? To talk with us ? You can join us on [Slack](https://join.slack.com/t/mltooling/shared_invite/zt-fxaj0qk7-SWy2_~EWyhj4x9SD6gbRvg)

# Installation

## Easy installation

You can install using `pip` or `conda` as follows.

**with pip**

```sh
pip install pytorch-tabnet
```

**with conda**

```sh
conda install -c conda-forge pytorch-tabnet
```

## Source code

If you wan to use it locally within a docker container:

- `git clone git@github.com:dreamquark-ai/tabnet.git`

- `cd tabnet` to get inside the repository

-----------------

#### CPU only

- `make start` to build and get inside the container

#### GPU

- `make start-gpu` to build and get inside the GPU container

-----------------

- `poetry install` to install all the dependencies, including jupyter

- `make notebook` inside the same terminal. You can then follow the link to a jupyter notebook with tabnet installed.

# What is new ?

- from version **> 4.0** attention is now embedding aware. This aims to maintain a good attention mechanism even with large number of embedding. It is also now possible to specify attention groups (using `grouped_features`). Attention is now done at the group level and not feature level. This is especially useful if a dataset has a lot of columns coming from on single source of data (exemple: a text column transformed using TD-IDF).

# Contributing

When contributing to the TabNet repository, please make sure to first discuss the change you wish to make via a new or already existing issue.

Our commits follow the rules presented [here](https://www.conventionalcommits.org/en/v1.0.0/).

# What problems does pytorch-tabnet handle?

- TabNetClassifier : binary classification and multi-class classification problems
- TabNetRegressor : simple and multi-task regression problems
- TabNetMultiTaskClassifier:  multi-task multi-classification problems

# How to use it?

TabNet is now scikit-compatible, training a TabNetClassifier or TabNetRegressor is really easy.

```python
from pytorch_tabnet.tab_model import TabNetClassifier, TabNetRegressor

clf = TabNetClassifier()  #TabNetRegressor()
clf.fit(
  X_train, Y_train,
  eval_set=[(X_valid, y_valid)]
)
preds = clf.predict(X_test)
```

or for TabNetMultiTaskClassifier :

```python
from pytorch_tabnet.multitask import TabNetMultiTaskClassifier
clf = TabNetMultiTaskClassifier()
clf.fit(
  X_train, Y_train,
  eval_set=[(X_valid, y_valid)]
)
preds = clf.predict(X_test)
```

The targets on `y_train/y_valid` should contain a unique type (e.g. they must all be strings or integers).

### Default eval_metric

A few classic evaluation metrics are implemented (see further below for custom ones):
- binary classification metrics : 'auc', 'accuracy', 'balanced_accuracy', 'logloss'
- multiclass classification : 'accuracy', 'balanced_accuracy', 'logloss'
- regression: 'mse', 'mae', 'rmse', 'rmsle'


Important Note : 'rmsle' will automatically clip negative predictions to 0, because the model can predict negative values.
In order to match the given scores, you need to use `np.clip(clf.predict(X_predict), a_min=0, a_max=None)` when doing predictions.


### Custom evaluation metrics

You can create a metric for your specific need. Here is an example for gini score (note that you need to specifiy whether this metric should be maximized or not):

```python
from pytorch_tabnet.metrics import Metric
from sklearn.metrics import roc_auc_score

class Gini(Metric):
    def __init__(self):
        self._name = "gini"
        self._maximize = True

    def __call__(self, y_true, y_score):
        auc = roc_auc_score(y_true, y_score[:, 1])
        return max(2*auc - 1, 0.)

clf = TabNetClassifier()
clf.fit(
  X_train, Y_train,
  eval_set=[(X_valid, y_valid)],
  eval_metric=[Gini]
)

```

A specific customization example notebook is available here : https://github.com/dreamquark-ai/tabnet/blob/develop/customizing_example.ipynb

# Semi-supervised pre-training

Added later to TabNet's original paper, semi-supervised pre-training is now available via the class `TabNetPretrainer`:

```python
# TabNetPretrainer
unsupervised_model = TabNetPretrainer(
    optimizer_fn=torch.optim.Adam,
    optimizer_params=dict(lr=2e-2),
    mask_type='entmax' # "sparsemax"
)

unsupervised_model.fit(
    X_train=X_train,
    eval_set=[X_valid],
    pretraining_ratio=0.8,
)

clf = TabNetClassifier(
    optimizer_fn=torch.optim.Adam,
    optimizer_params=dict(lr=2e-2),
    scheduler_params={"step_size":10, # how to use learning rate scheduler
                      "gamma":0.9},
    scheduler_fn=torch.optim.lr_scheduler.StepLR,
    mask_type='sparsemax' # This will be overwritten if using pretrain model
)

clf.fit(
    X_train=X_train, y_train=y_train,
    eval_set=[(X_train, y_train), (X_valid, y_valid)],
    eval_name=['train', 'valid'],
    eval_metric=['auc'],
    from_unsupervised=unsupervised_model
)
```

The loss function has been normalized to be independent of `pretraining_ratio`, `batch_size` and the number of features in the problem.
A self supervised loss greater than 1 means that your model is reconstructing worse than predicting the mean for each feature, a loss bellow 1 means that the model is doing better than predicting the mean.

A complete example can be found within the notebook `pretraining_example.ipynb`.

/!\ : current implementation is trying to reconstruct the original inputs, but Batch Normalization applies a random transformation that can't be deduced by a single line, making the reconstruction harder. Lowering the `batch_size` might make the pretraining easier.

# Data augmentation on the fly

It is now possible to apply custom data augmentation pipeline during training.
Templates for ClassificationSMOTE and RegressionSMOTE have been added in `pytorch-tabnet/augmentations.py` and can be used as is.


# Easy saving and loading

It's really easy to save and re-load a trained model, this makes TabNet production ready.
```
# save tabnet model
saving_path_name = "./tabnet_model_test_1"
saved_filepath = clf.save_model(saving_path_name)

# define new model with basic parameters and load state dict weights
loaded_clf = TabNetClassifier()
loaded_clf.load_model(saved_filepath)
```

# Useful links

- [explanatory video](https://youtu.be/ysBaZO8YmX8)
- [binary classification examples](https://github.com/dreamquark-ai/tabnet/blob/develop/census_example.ipynb)
- [multi-class classification examples](https://github.com/dreamquark-ai/tabnet/blob/develop/forest_example.ipynb)
- [regression examples](https://github.com/dreamquark-ai/tabnet/blob/develop/regression_example.ipynb)
- [multi-task regression examples](https://github.com/dreamquark-ai/tabnet/blob/develop/multi_regression_example.ipynb)
- [multi-task multi-class classification examples](https://www.kaggle.com/optimo/tabnetmultitaskclassifier)
- [kaggle moa 1st place solution using tabnet](https://www.kaggle.com/c/lish-moa/discussion/201510)

## Model parameters

- `n_d` : int (default=8)

    Width of the decision prediction layer. Bigger values gives more capacity to the model with the risk of overfitting.
    Values typically range from 8 to 64.

- `n_a`: int (default=8)

    Width of the attention embedding for each mask.
    According to the paper n_d=n_a is usually a good choice. (default=8)

- `n_steps` : int (default=3)

    Number of steps in the architecture (usually between 3 and 10)  

- `gamma` : float  (default=1.3)

    This is the coefficient for feature reusage in the masks.
    A value close to 1 will make mask selection least correlated between layers.
    Values range from 1.0 to 2.0.

- `cat_idxs` : list of int (default=[] - Mandatory for embeddings) 

    List of categorical features indices.

- `cat_dims` : list of int (default=[] - Mandatory for embeddings)

    List of categorical features number of modalities (number of unique values for a categorical feature)
    /!\ no new modalities can be predicted

- `cat_emb_dim` : list of int (optional)

    List of embeddings size for each categorical features. (default =1)

- `n_independent` : int  (default=2)

    Number of independent Gated Linear Units layers at each step.
    Usual values range from 1 to 5.

- `n_shared` : int (default=2)

    Number of shared Gated Linear Units at each step
    Usual values range from 1 to 5

- `epsilon` : float  (default 1e-15)

    Should be left untouched.

- `seed` : int (default=0)

    Random seed for reproducibility

- `momentum` : float

    Momentum for batch normalization, typically ranges from 0.01 to 0.4 (default=0.02)

- `clip_value` : float (default None)

    If a float is given this will clip the gradient at clip_value.
    
- `lambda_sparse` : float (default = 1e-3)

    This is the extra sparsity loss coefficient as proposed in the original paper. The bigger this coefficient is, the sparser your model will be in terms of feature selection. Depending on the difficulty of your problem, reducing this value could help.

- `optimizer_fn` : torch.optim (default=torch.optim.Adam)

    Pytorch optimizer function

- `optimizer_params`: dict (default=dict(lr=2e-2))

    Parameters compatible with optimizer_fn used initialize the optimizer. Since we have Adam as our default optimizer, we use this to define the initial learning rate used for training. As mentionned in the original paper, a large initial learning rate of ```0.02 ```  with decay is a good option.

- `scheduler_fn` : torch.optim.lr_scheduler (default=None)

    Pytorch Scheduler to change learning rates during training.

- `scheduler_params` : dict

    Dictionnary of parameters to apply to the scheduler_fn. Ex : {"gamma": 0.95, "step_size": 10}

- `model_name` : str (default = 'DreamQuarkTabNet')

    Name of the model used for saving in disk, you can customize this to easily retrieve and reuse your trained models.

- `verbose` : int (default=1)

    Verbosity for notebooks plots, set to 1 to see every epoch, 0 to get None.

- `device_name` : str (default='auto')
    'cpu' for cpu training, 'gpu' for gpu training, 'auto' to automatically detect gpu.

- `mask_type: str` (default='sparsemax')
    Either "sparsemax" or "entmax" : this is the masking function to use for selecting features.

- `grouped_features: list of list of ints` (default=None)
    This allows the model to share it's attention accross feature inside a same group.
    This can be especially useful when your preprocessing generates correlated or dependant features: like if you use a TF-IDF or a PCA on a text column.
    Note that feature importance will be exactly the same between features on a same group.
    Please also note that embeddings generated for a categorical variable are always inside a same group. 

- `n_shared_decoder` : int (default=1)

    Number of shared GLU block in decoder, this is only useful for `TabNetPretrainer`.

- `n_indep_decoder` : int (default=1)

    Number of independent GLU block in decoder, this is only useful for `TabNetPretrainer`.

## Fit parameters

- `X_train` : np.array or scipy.sparse.csr_matrix

    Training features

- `y_train` : np.array

    Training targets

- `eval_set`: list of tuple  

    List of eval tuple set (X, y).  
    The last one is used for early stopping  

- `eval_name`: list of str  
              List of eval set names.  

- `eval_metric` : list of str  
              List of evaluation metrics.  
              The last metric is used for early stopping.

- `max_epochs` : int (default = 200)

    Maximum number of epochs for trainng.
    
- `patience` : int (default = 10)

    Number of consecutive epochs without improvement before performing early stopping.

    If patience is set to 0, then no early stopping will be performed.

    Note that if patience is enabled, then best weights from best epoch will automatically be loaded at the end of `fit`.

- `weights` : int or dict (default=0)

    /!\ Only for TabNetClassifier
    Sampling parameter
    0 : no sampling
    1 : automated sampling with inverse class occurrences
    dict : keys are classes, values are weights for each class

- `loss_fn` : torch.loss or list of torch.loss

    Loss function for training (default to mse for regression and cross entropy for classification)
    When using TabNetMultiTaskClassifier you can set a list of same length as number of tasks,
    each task will be assigned its own loss function

- `batch_size` : int (default=1024)

    Number of examples per batch. Large batch sizes are recommended.

- `virtual_batch_size` : int (default=128)

    Size of the mini batches used for "Ghost Batch Normalization".
    /!\ `virtual_batch_size` should divide `batch_size`

- `num_workers` : int (default=0)

    Number or workers used in torch.utils.data.Dataloader

- `drop_last` : bool (default=False)

    Whether to drop last batch if not complete during training

- `callbacks` : list of callback function  
        List of custom callbacks

- `pretraining_ratio` : float

        /!\ TabNetPretrainer Only : Percentage of input features to mask during pretraining.

        Should be between 0 and 1. The bigger the harder the reconstruction task is.

- `warm_start` : bool (default=False)
    In order to match scikit-learn API, this is set to False.
    It allows to fit twice the same model and start from a warm start.

- `compute_importance` : bool (default=True)

    Whether to compute feature importance

<!-- README.md is generated from README.Rmd. Please edit that file -->

# tabnet

<!-- badges: start -->

[![R build
status](https://github.com/mlverse/tabnet/workflows/R-CMD-check/badge.svg)](https://github.com/mlverse/tabnet/actions)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![CRAN
status](https://www.r-pkg.org/badges/version/tabnet)](https://CRAN.R-project.org/package=tabnet)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/tabnet)](https://cran.r-project.org/package=tabnet)
[![Discord](https://img.shields.io/discord/837019024499277855?logo=discord)](https://discord.com/invite/s3D5cKhBkx)

<!-- badges: end -->

An R implementation of: [TabNet: Attentive Interpretable Tabular
Learning](https://arxiv.org/abs/1908.07442) [(Sercan O. Arik, Tomas
Pfister)](https://doi.org/10.48550/arXiv.1908.07442).  

The code in this repository started by an R port using the
[torch](https://github.com/mlverse/torch) package of
[dreamquark-ai/tabnet](https://github.com/dreamquark-ai/tabnet)
implementation.

TabNet is now augmented with

- [Coherent Hierarchical Multi-label Classification
  Networks](https://proceedings.neurips.cc//paper/2020/file/6dd4e10e3296fa63738371ec0d5df818-Paper.pdf)
  [(Eleonora Giunchiglia et
  Al.)](https://doi.org/10.48550/arXiv.2010.10151) for hierarchical
  outcomes

- [Optimizing ROC Curves with a Sort-Based Surrogate Loss for Binary
  Classification and Changepoint Detection (J Hillman, TD
  Hocking)](https://jmlr.org/papers/v24/21-0751.html) for imbalanced
  binary classification.

## Installation

Install [{tabnet} from CRAN](https://CRAN.R-project.org/package=tabnet)
with:

``` r
install.packages("tabnet")
```

The development version can be installed from
[GitHub](https://github.com/mlverse/tabnet) with:

``` r
# install.packages("pak")
pak::pak("mlverse/tabnet")
```

## Basic Binary Classification Example

Here we show a **binary classification** example of the `attrition`
dataset, using a **recipe** for dataset input specification.

``` r
library(tabnet)
suppressPackageStartupMessages(library(recipes))
library(yardstick)
library(ggplot2)
torch::torch_manual_seed(2026)

data("attrition", package = "modeldata")
test_idx <- sample.int(nrow(attrition), size = 0.2 * nrow(attrition))

train <- attrition[-test_idx,]
test <- attrition[test_idx,]

rec <- recipe(Attrition ~ ., data = train) %>% 
  step_normalize(all_numeric(), -all_outcomes())

fit <- tabnet_fit(rec, train, epochs = 30, valid_split=0.1, learn_rate = 5e-3)
autoplot(fit)
```

<img src="man/figures/README-model-fit-1.png" alt="A training loss line-plot along training epochs. Both validation loss and training loss are shown. Training loss line includes regular dots at epochs where a checkpoint is recorded." width="100%" />

The plots gives you an immediate insight about model over-fitting, and
if any, the available model checkpoints available before the
over-fitting

Keep in mind that **regression** as well as **multi-class
classification** are also available, and that you can specify dataset
through **data.frame** and **formula** as well. You will find them in
the package vignettes.

## Model performance results

As the standard method `predict()` is used, you can rely on your usual
metric functions for model performance results. Here we use {yardstick}
:

``` r
metrics <- metric_set(accuracy, precision, recall)

augment(fit, test) %>% 
  metrics(Attrition, estimate = .pred_class)
#> # A tibble: 3 × 3
#>   .metric   .estimator .estimate
#>   <chr>     <chr>          <dbl>
#> 1 accuracy  binary         0.847
#> 2 precision binary         0.853
#> 3 recall    binary         0.992
  
augment(fit, test, type = "prob") %>% 
  roc_auc(Attrition, .pred_No)
#> # A tibble: 1 × 3
#>   .metric .estimator .estimate
#>   <chr>   <chr>          <dbl>
#> 1 roc_auc binary         0.384
```

## Explain model on test-set with attention map

TabNet has intrinsic explainability feature through the visualization of
attention map, either **aggregated**:

``` r
explain <- tabnet_explain(fit, test)
autoplot(explain)
```

<img src="man/figures/README-model-explain-1.png" alt="An heatmap as explainability plot showing for each variable of the test-set on the y axis the importance along each observation on the x axis. The value is a mask agggregate." width="100%" />

or at **each layer** through the `type = "steps"` option:

``` r
autoplot(explain, type = "steps")
```

<img src="man/figures/README-step-explain-1.png" alt="An small-multiple heatmap as explainability plot for each step of the Tabnet network. Each plot shows for each variable of the test-set on the y axis the importance along each observation on the x axis." width="100%" />

## Self-supervised pretraining

For cases when a consistent part of your dataset has no outcome, TabNet
offers a self-supervised training step allowing to model to capture
predictors intrinsic features and predictors interactions, upfront the
supervised task.

``` r
pretrain <- tabnet_pretrain(rec, train, epochs = 50, valid_split=0.1, learn_rate = 1e-2)
autoplot(pretrain)
```

<img src="man/figures/README-step-pretrain-1.png" alt="A training loss line-plot along pre-training epochs. Both validation loss and training loss are shown. Training loss line includes regular dots at epochs where a checkpoint is recorded." width="100%" />

The example here is a toy example as the `train` dataset does actually
contain outcomes. The vignette
[`vignette("selfsupervised_training")`](https://mlverse.github.io/tabnet/articles/selfsupervised_training.html)
will gives you the complete correct workflow step-by-step.

## {tidymodels} integration

The integration within tidymodels workflows offers you unlimited
opportunity to compare {tabnet} models with challengers.

Don’t miss the
[`vignette("tidymodels-interface")`](https://mlverse.github.io/tabnet/articles/tidymodels-interface.html)
for that.

## Missing data in predictors

{tabnet} leverage the masking mechanism to deal with missing data, so
you don’t have to remove the entries in your dataset with some missing
values in the predictors variables.

See
[`vignette("Missing_data_predictors")`](https://mlverse.github.io/tabnet/articles/Missing_data_predictors.html)

## Imbalanced binary classification

{tabnet} includes a Area under the $Min(FPR,FNR)$ (AUM) loss function
`nn_aum_loss()` dedicated to your imbalanced binary classification
tasks.

Try it out in
[`vignette("aum_loss")`](https://mlverse.github.io/tabnet/articles/aum_loss.html)

# Comparison with other implementations

| Group | Feature | {tabnet} | dreamquark-ai | fast-tabnet |
|----|----|:--:|:--:|:--:|
| Input format | data-frame | ✅ | ✅ | ✅ |
|  | formula | ✅ |  |  |
|  | recipe | ✅ |  |  |
|  | Node | ✅ |  |  |
|  | missings in predictor | ✅ |  |  |
| Output format | data-frame | ✅ | ✅ | ✅ |
|  | workflow | ✅ |  |  |
| ML Tasks | self-supervised learning | ✅ | ✅ |  |
|  | classification (binary, multi-class) | ✅ | ✅ | ✅ |
|  | unbalanced binary classification | ✅ |  |  |
|  | regression | ✅ | ✅ | ✅ |
|  | multi-outcome | ✅ | ✅ |  |
|  | hierarchical multi-label classif. | ✅ |  |  |
| Model management | from / to file | ✅ | ✅ | v |
|  | resume from snapshot | ✅ |  |  |
|  | training diagnostic | ✅ |  |  |
| Interpretability |  | ✅ | ✅ | ✅ |
| Performance |  | 1 x | 2 - 4 x |  |
| Code quality | test coverage | 85% |  |  |
|  | continuous integration | 4 OS including GPU |  |  |

Alternative TabNet implementation features

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)

## -----------------------------------------------------------------------------
# library(tidymodels, quietly = TRUE)
# library(tabnet)
# data("ames", package = "modeldata")
# qplot(ames$Mas_Vnr_Area)

## -----------------------------------------------------------------------------
# col_with_zero_as_na <- ames %>%
#   select(where(is.numeric)) %>%
#   select(matches("_SF|Area|Misc_Val|[Pp]orch$")) %>%
#   summarise_each(min) %>%
#   select_if(~.x==0) %>%
#   names()
# ames_missing <- ames %>% mutate_at(col_with_zero_as_na, na_if, 0) %>%
#   mutate_at("Alley", na_if, "No_Alley_Access") %>%
#   mutate_at("Fence", na_if, "No_Fence") %>%
#   mutate_at(c("Garage_Cond", "Garage_Finish"), na_if, "No_Garage") %>%
#   mutate_at(c("Bsmt_Exposure", "BsmtFin_Type_1", "BsmtFin_Type_2"), na_if, "No_Basement")
# 
# visdat::vis_miss(ames_missing)

## -----------------------------------------------------------------------------
# ames_rec <- recipe(Sale_Price ~ ., data=ames) %>%
#   step_normalize(all_numeric())
# 
# cat_emb_dim <- map_dbl(ames %>% select_if(is.factor), ~log2(nlevels(.x)) %>% round)
# 
# ames_pretrain <- tabnet_pretrain(ames_rec, data=ames,  epoch=50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose=TRUE, batch=2930,
#                             early_stopping_patience = 3L, early_stopping_tolerance = 1e-4)
# autoplot(ames_pretrain)

## -----------------------------------------------------------------------------
# col_with_missings <- ames_missing %>%
#   summarise_all(~sum(is.na(.)) > 0) %>%
#   t %>% enframe(name = "Variable") %>%
#   rename(has_missing = "value")
# 
# vip_color <- function(object, col_has_missing) {
#   vip_data <- vip::vip(object)$data %>% arrange(Importance)
#   vis_miss_plus <- left_join(vip_data, col_has_missing , by = "Variable") %>%
#     mutate(Variable = factor(Variable, levels = vip_data$Variable))
#   vis_miss_plus
#   ggplot(vis_miss_plus, aes(x = Variable, y = Importance, fill = has_missing)) +
#     geom_col() + coord_flip() + scale_fill_grey()
# }
# vip_color(ames_pretrain, col_with_missings)

## -----------------------------------------------------------------------------
# ames_missing_rec <- recipe(Sale_Price ~ ., data = ames_missing) %>%
#   step_normalize(all_numeric())
# ames_missing_pretrain <- tabnet_pretrain(ames_missing_rec, data = ames_missing, epoch = 50,
#                                     cat_emb_dim = cat_emb_dim,
#                                     valid_split = 0.2, verbose = TRUE, batch = 2930,
#                                     pretraining_ratio = 0.37,
#                                     early_stopping_patience = 3L, early_stopping_tolerance = 1e-4)
# autoplot(ames_missing_pretrain)
# vip_color(ames_missing_pretrain, col_with_missings)

## -----------------------------------------------------------------------------
# ames_fit <- tabnet_fit(ames_rec, data = ames,  tabnet_model = ames_pretrain,
#                             epoch = 50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose = TRUE, batch = 2930,
#                             early_stopping_patience = 5L, early_stopping_tolerance = 1e-4)
# autoplot(ames_fit)
# vip_color(ames_fit, col_with_missings)

## -----------------------------------------------------------------------------
# ames_missing_fit <- tabnet_fit(ames_rec, data = ames_missing,  tabnet_model = ames_missing_pretrain,
#                             epoch = 50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose = TRUE, batch = 2930,
#                             early_stopping_patience = 5L, early_stopping_tolerance = 1e-4)
# autoplot(ames_missing_fit)
# vip_color(ames_missing_fit, col_with_missings)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
requireNamespace("vip", quietly = TRUE)

## ----setup--------------------------------------------------------------------
# library(tabnet)
# library(dplyr)
# library(data.tree)
# library(ggplot2)
# library(rsample)
# library(tibble)
# set.seed(202307)

## -----------------------------------------------------------------------------
# data(acme, package = "data.tree")
# acme$attributesAll
# print(acme, "cost", "p" , limit = 8)

## -----------------------------------------------------------------------------
# data(starwars, package = "dplyr")
# head(starwars, 4)
# 
# # erroneous Node construction
# starwars_tree <- starwars %>%
#   mutate(pathString = paste("StarWars_characters", species, sex, `name`, sep = "/")) %>%
#   as.Node()
# print(starwars_tree, "name","height", "mass", "eye_color", limit = 8)
# 

## -----------------------------------------------------------------------------
# # demonstration of reserved column modification in Node construction
# starwars_tree <- starwars %>%
#   rename(`_name` = "name", `_height` = "height") %>%
#   mutate(
#     species = coalesce(species, "Unknown_Species"),
#     sex     = coalesce(sex, "Unknown_Sex"),
#     pathString = paste("StarWars_characters", species, sex, `_name`, sep = "/")) %>%
#   as.Node()
# print(starwars_tree, "name", "_name","_height", "mass", "eye_color", limit = 8)

## -----------------------------------------------------------------------------
# starw_split <- starwars %>%
#   tidyr::unnest_longer(films) %>%
#   tidyr::unnest_longer(vehicles, keep_empty = TRUE) %>%
#   tidyr::unnest_longer(starships, keep_empty = TRUE) %>%
#   initial_split( prop = .8, strata = "species")

## -----------------------------------------------------------------------------
# # correct Node construction for hierarchical modeling
# starwars_train_tree <- starw_split %>%
#   training() %>%
#   # avoid reserved column names
#   rename(`_name` = "name", `_height` = "height") %>%
#   rowid_to_column() %>%
#   mutate(pathString = paste("StarWars_characters", species, sex, rowid, sep = "/")) %>%
#   # remove outcomes labels from predictors
#   select(-species, -sex, -`_name`, -rowid) %>%
#   # turn it as hierarchical Node
#   as.Node()
# 
# starwars_test_tree <- starw_split %>%
#   testing() %>%
#   rename(`_name` = "name", `_height` = "height") %>%
#   rowid_to_column() %>%
#   mutate(pathString = paste("StarWars_characters", species, sex, rowid, sep = "/")) %>%
#   select(-species, -sex, -`_name`, -rowid) %>%
#   as.Node()
# 
# starwars_train_tree$attributesAll

## ----model training-----------------------------------------------------------
# config <- tabnet_config(decision_width = 8, attention_width = 8, num_steps = 3, penalty = .003, cat_emb_dim = 2, valid_split = 0.2, learn_rate = 1e-3, lr_scheduler = "reduce_on_plateau", early_stopping_monitor = "valid_loss", early_stopping_patience = 4, verbose = FALSE)
# 
# starw_model <- tabnet_fit(starwars_train_tree, config = config, epoch = 170, checkpoint_epochs = 25)
# 

## -----------------------------------------------------------------------------
# autoplot(starw_model)

## -----------------------------------------------------------------------------
# vip::vip(starw_model)

## -----------------------------------------------------------------------------
# starwars_hat <- augment(starw_model, starwars_test_tree)
# tail(starwars_hat, n = 5)
# 

## -----------------------------------------------------------------------------
# starwars_explain <- tabnet_explain(starw_model, starwars_test_tree)
# autoplot(starwars_explain)
# autoplot(starwars_explain, type = "steps")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = torch::torch_is_installed()
)

## ----setup--------------------------------------------------------------------
library(tabnet)
suppressPackageStartupMessages(library(tidymodels))
library(modeldata)
data("lending_club", package = "modeldata")
set.seed(20250809)

## -----------------------------------------------------------------------------
class_ratio <- lending_club |> 
  summarise(sum( Class == "good") / sum( Class == "bad")) |> 
  pull() 

class_ratio

## -----------------------------------------------------------------------------
lending_club <- lending_club |>
  mutate(
    case_wts = if_else(Class == "bad", class_ratio, 1),
    case_wts = importance_weights(case_wts)
  )

split <- initial_split(lending_club, strata = Class)
train <- training(split)
test  <- testing(split)

tab_rec <- train |>
  recipe() |>
  update_role(Class, new_role = "outcome") |>
  update_role(-has_role(c("outcome", "id", "case_weights")), new_role = "predictor")

xgb_rec <- tab_rec |> 
  step_dummy(term, sub_grade, addr_state, verification_status, emp_length)

tab_mod <- tabnet(epochs = 100) |> 
  set_engine("torch", device = "cpu") |> 
  set_mode("classification")

xgb_mod <- boost_tree(trees = 100) |> 
  set_engine("xgboost") |> 
  set_mode("classification")

tab_wf <- workflow() |> 
  add_model(tab_mod) |> 
  add_recipe(tab_rec) |> 
  add_case_weights(case_wts)

xgb_wf <- workflow() |> 
  add_model(xgb_mod) |> 
  add_recipe(xgb_rec) |> 
  add_case_weights(case_wts)

## -----------------------------------------------------------------------------
tab_fit <- tab_wf |> fit(train)
xgb_fit <- xgb_wf |> fit(train)

tab_test <- tab_fit |> augment(test)
xgb_test <- xgb_fit |> augment(test)

tab_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot()

xgb_test |>
  pr_curve(Class, .pred_good) |>
  autoplot()


## -----------------------------------------------------------------------------
tab_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 

xgb_test |>
  pr_curve(Class, .pred_good, case_weights = case_wts) |>
  autoplot() 

## -----------------------------------------------------------------------------
# configure the AUM loss
tab_aum_mod <- tabnet(epochs = 100, loss = tabnet::nn_aum_loss, learn_rate = 0.02) |> 
  set_engine("torch", device = "cpu") |> 
  set_mode("classification")

# derive a workflow
tab_aum_wf <- workflow() |> 
  add_model(tab_aum_mod) |> 
  add_recipe(tab_rec) |> 
  add_case_weights(case_wts)

# fit and augment the test dataset with prediction
tab_aum_fit <- tab_aum_wf |> fit(train)
tab_aum_test <- tab_aum_fit |> augment(test)

## -----------------------------------------------------------------------------
tab_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot() 

tab_aum_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot() 

## -----------------------------------------------------------------------------
tab_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 


tab_aum_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
requireNamespace("vip", quietly = TRUE)

## -----------------------------------------------------------------------------
# library(tidymodels, quietly = TRUE)
# library(tabnet)
# data("ames", package = "modeldata")
# qplot(ames$Mas_Vnr_Area)

## -----------------------------------------------------------------------------
# col_with_zero_as_na <- ames %>%
#   select(where(is.numeric)) %>%
#   select(matches("_SF|Area|Misc_Val|[Pp]orch$")) %>%
#   summarise_each(min) %>%
#   select_if(~.x==0) %>%
#   names()
# ames_missing <- ames %>% mutate_at(col_with_zero_as_na, na_if, 0) %>%
#   mutate_at("Alley", na_if, "No_Alley_Access") %>%
#   mutate_at("Fence", na_if, "No_Fence") %>%
#   mutate_at(c("Garage_Cond", "Garage_Finish"), na_if, "No_Garage") %>%
#   mutate_at(c("Bsmt_Exposure", "BsmtFin_Type_1", "BsmtFin_Type_2"), na_if, "No_Basement")
# 
# visdat::vis_miss(ames_missing)

## -----------------------------------------------------------------------------
# ames_rec <- recipe(Sale_Price ~ ., data=ames) %>%
#   step_normalize(all_numeric())
# 
# cat_emb_dim <- map_dbl(ames %>% select_if(is.factor), ~log2(nlevels(.x)) %>% round)
# 
# ames_pretrain <- tabnet_pretrain(ames_rec, data=ames,  epoch=50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose=TRUE, batch=2930,
#                             early_stopping_patience = 3L, early_stopping_tolerance = 1e-4)
# autoplot(ames_pretrain)

## -----------------------------------------------------------------------------
# col_with_missings <- ames_missing %>%
#   summarise_all(~sum(is.na(.)) > 0) %>%
#   t %>% enframe(name = "Variable") %>%
#   rename(has_missing = "value")
# 
# vip_color <- function(object, col_has_missing) {
#   vip_data <- vip::vi(object) %>% arrange(Importance)
#   vis_miss_plus <- left_join(vip_data, col_has_missing , by = "Variable") %>%
#     mutate(Variable = factor(Variable, levels = vip_data$Variable))
#   vis_miss_plus
#   ggplot(vis_miss_plus, aes(x = Variable, y = Importance, fill = has_missing)) +
#     geom_col() + coord_flip() + scale_fill_grey()
# }
# vip_color(ames_pretrain, col_with_missings)

## -----------------------------------------------------------------------------
# ames_missing_rec <- recipe(Sale_Price ~ ., data = ames_missing) %>%
#   step_normalize(all_numeric())
# ames_missing_pretrain <- tabnet_pretrain(ames_missing_rec, data = ames_missing, epoch = 50,
#                                     cat_emb_dim = cat_emb_dim,
#                                     valid_split = 0.2, verbose = TRUE, batch = 2930,
#                                     pretraining_ratio = 0.37,
#                                     early_stopping_patience = 3L, early_stopping_tolerance = 1e-4)
# autoplot(ames_missing_pretrain)
# vip_color(ames_missing_pretrain, col_with_missings)

## -----------------------------------------------------------------------------
# ames_fit <- tabnet_fit(ames_rec, data = ames,  tabnet_model = ames_pretrain,
#                             epoch = 50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose = TRUE, batch = 2930,
#                             early_stopping_patience = 5L, early_stopping_tolerance = 1e-4)
# autoplot(ames_fit)
# vip_color(ames_fit, col_with_missings)

## -----------------------------------------------------------------------------
# ames_missing_fit <- tabnet_fit(ames_rec, data = ames_missing,  tabnet_model = ames_missing_pretrain,
#                             epoch = 50, cat_emb_dim = cat_emb_dim,
#                             valid_split = 0.2, verbose = TRUE, batch = 2930,
#                             early_stopping_patience = 5L, early_stopping_tolerance = 1e-4)
# autoplot(ames_missing_fit)
# vip_color(ames_missing_fit, col_with_missings)

add_parsnip_tabnet <- function() {

  parsnip::set_new_model("tabnet")
  parsnip::set_model_mode(model = "tabnet", mode = "classification")
  parsnip::set_model_mode(model = "tabnet", mode = "regression")

  parsnip::set_model_engine(
    "tabnet",
    mode = "classification",
    eng = "torch"
  )

  parsnip::set_model_engine(
    "tabnet",
    mode = "regression",
    eng = "torch"
  )

  parsnip::set_dependency("tabnet", eng = "torch", pkg = "tabnet")

  parsnip::set_fit(
    model = "tabnet",
    eng = "torch",
    mode = "classification",
    value = list(
      interface = "formula",
      protect = c("formula", "data", "weights"),
      func = c(pkg = "tabnet", fun = "tabnet_fit"),
      defaults = list()
    )
  )

  parsnip::set_fit(
    model = "tabnet",
    eng = "torch",
    mode = "regression",
    value = list(
      interface = "formula",
      protect = c("formula", "data", "weights"),
      func = c(pkg = "tabnet", fun = "tabnet_fit"),
      defaults = list()
    )
  )

  parsnip::set_encoding(
    model = "tabnet",
    eng = "torch",
    mode = "regression",
    options = list(
      predictor_indicators = "none",
      compute_intercept = FALSE,
      remove_intercept = FALSE,
      allow_sparse_x = FALSE
    )
  )

  parsnip::set_encoding(
    model = "tabnet",
    eng = "torch",
    mode = "classification",
    options = list(
      predictor_indicators = "none",
      compute_intercept = FALSE,
      remove_intercept = FALSE,
      allow_sparse_x = FALSE
    )
  )

  make_class_info <- function(type) {
    list(
      pre = NULL,
      post = NULL,
      func = c(fun = "predict"),
      args =
        list(
          object = quote(object$fit),
          new_data = quote(new_data),
          type = type
        )
    )
  }

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "cat_emb_dim",
    original = "cat_emb_dim",
    func = list(pkg = "tabnet", fun = "cat_emb_dim"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "decision_width",
    original = "decision_width",
    func = list(pkg = "tabnet", fun = "decision_width"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "attention_width",
    original = "attention_width",
    func = list(pkg = "tabnet", fun = "attention_width"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "num_steps",
    original = "num_steps",
    func = list(pkg = "tabnet", fun = "num_steps"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "mask_type",
    original = "mask_type",
    func = list(pkg = "tabnet", fun = "mask_type"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "mask_topk",
    original = "mask_topk",
    func = list(pkg = "tabnet", fun = "mask_topk"),
    has_submodel = FALSE
  )
  
  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "mlp_hidden_multiplier",
    original = "mlp_hidden_multiplier",
    func = list(pkg = "tabnet", fun = "mlp_hidden_multiplier"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "mlp_activation",
    original = "mlp_activation",
    func = list(pkg = "tabnet", fun = "mlp_activation"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "encoder_activation",
    original = "encoder_activation",
    func = list(pkg = "tabnet", fun = "encoder_activation"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "num_independent",
    original = "num_independent",
    func = list(pkg = "tabnet", fun = "num_independent"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "num_shared",
    original = "num_shared",
    func = list(pkg = "tabnet", fun = "num_shared"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "num_independent_decoder",
    original = "num_independent_decoder",
    func = list(pkg = "tabnet", fun = "num_independent_decoder"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "num_shared_decoder",
    original = "num_shared_decoder",
    func = list(pkg = "tabnet", fun = "num_shared_decoder"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "penalty",
    original = "penalty",
    func = list(pkg = "dials", fun = "penalty"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "feature_reusage",
    original = "feature_reusage",
    func = list(pkg = "tabnet", fun = "feature_reusage"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "momentum",
    original = "momentum",
    func = list(pkg = "dials", fun = "momentum"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "epochs",
    original = "epochs",
    func = list(pkg = "dials", fun = "epochs"),
    has_submodel = TRUE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "batch_size",
    original = "batch_size",
    func = list(pkg = "dials", fun = "batch_size"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "virtual_batch_size",
    original = "virtual_batch_size",
    func = list(pkg = "tabnet", fun = "virtual_batch_size"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "learn_rate",
    original = "learn_rate",
    func = list(pkg = "dials", fun = "learn_rate"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "optimizer",
    original = "optimizer",
    func = list(pkg = "tabnet", fun = "optimizer"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "loss",
    original = "loss",
    func = list(pkg = "tabnet", fun = "loss"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "clip_value",
    original = "clip_value",
    func = list(pkg = "tabnet", fun = "clip_value"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "drop_last",
    original = "drop_last",
    func = list(pkg = "tabnet", fun = "drop_last"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "lr_scheduler",
    original = "lr_scheduler",
    func = list(pkg = "tabnet", fun = "lr_scheduler"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "rate_decay",
    original = "lr_decay",
    func = list(pkg = "dials", fun = "rate_decay"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "rate_step_size",
    original = "step_size",
    func = list(pkg = "dials", fun = "rate_step_size"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "checkpoint_epochs",
    original = "checkpoint_epochs",
    func = list(pkg = "tabnet", fun = "checkpoint_epochs"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "verbose",
    original = "verbose",
    func = list(pkg = "tabnet", fun = "verbose"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "importance_sample_size",
    original = "importance_sample_size",
    func = list(pkg = "tabnet", fun = "importance_sample_size"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "early_stopping_monitor",
    original = "early_stopping_monitor",
    func = list(pkg = "tabnet", fun = "early_stopping_monitor"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "early_stopping_tolerance",
    original = "early_stopping_tolerance",
    func = list(pkg = "tabnet", fun = "early_stopping_tolerance"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "early_stopping_patience",
    original = "early_stopping_patience",
    func = list(pkg = "tabnet", fun = "early_stopping_patience"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "skip_importance",
    original = "skip_importance",
    func = list(pkg = "dials", fun = "skip_importance"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "tabnet_model",
    original = "tabnet_model",
    func = list(pkg = "tabnet", fun = "tabnet_model"),
    has_submodel = FALSE
  )

  parsnip::set_model_arg(
    model = "tabnet",
    eng = "torch",
    parsnip = "from_epoch",
    original = "from_epoch",
    func = list(pkg = "tabnet", fun = "from_epoch"),
    has_submodel = FALSE
  )

  parsnip::set_pred(
    model = "tabnet",
    eng = "torch",
    mode = "classification",
    type = "class",
    value = make_class_info("class")
  )

  parsnip::set_pred(
    model = "tabnet",
    eng = "torch",
    mode = "classification",
    type = "prob",
    value = make_class_info("prob")
  )

  parsnip::set_pred(
    model = "tabnet",
    eng = "torch",
    mode = "regression",
    type = "numeric",
    value = make_class_info("numeric")
  )

}

#' Parsnip compatible tabnet model
#'
#' @param mode A single character string for the type of model. Possible values
#'   for this model are "unknown", "regression", or "classification".
#' @inheritParams tabnet_config
#' @inheritParams tabnet_fit
#' @param rate_decay multiplies the initial learning rate by `rate_decay` every
#'   `rate_step_size` epochs. Unused if `lr_scheduler` is a `torch::lr_scheduler`
#'   or `NULL`.
#' @param rate_step_size the learning rate scheduler step size. Unused if
#'   `lr_scheduler` is a `torch::lr_scheduler` or `NULL`.
#'
#' @inheritSection tabnet_fit Threading
#' @seealso tabnet_fit
#'
#' @examplesIf torch::torch_is_installed()
#' library(parsnip)
#' data("ames", package = "modeldata")
#' model <- tabnet() %>%
#'   set_mode("regression") %>%
#'   set_engine("torch")
#' model %>%
#'   fit(Sale_Price ~ ., data = ames)
#'
#' @return A TabNet `parsnip` instance. It can be used to fit tabnet models using
#' `parsnip` machinery.
#'
#' @export
tabnet <- function(mode = "unknown",  cat_emb_dim = NULL, decision_width = NULL, attention_width = NULL,
                   num_steps = NULL, mask_type = NULL, mask_topk = NULL, num_independent = NULL, num_shared = NULL,
                   num_independent_decoder = NULL, num_shared_decoder = NULL, penalty = NULL,
                   feature_reusage = NULL, momentum = NULL, epochs = NULL, batch_size = NULL,
                   virtual_batch_size = NULL, learn_rate = NULL, optimizer = NULL, loss = NULL,
                   clip_value = NULL, drop_last = NULL, lr_scheduler = NULL, rate_decay = NULL, rate_step_size = NULL,
                   checkpoint_epochs = NULL, verbose = NULL, importance_sample_size = NULL,
                   early_stopping_monitor = NULL, early_stopping_tolerance = NULL,
                   early_stopping_patience = NULL, skip_importance = NULL,
                   tabnet_model = NULL, from_epoch = NULL
                   ) {

  if (!requireNamespace("parsnip", quietly = TRUE))
    runtime_error("Package {.pkg parsnip} is needed for this function to work. Please install it.")
  
  if (parsnip_is_missing_tabnet(tabnet_env)) {
    add_parsnip_tabnet()
    tabnet_env$parsnip_added <- TRUE
  }


  # Capture the arguments in quosures
  args <- list(
    cat_emb_dim = rlang::enquo(cat_emb_dim),
    decision_width = rlang::enquo(decision_width),
    attention_width = rlang::enquo(attention_width),
    num_steps = rlang::enquo(num_steps),
    mask_type = rlang::enquo(mask_type),
    mask_topk = rlang::enquo(mask_topk),
    num_independent = rlang::enquo(num_independent),
    num_shared = rlang::enquo(num_shared),
    num_independent_decoder = rlang::enquo(num_independent_decoder),
    num_shared_decoder = rlang::enquo(num_shared_decoder),
    penalty = rlang::enquo(penalty),
    feature_reusage = rlang::enquo(feature_reusage),
    momentum = rlang::enquo(momentum),
    epochs = rlang::enquo(epochs),
    batch_size = rlang::enquo(batch_size),
    virtual_batch_size = rlang::enquo(virtual_batch_size),
    learn_rate = rlang::enquo(learn_rate),
    optimizer = rlang::enquo(optimizer),
    loss = rlang::enquo(loss),
    clip_value = rlang::enquo(clip_value),
    drop_last = rlang::enquo(drop_last),
    lr_scheduler = rlang::enquo(lr_scheduler),
    lr_decay = rlang::enquo(rate_decay),
    step_size = rlang::enquo(rate_step_size),
    checkpoint_epochs = rlang::enquo(checkpoint_epochs),
    verbose = rlang::enquo(verbose),
    importance_sample_size = rlang::enquo(importance_sample_size),
    early_stopping_monitor = rlang::enquo(early_stopping_monitor),
    early_stopping_tolerance = rlang::enquo(early_stopping_tolerance),
    early_stopping_patience = rlang::enquo(early_stopping_patience),
    skip_importance = rlang::enquo(skip_importance),
    tabnet_model = rlang::enquo(tabnet_model),
    from_epoch = rlang::enquo(from_epoch)
  )

  # Save some empty slots for future parts of the specification
  out <- list(args = args, eng_args = NULL,
              mode = mode, method = NULL, engine = "torch")

  # set classes in the correct order
  class(out) <- parsnip::make_classes("tabnet")
  out
}

tabnet_env <- new.env()
tabnet_env$parsnip_added <- FALSE

#' @export
#' @importFrom parsnip multi_predict
multi_predict._tabnet_fit <- function(object, new_data, type = NULL, epochs = NULL, ...) {

  if (is.null(epochs))
    epochs <- object$fit$config$epochs

  p <- lapply(epochs, function(epoch) {
    pred <- predict(object$fit, new_data, type = type, epoch = epoch)
    nms <- names(pred)
    pred[["epochs"]] <- epoch
    pred[[".row"]] <- seq_len(nrow(new_data))
    pred[, c(".row", "epochs", nms)]
  })

  p <- do.call(rbind, p)
  p <- p[order(p$.row, p$epochs),]
  p <- split(p[,-1], p$.row)
  names(p) <- NULL
  tibble::tibble(.pred = p)
}

#' @export
#' @importFrom stats update
update.tabnet <- function(object, parameters = NULL, epochs = NULL, penalty = NULL, batch_size = NULL,
                          learn_rate = NULL, decision_width = NULL, attention_width = NULL,
                          num_steps = NULL, feature_reusage = NULL, virtual_batch_size = NULL,
                          num_independent = NULL, num_shared = NULL, momentum = NULL, ...) {
  rlang::check_installed("parsnip")
  eng_args <- parsnip::update_engine_parameters(object$eng_args, fresh=TRUE, ...)
  args <- list(
    epochs = rlang::enquo(epochs),
    penalty = rlang::enquo(penalty),
    batch_size = rlang::enquo(batch_size),
    learn_rate = rlang::enquo(learn_rate),
    decision_width = rlang::enquo(decision_width),
    attention_width = rlang::enquo(attention_width),
    num_steps = rlang::enquo(num_steps),
    feature_reusage = rlang::enquo(feature_reusage),
    virtual_batch_size = rlang::enquo(virtual_batch_size),
    num_independent = rlang::enquo(num_independent),
    num_shared = rlang::enquo(num_shared),
    momentum = rlang::enquo(momentum)
  )
  args <- parsnip::update_main_parameters(args, parameters)
  parsnip::new_model_spec(
    "tabnet",
    args = args,
    eng_args = eng_args,
    mode = object$mode,
    method = NULL,
    engine = object$engine
  )
}

#' Determine the minimum set of model fits
#'
#' `min_grid()` determines exactly what models should be fit in order to
#'  evaluate the entire set of tuning parameter combinations. This is for
#'  internal use only and the API may change in the near future.
#'
#' `fit_max_value()` can be used in other packages to implement a `min_grid()`
#' method.
#' @param x A model specification.
#' @param grid A tibble with tuning parameter combinations.
#' @param ... Not currently used.
#' @return A tibble with the minimum tuning parameters to fit and an additional
#' list column with the parameter combinations used for prediction.
#' @keywords internal
#' @examplesIf torch::torch_is_installed()
#' library(dials)
#' library(tune)
#' library(parsnip)
#'
#' tabnet_spec <- tabnet(decision_width = tune(), attention_width = tune()) %>%
#'   set_mode("regression") %>%
#'   set_engine("torch")
#'
#' tabnet_grid <-
#'   tabnet_spec %>%
#'   extract_parameter_set_dials() %>%
#'   grid_regular(levels = 3)
#'
#' min_grid(tabnet_spec, tabnet_grid)
#' 
#' @importFrom tune min_grid
#' @export
min_grid.tabnet <- function(x, grid, ...) tune::fit_max_value(x, grid, ...)

parsnip_is_missing_tabnet <- function(tabnet_env) {
  current <- parsnip::get_model_env()
  !(any(current$models == "tabnet") || tabnet_env$parsnip_added)
}

test_that("parsnip fit model works", {

  # default params
  expect_no_error(
    model <- tabnet() %>%
      parsnip::set_mode("regression") %>%
      parsnip::set_engine("torch")
  )

  expect_no_error(
    fit <- model %>%
      parsnip::fit(Sale_Price ~ ., data = small_ames)
  )

  # some setup params
  expect_no_error(
    model <- tabnet(epochs = 2, learn_rate = 1e-5) %>%
      parsnip::set_mode("regression") %>%
      parsnip::set_engine("torch")
  )

  expect_no_error(
    fit <- model %>%
      parsnip::fit(Sale_Price ~ ., data = small_ames)
  )

  # new batch of setup params
  expect_no_error(
    model <- tabnet(penalty = 0.2, verbose = FALSE, early_stopping_tolerance = 1e-3) %>%
      parsnip::set_mode("classification") %>%
      parsnip::set_engine("torch")
  )

  expect_no_error(
    fit <- model %>%
      parsnip::fit(Overall_Cond ~ ., data = small_ames)
  )

})

test_that("parsnip fit model works from a pretrained model", {

  # default params
  expect_no_error(
    model <- tabnet(tabnet_model = ames_pretrain, from_epoch = 1, epoch = 1) %>%
      parsnip::set_mode("regression") %>%
      parsnip::set_engine("torch")
  )

  expect_no_error(
    fit <- model %>%
      parsnip::fit(Sale_Price ~ ., data = small_ames)
  )




})

test_that("multi_predict works as expected", {

  model <- tabnet(checkpoint_epoch = 1) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  expect_no_error(
    fit <- model %>%
      parsnip::fit(Sale_Price ~ ., data = small_ames)
  )

  preds <- parsnip::multi_predict(fit, small_ames, epochs = c(1,2,3,4,5))

  expect_equal(nrow(preds), nrow(small_ames))
  expect_equal(nrow(preds$.pred[[1]]), 5)
})

test_that("Check we can finalize a workflow", {

  model <- tabnet(penalty = tune(), epochs = tune()) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .)

  wf <- tune::finalize_workflow(wf, tibble::tibble(penalty = 0.01, epochs = 1))

  expect_no_error(
    fit <- wf %>% parsnip::fit(data = small_ames)
  )

  expect_equal(rlang::eval_tidy(wf$fit$actions$model$spec$args$penalty), 0.01)
  expect_equal(rlang::eval_tidy(wf$fit$actions$model$spec$args$epochs), 1)
})

test_that("Check we can finalize a workflow from a tune_grid", {

  model <- tabnet(epochs = tune(), checkpoint_epochs = 1) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_args(epochs = 2) %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .)

  custom_grid <- tidyr::crossing(epochs = c(1,2,3))
  cv_folds <- small_ames %>%
    rsample::vfold_cv(v = 2, repeats = 1)

  at <- tune::tune_grid(
    object = wf,
    resamples = cv_folds,
    grid = custom_grid,
    metrics = yardstick::metric_set(yardstick::rmse),
    control = tune::control_grid(verbose = F)
  )

  best_rmse <- tune::select_best(at, metric = "rmse")

  expect_no_error(
    final_wf <- tune::finalize_workflow(wf, best_rmse)
  )
})

test_that("tabnet grid reduction - torch", {

  mod <- tabnet() %>%
    parsnip::set_engine("torch")

  # A typical grid
  reg_grid <- expand.grid(epochs = 1:3, penalty = 1:2)
  reg_grid_smol <- tune::min_grid(mod, reg_grid)

  expect_equal(reg_grid_smol$epochs, rep(3, 2))
  expect_equal(reg_grid_smol$penalty, 1:2)
  for (i in seq_len(nrow(reg_grid_smol))) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(epochs = 1:2))
  }

  # Unbalanced grid
  reg_ish_grid <- expand.grid(epochs = 1:3, penalty = 1:2)[-3, ]
  reg_ish_grid_smol <- tune::min_grid(mod, reg_ish_grid)

  expect_equal(reg_ish_grid_smol$epochs, 2:3)
  expect_equal(reg_ish_grid_smol$penalty, 1:2)
  for (i in 2:nrow(reg_ish_grid_smol)) {
    expect_equal(reg_ish_grid_smol$.submodels[[i]], list(epochs = 1:2))
  }

  # Grid with a third parameter
  reg_grid_extra <- expand.grid(epochs = 1:3, penalty = 1:2, batch_size = 10:12)
  reg_grid_extra_smol <- tune::min_grid(mod, reg_grid_extra)

  expect_equal(reg_grid_extra_smol$epochs, rep(3, 6))
  expect_equal(reg_grid_extra_smol$penalty, rep(1:2, each = 3))
  expect_equal(reg_grid_extra_smol$batch_size, rep(10:12, 2))
  for (i in seq_len(nrow(reg_grid_extra_smol))) {
    expect_equal(reg_grid_extra_smol$.submodels[[i]], list(epochs = 1:2))
  }

  # Only epochs
  only_epochs <- expand.grid(epochs = 1:3)
  only_epochs_smol <- tune::min_grid(mod, only_epochs)

  expect_equal(only_epochs_smol$epochs, 3)
  expect_equal(only_epochs_smol$.submodels, list(list(epochs = 1:2)))

  # No submodels
  no_sub <- tibble::tibble(epochs = 1, penalty = 1:2)
  no_sub_smol <- tune::min_grid(mod, no_sub)

  expect_equal(no_sub_smol$epochs, rep(1, 2))
  expect_equal(no_sub_smol$penalty, 1:2)
  for (i in seq_len(nrow(no_sub_smol))) {
    expect_length(no_sub_smol$.submodels[[i]], 0)
  }

  # different id names
  mod_1 <- tabnet(epochs = tune("Amos")) %>%
    parsnip::set_engine("torch")
  reg_grid <- expand.grid(Amos = 1:3, penalty = 1:2)
  reg_grid_smol <- tune::min_grid(mod_1, reg_grid)

  expect_equal(reg_grid_smol$Amos, rep(3, 2))
  expect_equal(reg_grid_smol$penalty, 1:2)
  for (i in seq_len(nrow(reg_grid_smol))) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(Amos = 1:2))
  }

  all_sub <- expand.grid(Amos = 1:3)
  all_sub_smol <- tune::min_grid(mod_1, all_sub)

  expect_equal(all_sub_smol$Amos, 3)
  expect_equal(all_sub_smol$.submodels[[1]], list(Amos = 1:2))

  mod_2 <- tabnet(epochs = tune("Ade Tukunbo")) %>%
    parsnip::set_engine("torch")
  reg_grid <- expand.grid(`Ade Tukunbo` = 1:3, penalty = 1:2, ` \t123` = 10:11)
  reg_grid_smol <- tune::min_grid(mod_2, reg_grid)

  expect_equal(reg_grid_smol$`Ade Tukunbo`, rep(3, 4))
  expect_equal(reg_grid_smol$penalty, rep(1:2, each = 2))
  expect_equal(reg_grid_smol$` \t123`, rep(10:11, 2))
  for (i in seq_len(nrow(reg_grid_smol))) {
    expect_equal(reg_grid_smol$.submodels[[i]], list(`Ade Tukunbo` = 1:2))
  }
})

test_that("Check workflow can use case_weight", {

  small_ames_cw <- small_ames %>% dplyr::mutate(case_weight = hardhat::frequency_weights(Year_Sold))
  model <- tabnet(epochs = 2) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .) %>%
    workflows::add_case_weights(case_weight)

  expect_no_error(
    fit <- wf %>% parsnip::fit(data = small_ames_cw)
  )
  expect_no_error(
    predict(fit, small_ames)
  )


})

test_that("Check we can use hardhat:::extract_parameter_set_dials() with {dial} tune()ed parameter", {

  model <- tabnet(batch_size = tune(), learn_rate = tune(), epochs = tune(),
                  momentum = tune(), penalty = tune(), rate_step_size = tune()) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .)

  expect_no_error(
    wf %>% hardhat::extract_parameter_set_dials()
  )
})

test_that("Check we can use hardhat:::extract_parameter_set_dials() with {tabnet} tune()ed parameter", {

  model <- tabnet(num_steps = tune(), num_shared = tune(), mask_type = tune(),
                  feature_reusage = tune(), attention_width = tune()) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .)

  expect_no_error(
    wf %>% hardhat::extract_parameter_set_dials()
  )
})

test_that("Check non supported tune()ed parameter raise an explicit error", {

  model <- tabnet(cat_emb_dim = tune(), checkpoint_epochs = 0) %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")

  wf <- workflows::workflow() %>%
    workflows::add_model(model) %>%
    workflows::add_formula(Sale_Price ~ .)

  expect_error(
    wf %>% hardhat::extract_parameter_set_dials(),
    regexp = "cannot be used as a .* parameter yet"
  )
})

#' Self-supervised learning loss
#'
#' Creates a criterion that measures the Autoassociative self-supervised learning loss between each
#' element in the input \eqn{y_pred} and target \eqn{embedded_x} on the values masked by \eqn{obfuscation_mask}. 
#' 
#' @noRd
#' @importFrom torch torch_mul torch_std torch_matmul torch_mean
nn_unsupervised_loss <- nn_module(
  "nn_unsupervised_loss",
  inherit = torch::nn_cross_entropy_loss,
  
  initialize = function(eps = 1e-9){
    super$initialize()
    self$eps = eps
  },
  
  forward = function(y_pred, embedded_x, obfuscation_mask){
    errors <- y_pred - embedded_x
    reconstruction_errors <- torch_mul(errors, obfuscation_mask) ^ 2
    batch_stds <- torch_std(embedded_x, dim = 1) ^ 2 + self$eps
    
    # compute the number of obfuscated variables to reconstruct
    nb_reconstructed_variables <- torch_sum(obfuscation_mask, dim = 2)
    
    # take the mean of the reconstructed variable errors
    features_loss <- torch_matmul(reconstruction_errors, 1 / batch_stds) / (nb_reconstructed_variables +  self$eps)
    loss <- torch_mean(features_loss, dim = 1)
    loss
  }
)


#' AUM loss
#'
#' Creates a criterion that measures the Area under the \eqn{Min(FPR, FNR)} (AUM) between each
#' element in the input \eqn{pred_tensor} and target \eqn{label_tensor}. 
#' 
#' This is used for measuring the error of a binary reconstruction within highly unbalanced dataset, 
#' where the goal is optimizing the ROC curve. Note that the targets \eqn{label_tensor} should be factor
#' level of the binary outcome, i.e. with values `1L` and `2L`.
#'
#' @examplesIf torch::torch_is_installed()
#' loss <- nn_aum_loss()
#' input <- torch::torch_randn(4, 6, requires_grad = TRUE)
#' target <- input > 1.5
#' output <- loss(input, target)
#' output$backward()
#' @export
#' @importFrom torch nn_module torch_sum torch_cat torch_minimum torch_long torch_argsort
#' @importFrom torch torch_tensor as_array
nn_aum_loss <- nn_module(
  "nn_aum_loss",
  inherit = torch::nn_mse_loss,
  initialize = function(){
    super$initialize()
    self$roc_aum <- tibble::tibble()
  },
  forward = function(pred_tensor, label_tensor){
    # thanks to https://tdhock.github.io/blog/2024/auto-grad-overhead/
    is_positive <- label_tensor == label_tensor$max()
    is_negative <- is_positive$bitwise_not()
    # manage case when prediction error is null (prevent division by 0)
    if(as.logical(torch_sum(is_positive) == 0) || as.logical(torch_sum(is_negative) == 0)){
      return(torch_sum(pred_tensor*0))
    }
    
    # pred tensor may be [prediction, case_wts] when add_case_weight() is used. We keep only prediction
    if (pred_tensor$ndim > label_tensor$ndim) {
      pred_tensor <- pred_tensor$slice(dim = 2, 0, 1)$squeeze(2) 
    }
    
    # nominal case
    fn_diff <- -1L * is_positive
    fp_diff <- is_negative$to(dtype = torch_long())
    fp_denom <- torch_sum(is_negative) # or 1 for AUM based on count instead of rate
    fn_denom <- torch_sum(is_positive) # or 1 for AUM based on count instead of rate
    sorted_pred_ids <- torch_argsort(pred_tensor, dim = 1, descending = TRUE)$squeeze(-1)
    
    sorted_fp_cum <- fp_diff[sorted_pred_ids]$cumsum(dim = 1) / fp_denom
    sorted_fn_cum <- -fn_diff[sorted_pred_ids]$flip(1)$cumsum(dim = 1)$flip(1) / fn_denom
    sorted_thresh_gr <- -pred_tensor[sorted_pred_ids]
    sorted_dedup <- sorted_thresh_gr$diff(dim = 1) != 0
    # pad to replace removed last element
    padding <- sorted_dedup$slice(dim = 1, 0, 1) # torch_tensor 1 w same dtype, same shape, same device 
    sorted_fp_end <- torch_cat(c(sorted_dedup, padding))
    sorted_fn_end <- torch_cat(c(padding, sorted_dedup))
    uniq_thresh_gr <- sorted_thresh_gr[sorted_fp_end]
    uniq_fp_after <- sorted_fp_cum[sorted_fp_end]
    uniq_fn_before <- sorted_fn_cum[sorted_fn_end]
    if (pred_tensor$ndim == 1) {
      FPR <- torch_cat(c(padding$logical_not(), uniq_fp_after)) # FPR with trailing 0
      FNR <-  torch_cat(c(uniq_fn_before, padding$logical_not())) # FNR with leading 0
      self$roc_aum <- list(
        FPR = FPR,
        FNR = FNR,
        TPR = 1 - FNR,
        "min(FPR,FNR)" = torch_minimum(FNR, FPR), # full-range min(FNR, FPR)
        constant_range_low = torch_cat(c(torch_tensor(-Inf), uniq_thresh_gr)),
        constant_range_high = torch_cat(c(uniq_thresh_gr, torch_tensor(Inf)))
      ) %>% purrr::map_dfc(as_array)
    }
    min_FPR_FNR <- torch_minimum(uniq_fp_after[1:-2], uniq_fn_before[2:N])
    constant_range_gr <- uniq_thresh_gr$diff() # range splits leading to {FPR, FNR } errors (see roc_aum row)
    torch_sum(min_FPR_FNR * constant_range_gr, dim = 1)
    
  }
)


#' Apply hierarchy constraints via max-pooling over descendants (MCM)
#'
#' Given neural network outputs x and ancestor matrix R, enforces that
#' if a class is predicted positive, all its ancestors must also be positive.
#' Implements: `final_out[i] = max{x[j] : R[i,j] = 1}`
#'
#' @param x A `torch_tensor` of shape `(batch_size, n_classes)`.
#' @param R A `torch_tensor` of shape `(1, n_classes, n_classes)` where 
#'   `R[1, i, j] = 1` iff class `i` is a descendant of class `j`.
#' @return A `torch_tensor` of shape `(batch_size, n_classes)` with constrained outputs.
get_constr_output <- function(x, R) {
  c_out <- x$unsqueeze(2)$expand(c(x$shape[1], R$shape[2], R$shape[2]))
  R_batch <- R$expand(c(x$shape[1], R$shape[2], R$shape[2]))
  final_out <- (R_batch * c_out)$clone()$max(dim = 3)
  final_out[[1]]
}


#' Max-Constraint Margin Loss (functional)
#'
#' Computes the hierarchy-constrained loss for multi-label classification.
#' Enforces that if a class is predicted positive, all its ancestors must 
#' also be positive, using the ancestor matrix R.
#'
#' The loss combines constrained outputs differently for positive and negative 
#' labels:
#' \itemize{
#'   \item For positive labels: uses constrained output of label-weighted predictions
#'   \item For negative labels: uses constrained raw predictions (penalizes ancestor violations)
#' }
#'
#' @param output A `torch_tensor` of raw network outputs (pre-sigmoid), 
#'   shape `(batch_size, n_classes)`.
#' @param target Binary target labels, shape `(batch_size, n_classes)`.
#' @param R Ancestor matrix tensor of shape `(1, n_classes, n_classes)` where 
#'   `R[1, i, j] = 1` iff class `i` is a descendant of class `j`.
#' @param to_eval Optional logical tensor of shape `(n_classes,)` indicating 
#'   which classes to include in the loss computation. If `NULL`, all classes 
#'   are evaluated.
#' @param criterion Loss function to apply after constraint propagation. 
#'   Default: `nnf_binary_cross_entropy_with_logits` (expects raw logits).
#'
#' @return A scalar `torch_tensor` containing the computed loss, or a tensor 
#'   of shape `(batch_size, n_classes)` if `reduction = "none"`.
#'
#' @seealso [nn_mc_loss()], [get_constr_output()]
#' @importFrom stats setNames
#' @export
nnf_mc_loss <- function(output, target, R, to_eval = NULL, 
                        criterion = nnf_binary_cross_entropy_with_logits) {
  # Ensuring double precision for numerical stability during constraint propagation is not available on mps platform
  # output_d <- output$double()
  
  # 1. Constrained output from raw predictions: max-pool over descendants
  constr_output <- get_constr_output(output, R)  # (batch, n_classes)
  
  # 2. Label-weighted output, then constrained (for positive label handling)
  labeled_output <- target * output
  train_output <- get_constr_output(labeled_output, R)
  
  # 3. Blend outputs based on ground-truth labels:
  #    - Positive labels: use constrained label-weighted output
  #    - Negative labels: use constrained raw output
  blended_output <- (1 - target) * constr_output + target * train_output
  
  # 4. Select classes to evaluate (if specified)
  if (!is.null(to_eval)) {
    blended_output <- blended_output[, to_eval, drop = FALSE]
    target <- target[, to_eval, drop = FALSE]
  }
  
  # 5. Apply the base loss function (e.g., BCE with logits)
  loss <- criterion(
    blended_output, 
    target
  )
  
  return(loss)
}


#' Max-Constraint Margin Loss (module)
#'
#' Module wrapper for [nnf_mc_loss()] with configurable parameters.
#' Stores the ancestor matrix R and evaluation mask for reuse across batches.
#'
#' @param R Ancestor matrix tensor of shape `(1, n_classes, n_classes)`.
#' @param to_eval Optional logical tensor of shape `(n_classes,)` indicating 
#'   which classes to include in loss computation.
#' @param criterion Loss function module or functional to apply after constraint 
#'   propagation. Default: `nn_binary_cross_entropy_with_logits()`.
#' @param reduction (string, optional): Reduction method: `'none'` | `'mean'` | `'sum'`.
#'
#' @section Shape:
#' - Input `output`: \eqn{(N, C)} where N = batch size, C = number of classes
#' - Input `target`: \eqn{(N, C)}, same shape as output, binary values
#' - Output: scalar by default. If `reduction = "none"`, then \eqn{(N, C')} 
#'   where C' is the number of evaluated classes
#'
#' @examples
#' \dontrun{
#' # Build ancestor matrix from hierarchy
#' R <- build_ancestor_matrix_from_outcomes(my_tree, processed$outcomes, device = "cuda")
#' 
#' # Create loss module
#' loss_fn <- nn_mc_loss(R = R, reduction = "mean")
#' 
#' # Forward pass
#' output <- model(x)  # (batch, n_classes)
#' loss <- loss_fn(output, labels)
#' loss$backward()
#' }
#'
#' @seealso [nnf_mc_loss()], [build_ancestor_matrix_from_outcomes()], [get_constr_output()]
#' @importFrom torch nnf_binary_cross_entropy_with_logits
#' @export
nn_mc_loss <- nn_module(
  "nn_mc_loss",
  inherit = torch::nn_l1_loss,
  
  initialize = function(R, to_eval = NULL, 
                        criterion = nnf_binary_cross_entropy_with_logits,
                        reduction = "mean") {
    super$initialize(reduction = reduction)
    
    # Store ancestor matrix (move to device if needed)
    self$R <- R
    self$to_eval <- to_eval
    # Resolve criterion based on its type
    self$criterion_fn <- .resolve_mc_criterion(criterion, reduction)
  },
  
  forward = function(output, target) {
    nnf_mc_loss(
      output = output,
      target = target,
      R = self$R,
      to_eval = self$to_eval,
      criterion = self$criterion_fn
    )
  }
)

#' Resolve criterion into a callable function(input, target, reduction)
#' @keywords internal
#' @noRd
.resolve_mc_criterion <- function(criterion, reduction) {
  # Case 1: Already an nn_module instance
  if (inherits(criterion, "nn_module")) {
    module_reduction <- criterion$reduction
    if (!is.null(module_reduction) && module_reduction != reduction) {
      warn(
        c(
          "The criterion module has reduction={.val {module_reduction}}",
          "but nn_mc_loss was called with reduction={.val {reduction}}.",
          "i" = "The module's reduction will be used."
        ),
        class = "mc_loss_reduction_mismatch"
      )
    }
    return(function(input, target) criterion(input, target))
  }
  
  # Case 2: A function (could be functional nnf_* or constructor nn_*)
  if (rlang::is_function(criterion)) {
    # Try to detect if it's a constructor by calling with just reduction
    # Constructors return nn_module, functionals need input/target
    maybe_module <- tryCatch(
      {
        result <- criterion(reduction = reduction)
        if (inherits(result, "nn_module")) result else NULL
      },
      error = function(e) NULL
    )
    
    if (!is.null(maybe_module)) {
      # It's a constructor (e.g., nn_bce_with_logits_loss)
      return(function(input, target) maybe_module(input, target))
    }
    
    # It's a functional (e.g., nnf_binary_cross_entropy_with_logits)
    return(function(input, target) {
      criterion(input, target, reduction = reduction)
    })
  }
  
  # Invalid type
  value_error(
    c(
      "`criterion` must be a function or an `nn_module`.",
      "x" = "Got: {.class {class(criterion)[1]}}"
    ),
    class = "mc_loss_invalid_criterion"
  )
}

#' Convert class_id tensor to binary one-hot tensor
#'
#' Transforms a tensor of class indices (one column per hierarchy level)
#' into a binary tensor where each column corresponds to a class.
#'
#' @param y A `torch_tensor` of shape `(batch_size, n_levels)` containing
#'   1-based class indices.
#' @param outcomes A tibble with factor columns (as from `hardhat::mold()$outcomes`).
#' @param device Torch device.
#' @return A `torch_tensor` of shape `(batch_size, n_classes)` with binary values.
#' @export
nnf_multilabel_one_hot <- function(y, outcomes, device = "cpu") {
  batch_size <- y$shape[1]
  n_levels <- y$shape[2]
  
  # Number of classes per level
  n_per_level <- lengths(lapply(outcomes, levels))
  n_classes <- sum(n_per_level)
  
  one_hot_list <- vector("list", n_levels)
  
  for (lvl in seq_len(n_levels)) {
    level_ids <- y[, lvl]$to(dtype = torch::torch_long())
    
    # Encode one-hot of each levels
    one_hot_list[[lvl]] <- torch::nnf_one_hot(
      level_ids, 
      num_classes = n_per_level[lvl]
    )
  }
  # concatenate along the columns axis; cast to float for BCE loss compatibility
  torch::torch_cat(one_hot_list, dim = 2)$to(device = device, dtype = torch::torch_float())
}

test_that("nn_unsupervised_loss is working as expected", {
  
  unsup_loss <- tabnet:::nn_unsupervised_loss()
  
  # the poor-guy expect_r6_class(x, class)
  expect_true(all(c("nn_weighted_loss","nn_loss","nn_module") %in% class(unsup_loss)))
  
  y_pred <- torch::torch_rand(3,5, requires_grad = TRUE)
  embedded_x <- torch::torch_rand(3,5)
  obfuscation_mask <- torch::torch_bernoulli(embedded_x, p = 0.5)
  output <- unsup_loss(y_pred, embedded_x, obfuscation_mask)
  output$backward()
  
  expect_tensor(output)
  expect_equal_to_r(output >= 0, TRUE) 
  expect_false(rlang::is_null(output$grad_fn))
  expect_equal(output$dim(), 0)
})


test_that("nn_aum_loss works as expected with 1-dim label", {
  
  aum_loss <- tabnet::nn_aum_loss()
  
  # the poor-guy expect_r6_class(x, class)
  expect_true(all(c("nn_mse_loss","nn_loss","nn_module") %in% class(aum_loss)))
  
  # 1-dim label
  label_tensor <- torch::torch_tensor(attrition$Attrition)
  pred_tensor <- torch::torch_rand(label_tensor$shape, requires_grad = TRUE) 
  output <- aum_loss(pred_tensor, label_tensor)
  output$backward()
  
  expect_tensor(output)
  expect_equal_to_r(output >= 0, TRUE) 
  expect_false(rlang::is_null(output$grad_fn))
  expect_equal(output$dim(), 0)
  
})


test_that("nn_aum_loss works as expected with 2-dim label", {
  
  aum_loss <- tabnet::nn_aum_loss()
  label_tensor <- torch::torch_tensor(attrition$Attrition)$unsqueeze(-1)
  pred_tensor <- torch::torch_rand(label_tensor$shape, requires_grad = TRUE)
  output <- aum_loss(pred_tensor, label_tensor)
  output$backward()
  
  expect_tensor(output)
  expect_equal_to_r(output >= 0, TRUE) 
  expect_false(rlang::is_null(output$grad_fn))
  expect_equal(output$dim(), 0)
})


test_that("nn_aum_loss works as expected with {n, 2} shape prediction", {
  
  aum_loss <- tabnet::nn_aum_loss()
  label_tensor <- torch::torch_tensor(attrition$Attrition)
  pred_tensor <- torch::torch_rand(c(label_tensor$shape, 2), requires_grad = TRUE) 
  output <- aum_loss(pred_tensor, label_tensor)
  output$backward()
  
  expect_tensor(output)
  expect_equal_to_r(output >= 0, TRUE) 
  expect_false(rlang::is_null(output$grad_fn))
  expect_equal(output$dim(), 0)
})


test_that("get_constr_output handles basic 2D input with identity constraint", {
  m <- matrix(c(1, 2, 
                3, 4), nrow = 2, ncol = 2)
  x <- torch_tensor(m, dtype = torch::torch_float32())
  R <- torch::torch_eye(2)
  result <- get_constr_output(x, R)
  expect_tensor(result)
  expect_tensor_shape(result, c(2, 2))
  expect_equal_to_r(result, m)
})

test_that("get_constr_output applies hierarchy constraint correctly", {
  x <- torch_tensor(matrix(c(1, 5, 
                             3, 2), nrow = 2, ncol = 2, byrow = TRUE))
  R <- torch_tensor(matrix(c(1, 1, 
                             0, 1), nrow = 2, ncol = 2, byrow = TRUE))
  result <- get_constr_output(x, R)
  expect_tensor_shape(result, c(2, 2))
  expected <- matrix(c(5, 5, 3, 2), nrow = 2, ncol = 2, byrow = TRUE)
  expect_equal_to_r(result, expected, tolerance = 1e-6)
})

test_that("get_constr_output handles batch dimension correctly", {
  x <- torch_tensor(matrix(1:12, nrow = 3, ncol = 4))
  R <- torch_tensor(matrix(c(1, 1, 0, 0, 
                             1, 1, 0, 0, 
                             0, 0, 1, 1, 
                             0, 0, 1, 1), nrow = 4, ncol = 4, byrow = TRUE))
  result <-get_constr_output(x, R)
  expect_tensor_shape(result, c(3, 4))
  
  for (i in 1:3) {
    row_result <- as_array(result[i, ])
    max_grp1 <- max(as_array(x[i, 1:2]))
    max_grp2 <- max(as_array(x[i, 3:4]))
    
    expect_equal(row_result[1:2], rep(max_grp1, 2), tolerance = 1e-6)
    expect_equal(row_result[3:4], rep(max_grp2, 2), tolerance = 1e-6)
  }
})
test_that("get_constr_output works with single sample", {
  x <- torch_tensor(matrix(c(2, 1, 4, 3), nrow = 1, ncol = 4, byrow = TRUE))
  R <- torch_tensor(matrix(c(1, 1, 0, 0, 
                             1, 1, 0, 0, 
                             0, 0, 1, 1, 
                             0, 0, 1, 1), nrow = 4, ncol = 4, byrow = TRUE))
  result <-get_constr_output(x, R)
  expect_tensor_shape(result, c(1, 4))
  expected <- matrix(c(2, 2, 4, 4), nrow = 1, byrow = TRUE)
  expect_equal_to_r(result, expected)
})

test_that("get_constr_output handles all-zeros constraint matrix", {
  x <- torch_tensor(matrix(1:6, nrow = 2, ncol = 3))
  R <- torch::torch_zeros(c(3, 3))
  result <-get_constr_output(x, R)
  expect_tensor_shape(result, c(2, 3))
  expect_equal_to_r(result, matrix(0, nrow = 2, ncol = 3))
})

test_that("get_constr_output handles all-ones constraint matrix", {
  x <- torch_tensor(matrix(c(1, 5, 3, 
                             2, 4, 6), nrow = 2, ncol = 3, byrow = TRUE))
  R <- torch::torch_ones(c(3, 3))
  result <-get_constr_output(x, R)
  expect_tensor_shape(result, c(2, 3))
  # Each row is filled with its own row-wise maximum
  expected <- matrix(c(5, 5, 5, 
                       6, 6, 6), nrow = 2, ncol = 3, byrow = TRUE)
  expect_equal_to_r(result, expected, tolerance = 1e-6)
})

test_that("get_constr_output throws error for dimension mismatch", {
  x <- torch_tensor(matrix(1:4, nrow = 2, ncol = 2))
  R <- torch::torch_eye(3)
  expect_error(get_constr_output(x, R), "must match the existing size")
})

test_that("get_constr_output throws error for non-2D R", {
  x <- torch_tensor(matrix(1:4, nrow = 2, ncol = 2))
  R <- torch_tensor(array(1:8, dim = c(1, 2, 2, 2)))
  expect_error(get_constr_output(x, R), "dimension")
})

test_that("get_constr_output handles negative values correctly", {
  x <- torch_tensor(matrix(c(-5, -1, 
                             -3, -2), nrow = 2, ncol = 2, byrow = TRUE))
  R <- torch_tensor(matrix(c(1, 1, 
                             0, 1), nrow = 2, ncol = 2, byrow = TRUE))
  result <- get_constr_output(x, R)
  expected <- matrix(c(-1, 0, 
                       -2, 0), nrow = 2, ncol = 2, byrow = TRUE)
  expect_equal_to_r(result, expected)
})

test_that("nn_mc_loss resolves functional criterion at initialization", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  # Functional criterion
  loss_fn <- nn_mc_loss(
    R = R,
    criterion = torch::nnf_binary_cross_entropy_with_logits,
    reduction = "mean"
  )
  
  expect_true(rlang::is_function(loss_fn$criterion_fn))
  
  output <- torch::torch_randn(2, 3, requires_grad = TRUE)
  target <- torch::torch_randint(0, 2, c(2, 3))
  
  expect_no_error(loss <- loss_fn(output, target))
  expect_tensor(loss)
})

test_that("nn_mc_loss resolves nn_module criterion at initialization (default)", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  # Functional criterion
  loss_fn <- nn_mc_loss(R = R)
  
  expect_true(rlang::is_function(loss_fn$criterion_fn))
  
  output <- torch::torch_randn(2, 3, requires_grad = TRUE)
  target <- torch::torch_randint(0, 2, c(2, 3))
  
  expect_no_error(loss <- loss_fn(output, target))
  expect_tensor(loss)
})

test_that("nn_mc_loss can use already instanciated nn_module criterion", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  # Module criterion
  loss_fn <- nn_mc_loss(
    R = R,
    criterion = torch::nn_bce_with_logits_loss(),
    reduction = "mean"
  )
  
  expect_true(rlang::is_function(loss_fn$criterion_fn))
  
  output <- torch::torch_randn(2, 3, requires_grad = TRUE)
  target <- torch::torch_randint(0, 2, c(2, 3))
  
  expect_no_error(loss <- loss_fn(output, target))
  expect_tensor(loss)
})

test_that("nn_mc_loss errors on invalid criterion type", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  expect_error(
    nn_mc_loss(R = R, criterion = "not_a_valid_criterion"),
    "must be a function or an"
  )
})

test_that("nn_mc_loss warns on reduction mismatch for module criterion", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  # Module with 'sum' reduction, but loss asks for 'mean'
  expect_warning(
    nn_mc_loss(
      R = R,
      criterion = torch::nn_bce_with_logits_loss(reduction = "sum"),
      reduction = "mean"
    ),
    "The criterion module has reduction"
  )
})

test_that("nn_mc_loss backward pass works without inplace errors", {
  R <- torch::torch_eye(3)$unsqueeze(1)
  
  loss_fn <- nn_mc_loss(R = R, reduction = "mean")
  
  output <- torch::torch_randn(2, 3, requires_grad = TRUE)
  target <- torch::torch_randint(0, 2, c(2, 3))
  
  # Forward
  loss <- loss_fn(output, target)
  
  # Backward should not throw inplace error
  expect_no_error(loss$backward())
  
  # Gradients should be computed
  expect_true(!is.null(output$grad))
  expect_tensor_shape(output$grad, output$shape)
})

check_dials <- function() {
  if (!requireNamespace("dials", quietly = TRUE))
    runtime_error("Package {.pkg dials} is needed for this function to work. Please install it.")
}


#' Parameters for the tabnet model
#'
#' @param range the default range for the parameter value
#' @param trans whether to apply a transformation to the parameter
#' @param values possible values for factor parameters
#'
#' These functions are used with `tune` grid functions to generate
#' candidates.
#'
#' @rdname tabnet_params
#' @return A `dials` parameter to be used when tuning TabNet models.
#' @export
#' @examplesIf (require("dials") && require("parsnip") && torch::torch_is_installed())
#'   model <- tabnet(attention_width = tune(), feature_reusage = tune(),
#'     momentum = tune(), penalty = tune(), rate_step_size = tune()) %>%
#'     parsnip::set_mode("regression") %>%
#'     parsnip::set_engine("torch")
#'
attention_width <- function(range = c(8L, 64L), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "integer",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(attention_width = "Width of the attention embedding for each mask"),
    finalize = NULL
  )
}

#' @rdname tabnet_params
#' @export
decision_width <- function(range = c(8L, 64L), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "integer",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(decision_width = "Width of the decision prediction layer"),
    finalize = NULL
  )
}


#' @rdname tabnet_params
#' @export
feature_reusage <- function(range = c(1, 2), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "double",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(feature_reusage = "Coefficient for feature reusage in the masks"),
    finalize = NULL
  )
}

#' @rdname tabnet_params
#' @export
momentum <- function(range = c(0.01, 0.4), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "double",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(momentum = "Momentum for batch normalization"),
    finalize = NULL
  )
}


#' @rdname tabnet_params
#' @export
mask_type <- function(values = c("sparsemax", "entmax")) {
  check_dials()
  dials::new_qual_param(
    type = "character",
    values = values,
    label = c(mask_type = "Final layer of feature selector, either 'sparsemax' or 'entmax'"),
    finalize = NULL
  )
}

#' @rdname tabnet_params
#' @export
num_independent <- function(range = c(1L, 5L), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "integer",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(num_independent = "Number of independent Gated Linear Units layers at each step"),
    finalize = NULL
  )
}

#' @rdname tabnet_params
#' @export
num_shared <- function(range = c(1L, 5L), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "integer",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(num_shared = "Number of shared Gated Linear Units at each step"),
    finalize = NULL
  )
}

#' @rdname tabnet_params
#' @export
num_steps <- function(range = c(3L, 10L), trans = NULL) {
  check_dials()
  dials::new_quant_param(
    type = "integer",
    range = range,
    inclusive = c(TRUE, TRUE),
    trans = trans,
    label = c(num_steps = "Number of steps in the architecture"),
    finalize = NULL
  )
}

#' Non-tunable parameters for the tabnet model
#'
#' @param range unused
#' @param trans unused
#' @rdname tabnet_non_tunable
#' @export
cat_emb_dim <- function(range = NULL, trans = NULL) {
  cli::cli_abort("{.var cat_emb_dim} cannot be used as a {.fun tune} parameter yet.")
}

#' @rdname tabnet_non_tunable
#' @export
checkpoint_epochs <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
drop_last <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
encoder_activation <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
lr_scheduler <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
mlp_activation <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
mlp_hidden_multiplier <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
num_independent_decoder <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
num_shared_decoder <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
optimizer <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
penalty <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
verbose <- cat_emb_dim

#' @rdname tabnet_non_tunable
#' @export
virtual_batch_size <- cat_emb_dim

#' Pipe operator
#'
#' See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
#'
#' @name %>%
#' @rdname pipe
#' @keywords internal
#' @export
#' @importFrom magrittr %>%
#' @importFrom zeallot %<-%
#' @usage lhs \%>\% rhs
#'
#' @return Returns `rhs(lhs)`.
NULL

#' @importFrom generics augment
#' @export
generics::augment

# credits to Yandex https://github.com/Qwicen/node/blob/master/lib/nn_utils.py
.make_ix_like <- function(input, dim = 1L) {
  d <- input$size(dim)
  rho <- torch::torch_arange(start = 1, end = d, device = input$device,
                             dtype = input$dtype)
  view <- rep(1, input$dim())
  view[1] <- -1
  rho$view(view)$transpose(1, dim)
}

.roll_last <- function(input, dim = 1L) {
  if (dim == -1L) {
    return(input)
  } else if (dim < 0) {
    dim <- input$dim() - dim
  }
  perm <- c(which(1:input$dim() != dim), dim)
  return(input$permute(perm))
}


#' Sparsemax compute with optimal threshold and support size.
#'
#' @param input The input tensor to compute thresholds over.
#' @param dim The dimension along which to apply sparsemax.
#' @param k The n largest elements to partial-sort input over. For optimal
#' performance, should be slightly bigger than the expected number of
#' non-zeros in the solution. If the solution is more than k-sparse,
#' this function is recursively called with a 2*k schedule. If `NULL`, full
#' sorting is performed from the beginning.
#'
#' @return A list containing:
#' \itemize{
#' \item{\code{tau}}{The threshold value for each vector, with all but the \code{dim} dimension intact.}
#' \item{\code{support_size}}{The number of non-zeros in each vector.}
#' }
#' @noRd
.sparsemax_threshold_and_support <- function(input, dim = -1L, k = NULL) {
  tau_ <- ss_ <- NULL # avoid NOTE
  if (is.null(k) || k >= input$size(dim)) { # do full sort
    topk <- input$sort(dim = dim, descending = TRUE)[[1]]
  } else {
    topk <- input$topk(k = k, dim = dim)[[1]]
  }
  
  topk_cumsum <- topk$cumsum(dim) - 1
  rhos <- .make_ix_like(topk, dim)
  support <- rhos * topk > topk_cumsum
  
  support_size <- support$sum(dim)$unsqueeze(dim)
  tau <- topk_cumsum$gather(dim, support_size)
  tau <- tau / support_size$to(dtype = input$dtype)
  
  if (!is.null(k) && k < input$size(dim)) {
    unsolved <- (support_size == k)$squeeze(dim)
    
    if (as.numeric(unsolved$sum()) > 0) {
      input_ <- .roll_last(input, dim)[unsolved]
      c(tau_, ss_) %<-% .sparsemax_threshold_and_support(input_, dim = -1L, k = 2 * k)
      tau_rl <- .roll_last(tau, dim)
      tau_rl[unsolved] <- tau_
      tau <- .roll_last(tau_rl, dim)
      support_size_rl <- .roll_last(support_size, dim)
      support_size_rl[unsolved] <- ss_
      support_size <- .roll_last(support_size_rl, dim)
      
    }
  }
  
  list(tau, support_size)
}


sparsemax_function <- torch::autograd_function(
  forward = function(ctx, input, dim = -1L, k = NULL) {
    max_val <- input$max(dim = dim, keepdim = TRUE)[[1]]
    input <- input - max_val # same numerical stability trick as for softmax
    c(tau, supp_size) %<-% .sparsemax_threshold_and_support(input, dim = dim, k = k)
    output <- torch::torch_clamp(input - tau, min = 0)
    ctx$save_for_backward(supp_size = supp_size, output = output, dim = dim, k = k)
    output
  },

  backward = function(ctx, grad_output) {
    c(supp_size, output, dim, k) %<-% ctx$saved_variables
    grad_input <- grad_output$clone()
    grad_input[output == 0] <- 0

    v_hat <- grad_input$sum(dim = dim) / supp_size$to(dtype = output$dtype)$squeeze(dim)
    v_hat <- v_hat$unsqueeze(dim)
    grad_input <- torch::torch_where(output != 0, grad_input - v_hat, grad_input)

    list(
      input = grad_input,
      dim = dim,
      k = k
    )
  }
)

#' Sparsemax 
#' 
#' Normalizing sparse transform (a la softmax).
#'
#' Solves the projection:
#' 
#' \eqn{\min_P ||input - P||_2 \text{ s.t. } P \geq0, \sum(P) ==1}
#'
#'
#' @param dim The dimension along which to apply sparsemax.
#' @param k The number of largest elements to partial-sort input over. For optimal
#' performance, `k` should be slightly bigger than the expected number of
#' non-zeros in the solution. If the solution is more than k-sparse,
#' this function is recursively called with a 2*k schedule. If `NULL`, full
#' sorting is performed from the beginning.
#'
#' @return The projection result, such that \eqn{\sum_{dim} P = 1 \forall dim} elementwise.
#'
#' @examplesIf torch::torch_is_installed()
#' \dontrun{
#' input <- torch::torch_randn(10, 5, requires_grad = TRUE)
#' # create a top3 alpha=1.5 sparsemax on last input dimension
#' nn_sparsemax <- sparsemax15(dim=1, k=3)
#' result <- nn_sparsemax(input)
#' print(result)
#' }
#' @export
sparsemax <- torch::nn_module(
  "sparsemax",
  initialize = function(dim = -1L) {
    self$dim <- dim
  },
  forward = function(input) {
    sparsemax_function(input, self$dim)
  }
)

#' Alpha-Sparsemax with alpha equal 1.5 
#' @rdname sparsemax
#' @export
sparsemax15 <- torch::nn_module(
  "sparsemax_15",
  initialize = function(dim = -1L, k=NULL) {
    self$dim <- dim
    self$k <- k
  },
  forward = function(input) {
    sparsemax_function(input, self$dim, self$k)
  }
)

#' 1.5-entmax compute with optimal threshold and support size.
#'
#' @param input The input tensor to compute thresholds over.
#' @param dim The dimension along which to apply 1.5-entmax.
#' @param k The n largest elements to partial-sort input over. For optimal
#'   performance, `k` should be slightly bigger than the expected number of
#'   non-zeros in the solution. If the solution is more than k-sparse,
#'   this function is recursively called with a 2*k schedule. If `NULL`, full
#'   sorting is performed from the beginning.
#'
#' @return A list containing:
#'   \itemize{
#'     \item{\code{tau_star}}{The threshold value for each vector, with all but the \code{dim} dimension intact.}
#'     \item{\code{support_size}}{The number of non-zeros in each vector.}
#'   }
#' @noRd
.entmax_threshold_and_support <- function(input, dim = -1L, k = NULL) {
  tau_ <- ss_ <- NULL # avoid NOTE
  if (is.null(k) || k >= input$size(dim)) { # do full sort
    sorted_input <- input$sort(dim = dim, descending = TRUE)[[1]]
  } else {
    sorted_input <- input$topk(k = k, dim = dim)[[1]]
  }
  
  rho <- .make_ix_like(sorted_input, dim)
  mean <- sorted_input$cumsum(dim) / rho
  mean_sq <- sorted_input$pow(2)$cumsum(dim) / rho
  ss <- rho * (mean_sq - mean$pow(2))
  delta <- (1 - ss) / rho
  
  # NOTE this is not exactly the same as in reference algo
  # Fortunately it seems the clamped values never wrongly
  # get selected by tau <= sorted_z. Prove this!
  delta_nz <- torch::torch_clamp(delta,0)
  tau <- mean - torch::torch_sqrt(delta_nz)
  
  support_size <- (tau <= sorted_input)$sum(dim)$unsqueeze(dim)
  tau_star <- tau$gather(dim, support_size)
  
  if (!is.null(k) && k < input$size(dim)) {
    unsolved <- (support_size == k)$squeeze(dim)
    
    if (as.numeric(unsolved$sum()) > 0) {
      input_ <- .roll_last(input, dim)[unsolved]
      c(tau_, ss_ ) %<-% .entmax_threshold_and_support(input_, dim = -1L, k = 2 * k)
      tau_star_rl <- .roll_last(tau_star, dim)
      tau_star_rl[unsolved] <- tau_
      tau_star <- .roll_last(tau_star_rl, dim)
      support_size_rl <- .roll_last(support_size, dim)
      support_size_rl[unsolved] <- ss_
      support_size <- .roll_last(support_size_rl, dim)
    }
  }
  
  list( tau_star, support_size)
}


entmax_function <- torch::autograd_function(
  forward = function(ctx, input, dim = -1L) {
    max_val <- input$max(dim=dim, keepdim=TRUE)[[1]]
    input$sub_(max_val) # same numerical stability trick as for softmax
    input <- input / 2

    tau_supp <- .entmax_threshold_and_support(input, dim = dim)
    output <- torch::torch_clamp(input - tau_supp[[1]], min = 0) ^ 2
    ctx$save_for_backward(supp_size = tau_supp[[2]], output = output, dim = dim)
    output
  },

  backward = function(ctx, grad_output) {
    # supp_size, output = ctx$saved_variables
    saved <- ctx$saved_variables
    dim <- saved$dim
    output <- saved$output
    gppr <- output$sqrt()
    dX <- grad_output * gppr
    q <- dX$sum(dim) / gppr$sum(dim)
    q <- q$unsqueeze(dim)
    dX$sub_(q*gppr)

    list(
      input = dX,
      dim = dim
    )
  }
)

#' @export
#' @rdname entmax15
entmax <- torch::nn_module(
  "entmax",
  initialize = function(dim = -1) {
    self$dim <- dim
  },
  forward = function(input) {
    if (self$dim == -1) {
      self$dim <- input$dim()
    }
    entmax_function(input, self$dim)
  }
)

entmax_15_function <- torch::autograd_function(
  forward = function(ctx, input, dim = 1L, k = NULL) {
    max_val <- input$max(dim = dim, keepdim = TRUE)[[1]]
    input <- (input - max_val)/2 # same numerical stability trick as for softmax

    tau_star <- .entmax_threshold_and_support(input, dim = dim, k = k)[[1]]
    output <- torch::torch_clamp(input - tau_star, min = 0) ^ 2
    ctx$save_for_backward(output = output, dim = dim, k = k)
    output
  },

  backward = function(ctx, grad_output) {
    # supp_size, output = ctx$saved_variables
    s <- ctx$saved_variables
    gppr <- s$output$sqrt()
    dX <- grad_output * gppr
    q <- dX$sum(s$dim) / gppr$sum(s$dim)
    q <- q$unsqueeze(s$dim)
    dX$sub_(q*gppr)

    list(
      input = dX,
      dim = dim,
      k = s$k
    )
  }
)

#' Alpha-entmax 
#' 
#' With alpha = 1.5 and normalizing sparse transform (a la softmax).
#'
#' Solves the optimization problem:
#' \eqn{\max_p <input, P> - H_{1.5}(P) \text{ s.t. } P \geq 0, \sum(P) == 1}
#' where \eqn{H_{1.5}(P)} is the Tsallis alpha-entropy with \eqn{\alpha=1.5}.
#'
#' @param dim The dimension along which to apply 1.5-entmax.
#' @param k The number of largest elements to partial-sort input over. For optimal
#' performance, should be slightly bigger than the expected number of
#' non-zeros in the solution. If the solution is more than k-sparse,
#' this function is recursively called with a 2*k schedule. If `NULL`, full
#' sorting is performed from the beginning.
#'
#' @return The projection result P  of the same shape as input, such that
#'   \eqn{\sum_{dim} P = 1 \forall dim} elementwise.
#'
#' @examplesIf torch::torch_is_installed()
#' \dontrun{
#' input <- torch::torch_randn(10,5, requires_grad = TRUE)
#' # create a top3 alpha=1.5 entmax on last input dimension
#' nn_entmax <- entmax15(dim=-1L, k = 3)
#' result <- nn_entmax(input)
#' }
#' @export
entmax15 <- torch::nn_module(
  "entmax_15",
  initialize = function(dim = -1L, k = NULL) {
    self$dim <- dim
    self$k <- k
  },
  forward = function(input) {
    if (self$dim == -1L) {
      self$dim <- input$dim()
    }
    entmax_15_function(input, self$dim, self$k)
  }
)

#' Optimal threshold (tau) computation for 1.5-entmax
#'
#' @param input The input tensor to compute thresholds over.
#' @param dim The dimension along which to apply 1.5-entmax. Default is -1.
#' @param k The number of largest elements to partial-sort over. For optimal
#'   performance, should be slightly bigger than the expected number of
#'   non-zeros in the solution. If the solution is more than k-sparse,
#'   this function is recursively called with a 2*k schedule. If `NULL`, 
#'   full sorting is performed from the beginning. Default is NULL.
#'
#' @return The threshold value for each vector, with all but the `dim` 
#'   dimension intact.
get_tau <- function(input, dim = -1L, k = NULL) {
  tau_ <- ss_ <- NULL # avoid note 
  if (is.null(k) || k >= input$size(dim)) { # do full sort
    sorted_input <- input$sort(dim = dim, descending = TRUE)[[1]]
  } else {
    sorted_input <- input$topk(k = k, dim = dim)[[1]]
  }
  
  rho <- .make_ix_like(sorted_input, dim)
  mean <- sorted_input$cumsum(dim) / rho
  mean_sq <- sorted_input$pow(2)$cumsum(dim) / rho
  ss <- rho * (mean_sq - mean$pow(2))
  delta <- (1 - ss) / rho
  
  # NOTE this is not exactly the same as in reference algo
  # Fortunately it seems the clamped values never wrongly
  # get selected by tau <= sorted_z. Prove this!
  delta_nz <- torch::torch_clamp(delta, 0)
  tau <- mean - delta_nz$sqrt()
  
  support_size <- (tau <= sorted_input)$sum(dim)$unsqueeze(dim)
  tau_star <-  tau$gather(dim, support_size)
  
  if (!is.null(k) && k < input$size(dim)) {
    unsolved <- (support_size == k)$squeeze(dim)
    
    if (as.numeric(unsolved$sum()) > 0) {
      input_ <- .roll_last(input, dim)[unsolved]
      c(tau_, ss_) %<-% .entmax_threshold_and_support(input_, dim = -1L, k = 2 * k)
      tau_star[unsolved] <- tau_
      support_size[unsolved] <- ss_
    }
  }  
  return(tau_star)
}

relu15_function <- torch::autograd_function(
  forward = function(ctx, input, dim = 0L, tau = 0) {
    logit <- torch::torch_clamp(input / 2 - tau, min = 0)
    ctx$save_for_backward(logit)
    output <- logit$pow(2)
    output
  },
  backward = function(ctx, dY) {
    logit <- ctx$get_saved_tensors()
    return(list(logit, NULL, NULL))
  }
)

relu15 <- torch::nn_module(
  "relu15",
  initialize = function(dim = -1L, tau = 0.0) {
    self$dim <- dim
    self$tau <- tau
  },
  forward = function(input) {
    return(relu15_function$apply(input, self$dim, self$tau))
  }
)

test_that(".sparsemax_threshold_and_support works as expected with default values", {
  input <- torch::torch_randn(10, 5)
  expect_no_error(
    result <- tabnet:::.sparsemax_threshold_and_support(input)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau)
  expect_tensor_shape(tau, c(input$shape[1], 1))
  expect_tensor(support_size)
})


test_that(".sparsemax_threshold_and_support works as expected with k < input$size(dim)", {
  input <- torch::torch_randn(10, 5)
  dim <- 1L
  k <- 3
  expect_no_error(
    result <- tabnet:::.sparsemax_threshold_and_support(input, dim, k)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau)
  expect_tensor_shape(tau, c(1, input$shape[2]))
  expect_tensor(support_size)
  expect_tensor_shape(support_size, c(1, input$shape[2]))
  
})

test_that(".sparsemax_threshold_and_support works as expected with k >= input$size(dim)", {
  input <- torch::torch_randn(10, 5)
  dim <- -2L
  k <- 7
  expect_no_error(
    result <- tabnet:::.sparsemax_threshold_and_support(input, dim, k)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau)
  expect_tensor_shape(tau, c(1, input$shape[2]))
  expect_tensor(support_size)
  expect_tensor_shape(support_size, c(1, input$shape[2]))
})



test_that(".entmax_threshold_and_support works as expected with default values", {
  input <- torch::torch_randn(10, 5)
  expect_no_error(
    result <- tabnet:::.entmax_threshold_and_support(input)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau_star <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau_star)
  expect_tensor_shape(tau_star, c(input$shape[1], 1))
  expect_tensor(support_size)
  expect_tensor_shape(support_size, c(input$shape[1], 1))
})


test_that(".entmax_threshold_and_support works as expected with k < input$size(dim)", {
  input <- torch::torch_randn(10, 5)
  dim <- 1L
  k <- 3
  expect_no_error(
    result <- tabnet:::.entmax_threshold_and_support(input, dim, k)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau_star <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau_star)
  expect_tensor_shape(tau_star, c(1, input$shape[2]))
  expect_tensor(support_size)
  expect_tensor_shape(support_size, c(1, input$shape[2]))
})


test_that(".entmax_threshold_and_support works as expected with k >= input$size(dim)", {
  input <- torch::torch_randn(10, 5)
  dim <- 2L
  k <- 12
  expect_no_error(
    result <- tabnet:::.entmax_threshold_and_support(input, dim, k)
  )
  expect_type(result, "list")
  expect_length(result, 2)
  
  tau_star <- result[[1]]
  support_size <- result[[2]]
  expect_tensor(tau_star)
  expect_tensor_shape(tau_star, c(input$shape[1], 1))
  expect_tensor(support_size)
  expect_tensor_shape(support_size, c(input$shape[1], 1))
})


test_that("fit works with entmax mask-type", {
  
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  
  expect_no_error(
    tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25, verbose = TRUE,
               config = tabnet_config( mask_type = "entmax"))
  )
})


test_that("sparsemax_15_function works as a proper autograd", {
  
  input = torch::torch_rand(10,2, requires_grad = TRUE)

  expect_no_error(
    output <- tabnet:::sparsemax_function(input, 2L, 3)
  )
  expect_no_error(
    output$backward
  )
})

test_that("entmax_15_function works as a proper autograd", {
  
  input = torch::torch_rand(10,2, requires_grad = TRUE)

  expect_no_error(
    output <- tabnet:::entmax_15_function(input, 2L, 3)
  )
  expect_no_error(
    output$backward
  )
})


test_that("fit works with sparsemax15 mask-type", {
  
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  
  expect_no_error(
    tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25, verbose = TRUE,
               config = tabnet_config( mask_type = "sparsemax15"))
  )
  expect_no_error(
    tabnet_fit(rec, attrition, epochs = 1, verbose = TRUE,
               config = tabnet_config( mask_type = "sparsemax15", mask_topk = 12))
  )
})

test_that("fit works with entmax15 mask-type", {
  
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  
  expect_no_error(
    tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25, verbose = TRUE,
               config = tabnet_config( mask_type = "entmax15"))
  )
  expect_no_error(
    tabnet_fit(rec, attrition, epochs = 1, verbose = TRUE,
               config = tabnet_config( mask_type = "entmax15", mask_topk = 12))
  )
})

train_batch_un <- function(network, optimizer, batch, config) {
  # forward pass
  output <- network(batch$x, batch$x_na_mask)
  loss <- config$loss_fn(output[[1]], output[[2]], output[[3]])

  # step of the backward pass and optimization
  optimizer$zero_grad()
  loss$backward()
  if (!is.null(config$clip_value)) {
    torch::nn_utils_clip_grad_norm_(network$parameters, config$clip_value)
  }
  optimizer$step()

  list(
    loss = loss$item()
  )
}

valid_batch_un <- function(network, batch, config) {
  # forward pass
  output <- network(batch$x, batch$x_na_mask)
  # we inverse the batch_na_mask here to avoid nan in the loss
  loss <- config$loss_fn(output[[1]], output[[2]], output[[3]]$logical_not())

  list(
    loss = loss$item()
  )
}

transpose_metrics <- function(metrics) {
  nms <- names(metrics[1])
  out <- vector(mode = "list", length = length(nms))
  for (nm in nms) {
    out[[nm]] <- vector("numeric", length = length(metrics))
  }

  for (i in seq_along(metrics)) {
    for (nm in nms) {
      out[[nm]][i] <- metrics[i][[nm]]
    }
  }

  out[-1]
}


tabnet_train_unsupervised <- function(x, config = tabnet_config(), epoch_shift = 0L) {
  torch::torch_manual_seed(sample.int(1e6, 1))

  device <- get_device_from_config(config)

  # validation dataset & dataloaders
  has_valid <- config$valid_split > 0
  if (has_valid) {
    n <- nrow(x)
    valid_idx <- sample.int(n, n*config$valid_split)
    valid_x <- x[valid_idx, ]
    valid_ds <-   torch::dataset(
      initialize = function() {},
      .getbatch = function(batch) {resolve_data(valid_x[batch,], matrix(1L, nrow = length(batch)))},
      .length = function() {nrow(valid_x)}
    )()

    valid_dl <- torch::dataloader(
      valid_ds,
      batch_size = config$batch_size,
      shuffle = FALSE ,
      num_workers = config$num_workers
    )

    x <- x[-valid_idx, ]
  }

  # training dataset & dataloader
  train_ds <-   torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {resolve_data(x[batch,], matrix(1L, nrow = length(batch)))},
    .length = function() {nrow(x)}
  )()
  # we can get training_set parameters from the 2 first samples
  train <- train_ds$.getbatch(batch = c(1:2))

  train_dl <- torch::dataloader(
    train_ds,
    batch_size = config$batch_size,
    drop_last = config$drop_last,
    shuffle = TRUE ,
    num_workers = config$num_workers
  )

  # resolve loss (shortcutted from config)
  config$loss_fn <- nn_unsupervised_loss()

  # create network
  network <- tabnet_pretrainer(
    input_dim = train$input_dim,
    cat_idxs = train$cat_idx,
    cat_dims = train$cat_dims,
    pretraining_ratio = config$pretraining_ratio,
    n_d = config$n_d,
    n_a = config$n_a,
    n_steps = config$n_steps,
    gamma = config$gamma,
    virtual_batch_size = config$virtual_batch_size,
    cat_emb_dim = config$cat_emb_dim,
    n_independent = config$n_independent,
    n_shared = config$n_shared,
    n_independent_decoder = config$n_independent_decoder,
    n_shared_decoder = config$n_shared_decoder,
    momentum = config$momentum,
    mask_type = config$mask_type,
    mask_topk = config$mask_topk
  )

  network$to(device = device)

  # instantiate optimizer
  if (is_optim_generator(config$optimizer)) {
    optimizer <- config$optimizer(network$parameters, config$learn_rate)
  } else {
    type_error("{.var optimizer} must be resolved into a torch optimizer generator.")
  }
  

  # define scheduler
  if (is.null(config$lr_scheduler)) {
    scheduler <- list(step = function() {})
  } else if (rlang::is_function(config$lr_scheduler)) {
    scheduler <- config$lr_scheduler(optimizer)
  } else if (config$lr_scheduler == "reduce_on_plateau") {
    scheduler <- torch::lr_reduce_on_plateau(optimizer, factor = config$lr_decay, patience = config$step_size)
  } else if (config$lr_scheduler == "step") {
    scheduler <- torch::lr_step(optimizer, config$step_size, config$lr_decay)
  } else {
    not_implemented_error("Currently only the {.str step} and {.str reduce_on_plateau} scheduler are supported.", call. = FALSE)
  }

  # initialize metrics & checkpoints
  metrics <- list()
  checkpoints <- list()
  patience_counter <- 0L

  # main loop
  for (epoch in seq_len(config$epochs) + epoch_shift) {

    metrics[[epoch]] <- list()
    train_metrics <- c()
    valid_metrics <- c()

    network$train()

    if (config$verbose)
      pb <- progress::progress_bar$new(
        total = length(train_dl),
        format = "[:bar] loss= :loss"
      )

    coro::loop(for (batch in train_dl) {
      m <- train_batch_un(network, optimizer, to_device(batch, device), config)
      if (config$verbose) pb$tick(tokens = m)
      train_metrics <- c(train_metrics, m)
    })
    metrics[[epoch]][["train"]] <- transpose_metrics(train_metrics)$loss

    if (config$checkpoint_epochs > 0 && epoch %% config$checkpoint_epochs == 0) {
      checkpoints[[length(checkpoints) + 1]] <- state_to_raw(network)
      metrics[[epoch]][["checkpoint"]] <- TRUE
    }

    network$eval()
    if (has_valid) {
      coro::loop(for (batch in valid_dl) {
        m <- valid_batch_un(network, to_device(batch, device), config)
        valid_metrics <- c(valid_metrics, m)
      })
      metrics[[epoch]][["valid"]] <- transpose_metrics(valid_metrics)$loss
    }

    if (config$verbose && !has_valid)
      message(gettextf("[Epoch %03d] Loss: %3f", epoch, mean(metrics[[epoch]]$train)))
    if (config$verbose && has_valid)
      message(gettextf("[Epoch %03d] Loss: %3f, Valid loss: %3f", epoch, mean(metrics[[epoch]]$train), mean(metrics[[epoch]]$valid)))

    # Early-stopping checks
    if (config$early_stopping && config$early_stopping_monitor=="valid_loss"){
      current_loss <- mean(metrics[[epoch]]$valid)
    } else {
      current_loss <- mean(metrics[[epoch]]$train)
    }
    if (config$early_stopping && epoch > 1+epoch_shift) {
      # compute relative change, and compare to best_metric
      change <- (current_loss - best_metric) / current_loss
      if (change > config$early_stopping_tolerance) {
        patience_counter <- patience_counter + 1
        if (patience_counter >= config$early_stopping_patience) {
          if (config$verbose)
            cli::cli_alert_success(gettextf("Early-stopping at epoch {.val epoch}"))
          break
        }
      } else {
        # reset the patience counter
        best_metric <- current_loss
        patience_counter <- 0L
      }
    }
    if (config$early_stopping && epoch == 1 + epoch_shift) {
      # initialise best_metric
      best_metric <- current_loss
    }

    if ("metrics" %in% names(formals(scheduler$step))) {
      scheduler$step(current_loss)
    } else {
      scheduler$step()
    }
  }

  network$to(device = "cpu")
  if(!config$skip_importance) {
    importance_sample_size <- config$importance_sample_size
    if (is.null(config$importance_sample_size) && train_ds$.length() > 1e5) {
      warn("Computing importances for a dataset with size {.val {train_ds$.length()}}. 
           This can consume too much memory. We are going to use a sample of size 1e5. 
           You can disable this message by using the `importance_sample_size` argument.")
      importance_sample_size <- 1e5
    }
    indexes <- as.numeric(torch::torch_randint(
      1, train_ds$.length(), min(importance_sample_size, train_ds$.length()),
      dtype = torch::torch_long()
    ))
    importances <- tibble::tibble(
      variables = colnames(x),
      importance = compute_feature_importance(
        network,
        train_ds$.getbatch(batch =indexes)$x$to(device = "cpu"),
        train_ds$.getbatch(batch =indexes)$x_na_mask$to(device = "cpu"))
    )
  } else {
    importances <- NULL
  }

  list(
    network = network,
    metrics = metrics,
    config = config,
    checkpoints = checkpoints,
    importances = importances
  )
}

# Run before any test
suppressPackageStartupMessages(library(recipes))
suppressPackageStartupMessages(library(ggplot2))
suppressPackageStartupMessages(library(data.tree))


# ames small data
utils::data("ames", package = "modeldata")
ids <- sample(nrow(ames), 256)
small_ames <- ames[ids,]
x <- ames[ids,-which(names(ames) == "Sale_Price")]
y <- ames[ids,]$Sale_Price


# attrition small data
utils::data("attrition", package = "modeldata")
ids <- sample(nrow(attrition), 256)

# attrition common models
attrix <- attrition[ids,-which(names(attrition) == "Attrition")]
attri_mult_x <- attrix[-which(names(attrix) == "JobSatisfaction")]

attriy <- attrition[ids,]$Attrition

# data.tree Node dataset
utils::data("acme", package = "data.tree")
acme_df <-  data.tree::ToDataFrameTypeCol(acme, acme$attributesAll) %>%
  select(-starts_with("level_"))
# acme2 <- acme$clone()
# acme2$RemoveAttribute("level_3")

attrition_tree <- attrition %>%
  tibble::rowid_to_column() %>%
  mutate(pathString = paste("attrition", Department, JobRole, rowid, sep = "/")) %>%
  select(-Department, -JobRole, -rowid) %>%
  data.tree::as.Node()

merge_config_and_dots <- function(config, ...) {
  default_config <- tabnet_config()
  new_config <- do.call(tabnet_config, list(...))
  # TODO currently we cannot not compare two nn_optimizer nor nn_loss values 
  new_config <- new_config[
    mapply(
      function(x, y) ifelse(is_null_or_optim_generator_or_loss(x), 
                            !is_null_or_optim_generator_or_loss(y), # TRUE 
                            ifelse(is_optim_generator_or_loss(y), TRUE, x != y)), # FALSE 
      default_config,
      new_config)
  ]
  merged_config <- utils::modifyList(config, as.list(new_config))
  merged_config$optimizer <- resolve_optimizer(merged_config$optimizer)
  merged_config
}
# 
# is_different_param <- function(x, y) {
#   if (rlang::inherits_any(x, c("nn_loss", "nn_optim_generatorclass"))) {
#     
#   }
# }

check_net_is_empty_ptr <- function(object) {
  is_null_external_pointer(object$fit$network$.check$ptr)
}

# https://stackoverflow.com/a/27350487/3297472
is_null_external_pointer <- function(pointer) {
  a <- attributes(pointer)
  attributes(pointer) <- NULL
  out <- identical(pointer, methods::new("externalptr"))
  attributes(pointer) <- a
  out
}

#' Check that Node object names are compliant
#'
#' @param node the Node object, or a dataframe ready to be parsed by `data.tree::as.Node()`
#'
#' @return node if it is compliant, else an Error with the column names to fix
#' @export
#'
#' @examplesIf (require("data.tree") || require("dplyr"))
#' library(dplyr)
#' library(data.tree)
#' data(starwars)
#' starwars_tree <- starwars %>%
#'   mutate(pathString = paste("tree", species, homeworld, `name`, sep = "/"))
#'
#' # pre as.Node() check
#' try(check_compliant_node(starwars_tree))
#'
#' # post as.Node() check
#' check_compliant_node(as.Node(starwars_tree))
#'
check_compliant_node <- function(node) {
  #  prevent reserved data.tree Node colnames and the level_1 ... level_n names used for coercion
  if (inherits(node, "Node")) {
    # Node has already lost its reserved colnames
    reserved_names <- paste0("level_", c(1:node$height))
    actual_names <- node$attributesAll
  } else if (inherits(node, "data.frame") && "pathString" %in% colnames(node)) {
    node_height <- max(stringr::str_count(node$pathString, "/"))
    reserved_names <- c(paste0("level_", c(1:node_height)), data.tree::NODE_RESERVED_NAMES_CONST)
    actual_names <- colnames(node)[!colnames(node) %in% "pathString"]
  } else {
    type_error("The provided hierarchical object is not recognized with a valid format that can be checked")
  }
  
  if (any(actual_names %in% reserved_names)) {
    value_error("The attributes or colnames in the provided hierarchical object use the following reserved names:
                {.vars {actual_names[actual_names %in% reserved_names]}}. 
                Please change those names as they will lead to unexpected tabnet behavior.")
  }
  

  
  
  invisible(node)
}

#' Turn a Node object into predictor and outcome.
#'
#' @param x Node object
#' @param drop_last_level TRUE unused
#'
#' @return a named list of x and y, being respectively the predictor data-frame and the outcomes data-frame,
#'   as expected inputs for `hardhat::mold()` function.
#' @export
#'
#' @examplesIf (require("data.tree") || require("dplyr"))
#' library(dplyr)
#' library(data.tree)
#' data(starwars)
#' starwars_tree <- starwars %>%
#'   mutate(pathString = paste("tree", species, homeworld, `name`, sep = "/")) %>%
#'   as.Node()
#' node_to_df(starwars_tree)$x %>% head()
#' node_to_df(starwars_tree)$y %>% head()
#' @importFrom dplyr last_col mutate mutate_if select starts_with where
node_to_df <- function(x, drop_last_level = TRUE) {
  # TODO get rid of all those import through base R equivalent
  xy_df <- data.tree::ToDataFrameTypeCol(x, x$attributesAll)
  x_df <- xy_df %>%
    select(-starts_with("level_")) %>%
    mutate_if(is.character, as.factor)
  y_df <- xy_df %>%
    select(starts_with("level_")) %>%
    # drop first (and all zero-variance) column
    select(where(~ nlevels(as.factor(.x)) > 1 )) %>%
    # TODO take the drop_last_level param into account
    # drop last level column
    select(-last_col()) %>%
    # TODO impute "NA" with parent through coalesce() via an option
    mutate_if(is.character, as.factor)
  return(list(x = x_df, y = y_df))
}


model_to_raw <- function(model) {
  con <- rawConnection(raw(), open = "wr")
  torch::torch_save(model, con)
  on.exit({close(con)}, add = TRUE)
  r <- rawConnectionValue(con)
  r
}

# Serialize a CPU state dict without moving the live model.
# Moving the model between devices during training breaks optimizer parameter
# references on MPS/CUDA, causing the optimizer to stop updating weights.
state_to_raw <- function(model) {
  con <- rawConnection(raw(), open = "wr")
  on.exit(close(con), add = TRUE)
  cpu_state <- lapply(model$state_dict(), function(t) t$cpu())
  torch::torch_save(cpu_state, con)
  rawConnectionValue(con)
}

# generalize torch to_device to nested list of tensors
to_device <- function(x, device) {
  lapply(x, function(x) {
    if (inherits(x, "torch_tensor")) {
      x$to(device=device)
    } else if (is.list(x)) {
      lapply(x, to_device)
    } else {
      x
    }
  })
}

# `optim_ignite_*` requires a minimum torch version
torch_has_optim_ignite <- function() {
  utils::compareVersion(as.character(utils::packageVersion("torch")), "0.14.0") >= 0
}

# turn "adam" or a torch_optim_* generator into a proper torch_optim_ generator
resolve_optimizer <- function(optimizer) {
  if (is_optim_generator(optimizer)) {
    torch_optimizer <- optimizer
  } else if (is.character(optimizer)) {
    if (optimizer == "adam" && torch_has_optim_ignite()) {
      torch_optimizer <- torch::optim_ignite_adam
    } else if (optimizer == "adam") {
      torch_optimizer <- torch::optim_adam
    } else {
      value_error("Currently only {.val adam} is supported as character for {.var optimizer}.")
    }
  } else {
    value_error("Currently only {.val adam} is supported as character for {.var optimizer}.")
  }
  torch_optimizer
  
}


is_optim_generator <- function(x) {
  inherits(x, "torch_optimizer_generator")
}

is_loss_generator <- function(x) {
  rlang::inherits_all(x, c("nn_loss", "nn_module_generator"))
}

is_null_or_optim_generator_or_loss <- function(x) {
  is.null(x) || is_optim_generator(x) || inherits(x, "nn_loss")
}

is_optim_generator_or_loss <- function(x) {
  is_optim_generator(x) || inherits(x, "nn_loss")
}


value_error <- function(..., env = rlang::caller_env()) {
  cli::cli_abort(gettext(..., domain = "R-tabnet")[[1]], .envir = env)
}

type_error <- function(..., env = rlang::caller_env()) {
  cli::cli_abort(gettext(..., domain = "R-tabnet")[[1]], .envir = env)
}

runtime_error <- function(..., env = rlang::caller_env()) {
  cli::cli_abort(gettext(..., domain = "R-tabnet")[[1]], .envir = env)
}

not_implemented_error <- function(..., env = rlang::caller_env()) {
  cli::cli_abort(gettext(..., domain = "R-tabnet")[[1]], .envir = env)
}

warn <- function(..., env = rlang::caller_env()) {
  cli::cli_warn(gettext(..., domain = "R-tabnet")[[1]], .envir = env)
}

test_that("errors when using an argument that do not exist", {

  expect_error(
    fit <- tabnet_fit(x, y, epochsas = 1),
    "unused argument"
  )

  expect_error(
    fit <- tabnet_fit(x, y, config = tabnet_config(epochsas = 1)),
    "unused argument"
  )

})

test_that("merging parameters from config and dots works", {
  
  # we can merge non-atomic values on different variables (no comparison)
  expect_no_error(
    merged_conf <- tabnet:::merge_config_and_dots(
      config = tabnet_config(optimizer = torch::optim_adamw),
      loss = torch::nn_bce_loss)
  )
  expect_identical_modules(merged_conf$optimizer, torch::optim_adamw)
  expect_identical_modules(merged_conf$loss, torch::nn_bce_loss)
  
  # we can merge non-atomic values on different variables while resolving optimizer
  expect_no_error(
    merged_conf <- tabnet:::merge_config_and_dots(
      config = tabnet_config(optimizer = "adam"),
      loss = torch::nn_bce_loss)
  )
  if(tabnet:::torch_has_optim_ignite()) {
    expect_identical_modules(merged_conf$optimizer, torch::optim_ignite_adam)
  } else {
    expect_identical_modules(merged_conf$optimizer, torch::optim_adam)
  }
  expect_identical_modules(merged_conf$loss, torch::nn_bce_loss)
  
  # ... value wins over tabnet_config() value
  expect_no_error(
    merged_conf <- tabnet:::merge_config_and_dots(
      config = tabnet_config(batch_size = 200),
      batch_size = 400)
  )
  expect_identical(merged_conf$batch_size, 400)
  
  # ... value wins over tabnet_config() value for non-atomic parameter
  expect_no_error(
    merged_conf <- tabnet:::merge_config_and_dots(
      config = tabnet_config(loss = torch::nn_cross_entropy_loss),
      loss = torch::nn_bce_loss)
  )
  expect_identical_modules(merged_conf$loss, torch::nn_bce_loss)
  
  # NULL value get replaced and optimizer is resolved even lately
  expect_no_error(
    merged_conf <- tabnet:::merge_config_and_dots(
      config = tabnet_config(loss = NULL, device = "cuda"),
      loss = torch::nn_bce_loss, optimizer = "adam")
  )
  if(tabnet:::torch_has_optim_ignite()) {
    expect_identical_modules(merged_conf$optimizer, torch::optim_ignite_adam)
  } else {
    expect_identical_modules(merged_conf$optimizer, torch::optim_adam)
  }
  expect_identical_modules(merged_conf$loss, torch::nn_bce_loss)

  
})

test_that("pretrain and fit both work with early stopping", {

  expect_message(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 100, valid_split = 0.5, verbose=TRUE,
                                early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
    "Early-stopping at epoch"
  )
  expect_lt(length(pretrain$fit$metrics),100)

  expect_message(
    fit <- tabnet_fit(attrix, attriy, epochs = 100, valid_split = 0.5, verbose=TRUE,
                      early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
    "Early-stopping at epoch"
  )
  expect_lt(length(fit$fit$metrics),100)

})

test_that("early stopping works wo validation split", {

  # tabnet_pretrain
  expect_message(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 100, verbose=TRUE,
                                early_stopping_monitor="train_loss",
                                early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
    "Early-stopping at epoch"
  )
  expect_lt(length(pretrain$fit$metrics),100)

  expect_error(
    tabnet_pretrain(attrix, attriy, epochs = 100, verbose=TRUE,
                                early_stopping_monitor="cross_validation_loss",
                                early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
    regexp = "not a valid early-stopping metric to monitor"
  )

  # tabnet_fit
  expect_message(
    fit <- tabnet_fit(attrix, attriy, epochs = 200, verbose=TRUE,
                      early_stopping_monitor="train_loss",
                      early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.3),
    "Early-stopping at epoch"
  )
  expect_lt(length(fit$fit$metrics),200)

  expect_error(
    tabnet_fit(attrix, attriy, epochs = 200, verbose=TRUE,
                      early_stopping_monitor="cross_validation_loss",
                      early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
    regexp = "not a valid early-stopping metric to monitor"
  )

})

test_that("configuration with categorical_embedding_dimension vector works", {

  config <- tabnet_config(cat_emb_dim=c(1,1,2,2,1,1,1,2,1,1,1,2,2,2))

  expect_no_error(
    fit <- tabnet_fit(attrix, attriy, epochs = 1, valid_split = 0.2, config=config)
  )
})

test_that("explicit error message when categorical embedding dimension vector has wrong size", {

  config <- tabnet_config(cat_emb_dim=c(1,1,2,2))

  expect_error(
    fit <- tabnet_fit(attrix, attriy, epochs = 1, valid_split = 0.2, config=config),
    regexp = "number of categorical predictors"
  )
})

test_that("step scheduler works", {

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, lr_scheduler = "step",
                      lr_decay = 0.1, step_size = 1)
  )

  sc_fn <- function(optimizer) {
    torch::lr_step(optimizer, step_size = 1, gamma = 0.1)
  }

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, lr_scheduler = sc_fn,
                      lr_decay = 0.1, step_size = 1)
  )

})

test_that("configuring optimizer works", {

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, config = tabnet_config(optimizer = "adam"))
  )

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, optimizer = torch::optim_adamw)
  )

  expect_error(
    fit <- tabnet_fit(x, y, epochs = 3, optimizer = "adamw"),
    "Currently only \"adam\" is supported"
  )

})

test_that("reduce_on_plateau scheduler works", {

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, lr_scheduler = "reduce_on_plateau",
                      lr_decay = 0.1, step_size = 1)
  )

  expect_error(
    fit <- tabnet_fit(x, y, epochs = 3, lr_scheduler = "multiplicative",
                      lr_decay = 0.1, step_size = 1),
    "only the \"step\" and \"reduce_on_plateau\" scheduler"
  )

  sc_fn <- function(optimizer) {
    torch::lr_reduce_on_plateau(optimizer, factor = 0.1, patience = 10)
  }

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, lr_scheduler = sc_fn,
                      lr_decay = 0.1, step_size = 1)
  )

})

test_that("fit uses config parameters mix from config= and ...", {

  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit <- tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25, verbose = TRUE,
                    config = tabnet_config(decision_width=3, attention_width=5, cat_emb_dim = 2))
  expect_no_error(
    predict(fit, attrition)
  )

  expect_equal(fit$fit$config$verbose, TRUE)
  expect_equal(fit$fit$config$valid_split, 0.25)
  expect_equal(fit$fit$config$n_d, 3)
  expect_equal(fit$fit$config$n_a, 5)

})


test_that("fit raise an error with non-supported mask-type", {

  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  expect_error(
    tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25, verbose = TRUE,
                      config = tabnet_config( mask_type="max_entropy")),
    regexp = "either \"sparsemax\", \"sparsemax15\", \"entmax\" or \"entmax15\" as"
  )

})

test_that("config$loss=`auto` adapt to recipe outcome str()", {

  testthat::skip_on_ci()

  # nominal outcome
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit_auto <- tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                      config = tabnet_config( loss="auto"))
  expect_equal(fit_auto$fit$config$loss_fn, torch::nn_cross_entropy_loss(), ignore_function_env = TRUE)

  # numerical outcome
  rec <- recipe(MonthlyIncome ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit_auto <- tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                      config = tabnet_config( loss="auto"))
  expect_equal(fit_auto$fit$config$loss_fn, torch::nn_mse_loss(), ignore_function_env = TRUE)

})

test_that("config$loss not adapted to recipe outcome raise an explicit error", {

  testthat::skip_on_ci()

  # nominal outcome with numerical loss
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  expect_error(tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                          config = tabnet_config( loss="mse")),
              regexp = "is not a valid loss for outcome of type"
  )
  # numerical outcome
  rec <- recipe(MonthlyIncome ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  expect_error(tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                      config = tabnet_config( loss="cross_entropy")),
               regexp = "is not a valid loss for outcome of type"
  )
})


test_that("config$loss can be a function", {

  testthat::skip_on_ci()

  # nominal outcome loss
  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit_auto <- tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                      config = tabnet_config( loss = torch::nn_nll_loss()))
  expect_equal(fit_auto$fit$config$loss_fn, torch::nn_nll_loss(), ignore_function_env = TRUE)

  # numerical outcome loss
  rec <- recipe(MonthlyIncome ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit_auto <- tabnet_fit(rec, attrition[ids, ], epochs = 1, verbose = TRUE,
                      config = tabnet_config( loss = torch::nn_poisson_nll_loss()))
  expect_equal(fit_auto$fit$config$loss_fn, torch::nn_poisson_nll_loss(), ignore_function_env = TRUE)

})

test_that("Training hierarchical classification for {data.tree} Node attrition_tree", {
  expect_no_error(
    fit <- tabnet_fit(attrition_tree, epochs = 1)
  )
  expect_no_error(
    result <- predict(fit, attrition_tree, type = "prob")
  )
  expect_no_error(
    augmented <- augment(fit, attrition_tree, type = "prob")
  )

  expect_equal(ncol(result), 2) # 2 outcomes levels_

  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  # we get back outcomes vars with a `.pred_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_"), names(outcome_nlevels))

  # result columns are tibbles of resp 2, 2, 4 columns
  expect_true(all(purrr::map_lgl(result, tibble::is_tibble)))
  expect_equal(unname(purrr::map_dbl(result, ncol)), unname(outcome_nlevels), ignore_attr = TRUE)

})

test_that("Training hierarchical classification for {data.tree} Node with validation split", {

  expect_no_error(
    fit <- tabnet_fit(attrition_tree, valid_split = 0.2, epochs = 1)
  )
  expect_true( "ancestor" %in% names(fit$fit$config))
  
  expect_no_error(
    result <- predict(fit, attrition_tree, type = "prob")
  )

  expect_equal(ncol(result), 2) # 2 outcomes levels_

  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  # we get back outcomes vars with a `.pred_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_"), names(outcome_nlevels))

  # result columns are tibbles of resp 2, 2, 4 columns
  expect_true(all(purrr::map_lgl(result, tibble::is_tibble)))
  expect_equal(unname(purrr::map_dbl(result, ncol)), unname(outcome_nlevels), ignore_attr = TRUE)

  expect_no_error(
    result <- predict(fit, attrition_tree)
  )
  expect_equal(ncol(result), 2) # 2 outcomes levels_

  # we get back outcomes vars with a `.pred_class_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_class_"), names(fit$blueprint$ptypes$outcomes))
})

test_that("hierarchical classification for {data.tree} Node is explainable", {

    fit <- tabnet_fit(attrition_tree, epochs = 1)

  expect_no_error(
    explain <- tabnet_explain(fit, attrition_tree)
  )

  expect_no_error(
    autoplot(explain)
  )

})

test_that("Training hierarchical classification for {data.tree} Node for starwars", {

  starwars_tree <- starwars %>% 
    rename(`_name` = "name", `_height` = "height") %>% 
    mutate(species = coalesce(species, "Unknown_Species"),
           sex     = coalesce(sex, "Unknown_Sex"),
           pathString = paste("StarWars_characters", species, sex, `_name`, sep = "/")) %>%
    as.Node()
  
  expect_no_error(  fit <- tabnet_fit(starwars_tree, epochs = 1))
})

test_that("we properly check non-compliant colnames", {

  # try to use starwars dataset with two forbidden column name
  starwars_tree <- starwars %>%
    mutate(species = coalesce(species, "Unknown_Species"),
           sex     = coalesce(sex, "Unknown_Sex"),
           pathString = paste("tree", species, homeworld, `name`, sep = "/"))
  expect_error(
    check_compliant_node(starwars_tree)
    ,"reserved names")

  # augment acme dataset with a forbidden column name with no impact on predictor is ok
  acme$Do(function(x) {
    x$level_4 <- as.character(data.tree::Aggregate(node = x,
                           attribute = "p",
                           aggFun = sum))
  },
  traversal = "post-order")
  expect_no_error(check_compliant_node(acme))

  expect_no_error(tabnet_fit(acme, epochs = 1))

  # augment acme dataset with a used forbidden column name raise error
  acme$Do(function(x) {
    x$level_3 <- data.tree::Aggregate(node = x,
                           attribute = "p",
                           aggFun = sum)
  },
  traversal = "post-order")
  expect_error(
    check_compliant_node(acme)
    ,"reserved names")

  expect_error(
    tabnet_fit(acme, epochs = 1)
    ,"reserved names")

})

.onLoad <- function(...) {
  vctrs::s3_register("parsnip::multi_predict", "_tabnet_fit")
  vctrs::s3_register("vip::vi_model", "tabnet_fit")
  vctrs::s3_register("vip::vi_model", "tabnet_pretrain")
  vctrs::s3_register("ggplot2::autoplot", "tabnet_fit")
  vctrs::s3_register("ggplot2::autoplot", "tabnet_pretrain")
  vctrs::s3_register("ggplot2::autoplot", "tabnet_explain")
  vctrs::s3_register("torch::nn_prune_head", "tabnet_fit")
  vctrs::s3_register("torch::nn_prune_head", "tabnet_pretrain")
  vctrs::s3_register("tune::min_grid", "tabnet")
}


globalVariables(c("batch_size",
                  "checkpoint",
                  "dataset",
                  "epoch",
                  "has_checkpoint",
                  "loss",
                  "mask_agg",
                  "mean_loss",
                  "row_number",
                  "rowname",
                  "step",
                  "value",
                  "variable",
                  ".."))

#' Tabnet model
#'
#' Fits the [TabNet: Attentive Interpretable Tabular Learning](https://arxiv.org/abs/1908.07442) model
#'
#' @param x Depending on the context:
#'
#'   * A __data frame__ of predictors.
#'   * A __matrix__ of predictors.
#'   * A __recipe__ specifying a set of preprocessing steps
#'     created from [recipes::recipe()].
#'   * A __Node__ where tree will be used as hierarchical outcome,
#'     and attributes will be used as predictors.
#'
#'  The predictor data should be standardized (e.g. centered or scaled).
#'  The model treats categorical predictors internally thus, you don't need to
#'  make any treatment.
#'  The model treats missing values internally thus, you don't need to make any
#'  treatment.
#'
#' @param y When `x` is a __data frame__ or __matrix__, `y` is the outcome
#' specified as:
#'
#'   * A __data frame__ with 1 or many numeric column (regression) or 1 or many categorical columns (classification) .
#'   * A __matrix__ with 1 column.
#'   * A __vector__, either numeric or categorical.
#'
#' @param data When a __recipe__ or __formula__ is used, `data` is specified as:
#'
#'   * A __data frame__ containing both the predictors and the outcome.
#'
#' @param formula A formula specifying the outcome terms on the left-hand side,
#'  and the predictor terms on the right-hand side.
#' @param tabnet_model A previously fitted `tabnet_model` object to continue the fitting on.
#'  if `NULL` (the default) a brand new model is initialized.
#' @param config A set of hyperparameters created using the `tabnet_config` function.
#'  If no argument is supplied, this will use the default values in [tabnet_config()].
#' @param from_epoch When a `tabnet_model` is provided, restore the network weights from a specific epoch.
#'  Default is last available checkpoint for restored model, or last epoch for in-memory model.
#' @param weights Unused. Placeholder for hardhat::importance_weight() variables.
#' @param ... Model hyperparameters.
#' Any hyperparameters set here will update those set by the config argument.
#' See [tabnet_config()] for a list of all possible hyperparameters.
#'
#' @section Fitting a pre-trained model:
#'
#' When providing a parent `tabnet_model` parameter, the model fitting resumes from that model weights
#' at the following epoch:
#'    * last fitted epoch for a model already in torch context
#'    * Last model checkpoint epoch for a model loaded from file
#'    * the epoch related to a checkpoint matching or preceding the `from_epoch` value if provided
#' The model fitting metrics append on top of the parent metrics in the returned TabNet model.
#'
#' @section Multi-outcome:
#'
#' TabNet allows multi-outcome prediction, which is usually named [multi-label classification](https://en.wikipedia.org/wiki/Multi-label_classification)
#'   or multi-output regression when outcomes are numerical.
#' Multi-outcome currently expect outcomes to be either all numeric or all categorical.
#'
#' @section Threading:
#'
#' TabNet uses `torch` as its backend for computation and `torch` uses all
#' available threads by default.
#'
#' You can control the number of threads used by `torch` with:
#'
#' ```
#' torch::torch_set_num_threads(1)
#' torch::torch_set_num_interop_threads(1)
#' ```
#'
#' @examplesIf (torch::torch_is_installed() && require("modeldata"))
#' \dontrun{
#' data("ames", package = "modeldata")
#' data("attrition", package = "modeldata")
#'
#' ## Single-outcome regression using formula specification
#' fit <- tabnet_fit(Sale_Price ~ ., data = ames, epochs = 4)
#'
#' ## Single-outcome classification using data-frame specification
#' attrition_x <- attrition[ids,-which(names(attrition) == "Attrition")]
#' fit <- tabnet_fit(attrition_x, attrition$Attrition, epochs = 4, verbose = TRUE)
#'
#' ## Multi-outcome regression on `Sale_Price` and `Pool_Area` in `ames` dataset using formula,
#' ames_fit <- tabnet_fit(Sale_Price + Pool_Area ~ ., data = ames, epochs = 4, valid_split = 0.2)
#'
#' ## Multi-label classification on `Attrition` and `JobSatisfaction` in
#' ## `attrition` dataset using recipe
#' library(recipes)
#' rec <- recipe(Attrition + JobSatisfaction ~ ., data = attrition) %>%
#'   step_normalize(all_numeric(), -all_outcomes())
#'
#' attrition_fit <- tabnet_fit(rec, data = attrition, epochs = 4, valid_split = 0.2)
#'
#' ## Hierarchical classification on  `acme`
#' data(acme, package = "data.tree")
#'
#' acme_fit <- tabnet_fit(acme, epochs = 4, verbose = TRUE)
#'
#' # Note: Model's number of epochs should be increased for publication-level results.
#' }
#' @return A TabNet model object. It can be used for serialization, predictions, or further fitting.
#'
#' @export
tabnet_fit <- function(x, ...) {
  UseMethod("tabnet_fit")
}

#' @export
#' @rdname tabnet_fit
tabnet_fit.default <- function(x, ...) {
  type_error("{.fn tabnet_fit} is not defined for a {.type {class(x)[1])}}.")
}

#' @export
#' @rdname tabnet_fit
tabnet_fit.data.frame <- function(x, y, tabnet_model = NULL, config = tabnet_config(), ...,
                                  from_epoch = NULL, weights = NULL) {
  processed <- hardhat::mold(x, y)
  check_type(processed$outcomes)

  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "supervised")
}

#' @export
#' @rdname tabnet_fit
tabnet_fit.formula <- function(formula, data, tabnet_model = NULL, config = tabnet_config(), ...,
                               from_epoch = NULL, weights = NULL) {
  processed <- hardhat::mold(
    formula, data,
    blueprint = hardhat::default_formula_blueprint(
      indicators = "none",
      intercept = FALSE
    )
  )
  check_type(processed$outcomes)

  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "supervised")
}

#' @export
#' @rdname tabnet_fit
tabnet_fit.recipe <- function(x, data, tabnet_model = NULL, config = tabnet_config(), ...,
                              from_epoch = NULL, weights = NULL) {
  processed <- hardhat::mold(x, data)
  check_type(processed$outcomes)

  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "supervised")
}

#' @export
#' @rdname tabnet_fit
#'
#' @importFrom dplyr filter mutate select mutate_all mutate_if
#' @importFrom tidyr replace_na
#'
tabnet_fit.Node <- function(x, tabnet_model = NULL, config = tabnet_config(), ..., from_epoch = NULL) {
  # ensure there is no level_* col in the Node object
  check_compliant_node(x)
  # get tree leaves and extract attributes into data.frames
  xy_df <- node_to_df(x)
  processed <- hardhat::mold(xy_df$x, xy_df$y)
  check_type(processed$outcomes)
  
  config <- merge_config_and_dots(config, ...)
  # add ancestor boolean sparse matrix to config
  # check_dag_compliance(xy_df$y)
  config$ancestor <- build_ancestor_matrix_from_outcomes(x, processed$outcomes)
  # make outcomes levels available so that batched y could be one-hot encoded.
  config$outcomes <- processed$outcomes
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "supervised")
}

new_tabnet_fit <- function(fit, blueprint) {

  serialized_net <- model_to_raw(fit$network)

  hardhat::new_model(
    fit = fit,
    serialized_net = serialized_net,
    blueprint = blueprint,
    class = "tabnet_fit"
  )
}

#' Tabnet model
#'
#' Pretrain the [TabNet: Attentive Interpretable Tabular Learning](https://arxiv.org/abs/1908.07442) model
#' on the predictor data exclusively (unsupervised training).
#'
#' @param x Depending on the context:
#'
#'   * A __data frame__ of predictors.
#'   * A __matrix__ of predictors.
#'   * A __recipe__ specifying a set of preprocessing steps
#'     created from [recipes::recipe()].
#'   * A __Node__ where tree leaves will be left out,
#'     and attributes will be used as predictors.
#'
#'  The predictor data should be standardized (e.g. centered or scaled).
#'  The model treats categorical predictors internally thus, you don't need to
#'  make any treatment.
#'  The model treats missing values internally thus, you don't need to make any
#'  treatment.
#'
#' @param y (optional) When `x` is a __data frame__ or __matrix__, `y` is the outcome
#' @param data When a __recipe__ or __formula__ is used, `data` is specified as:
#'
#'   * A __data frame__ containing both the predictors and the outcome.
#'
#' @param formula A formula specifying the outcome terms on the left-hand side,
#'  and the predictor terms on the right-hand side.
#' @param tabnet_model A pretrained `tabnet_model` object to continue the fitting on.
#'  if `NULL` (the default) a brand new model is initialized.
#' @param config A set of hyperparameters created using the `tabnet_config` function.
#'  If no argument is supplied, this will use the default values in [tabnet_config()].
#' @param from_epoch When a `tabnet_model` is provided, restore the network weights from a specific epoch.
#'  Default is last available checkpoint for restored model, or last epoch for in-memory model.
#' @param ... Model hyperparameters.
#' Any hyperparameters set here will update those set by the config argument.
#' See [tabnet_config()] for a list of all possible hyperparameters.
#'
#'
#' @section outcome:
#'
#' Outcome value are accepted here only for consistent syntax with `tabnet_fit`, but
#' by design the outcome, if present, is ignored during pre-training.
#'
#' @section pre-training from a previous model:
#'
#' When providing a parent `tabnet_model` parameter, the model pretraining resumes from that model weights
#' at the following epoch:
#'    * last pretrained epoch for a model already in torch context
#'    * Last model checkpoint epoch for a model loaded from file
#'    * the epoch related to a checkpoint matching or preceding the `from_epoch` value if provided
#' The model pretraining metrics append on top of the parent metrics in the returned TabNet model.
#'
#' @section Threading:
#'
#' TabNet uses `torch` as its backend for computation and `torch` uses all
#' available threads by default.
#'
#' You can control the number of threads used by `torch` with:
#'
#' ```
#' torch::torch_set_num_threads(1)
#' torch::torch_set_num_interop_threads(1)
#' ```
#'
#' @examplesIf torch::torch_is_installed()
#' data("ames", package = "modeldata")
#' pretrained <- tabnet_pretrain(Sale_Price ~ ., data = ames, epochs = 1)
#'
#' @return A TabNet model object. It can be used for serialization, predictions, or further fitting.
#'
#' @export
tabnet_pretrain <- function(x, ...) {
  UseMethod("tabnet_pretrain")
}

#' @export
#' @rdname tabnet_pretrain
tabnet_pretrain.default <- function(x, ...) {
  type_error("{.fn tabnet_pretrain} is not defined for a {.type {class(x)[1])}}.")
}



#' @export
#' @rdname tabnet_pretrain
tabnet_pretrain.data.frame <- function(x, y = NULL, tabnet_model = NULL, config = tabnet_config(), ..., from_epoch = NULL) {
  processed <- hardhat::mold(x, y)

  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "unsupervised")
}

#' @export
#' @rdname tabnet_pretrain
tabnet_pretrain.formula <- function(formula, data, tabnet_model = NULL, config = tabnet_config(), ..., from_epoch = NULL) {
  processed <- hardhat::mold(
    formula, data,
    blueprint = hardhat::default_formula_blueprint(
      indicators = "none",
      intercept = FALSE
    )
  )
  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "unsupervised")
}

#' @export
#' @rdname tabnet_pretrain
tabnet_pretrain.recipe <- function(x, data, tabnet_model = NULL, config = tabnet_config(), ..., from_epoch = NULL) {
  processed <- hardhat::mold(x, data)

  config <- merge_config_and_dots(config, ...)
  tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "unsupervised")
}

#' @export
#' @rdname tabnet_pretrain
tabnet_pretrain.Node <- function(x, tabnet_model = NULL, config = tabnet_config(), ..., from_epoch = NULL) {
  # ensure there is no level_* col in the Node object
  check_compliant_node(x)
  # get tree leaves and extract attributes into data.frames
  xy_df <- node_to_df(x)
  tabnet_pretrain(xy_df$x, tabnet_model = tabnet_model, config = config, ..., from_epoch = from_epoch)
}

new_tabnet_pretrain <- function(pretrain, blueprint) {

  serialized_net <- model_to_raw(pretrain$network)

  hardhat::new_model(
    fit = pretrain,
    serialized_net = serialized_net,
    blueprint = blueprint,
    class = "tabnet_pretrain"
  )
}

#' Triple dispatch on task, resume training and resume epoch
#'
#' Perform the triple dispatch and initialize the model (if needed) or
#'  resume the model network weight to the right epoch
#'
#' @param processed the hardhat prerocessed dataset
#' @param config the tabnet network config list of parameters
#' @param tabnet_model the tabnet model to resume training on
#' @param from_epoch the epoch to resume training from
#' @param task "supervised" or "unsupervised"
#'
#' @return a fitted tabnet_model/tabnet_pretrain object list
#' @noRd
tabnet_bridge <- function(processed, config = tabnet_config(), tabnet_model, from_epoch, task="supervised") {
  predictors <- processed$predictors
  outcomes <- processed$outcomes
  epoch_shift <- 0L

  if (!(is.null(tabnet_model) || inherits(tabnet_model, "tabnet_fit") || inherits(tabnet_model, "tabnet_pretrain")))
    type_error("{.var {tabnet_model}} is not recognised as a proper TabNet model")

  if (!is.null(from_epoch) && !is.null(tabnet_model)) {
    # model must be loaded from checkpoint

    if (from_epoch > (length(tabnet_model$fit$checkpoints) * tabnet_model$fit$config$checkpoint_epoch))
      value_error("The model was trained for less than {.val {from_epoch}} epochs")

    # find closest checkpoint for that epoch
    closest_checkpoint <- from_epoch %/% tabnet_model$fit$config$checkpoint_epoch

    if (check_net_is_empty_ptr(tabnet_model)) {
      tabnet_model$fit$network <- reload_model(tabnet_model$serialized_net)
    }
    apply_checkpoint(tabnet_model$fit$network, tabnet_model$fit$checkpoints[[closest_checkpoint]])
    epoch_shift <- closest_checkpoint * tabnet_model$fit$config$checkpoint_epoch
    tabnet_model$fit$metrics <- tabnet_model$fit$metrics[seq(epoch_shift)]

  }
  if (task == "supervised") {
    if (sum(is.na(outcomes)) > 0) {
      value_error("Found missing values in the {.var {names(outcomes)}} outcome column.")
    }
    if (is.null(tabnet_model)) {
      # new supervised model needs network initialization
      tabnet_model_lst <- tabnet_initialize(predictors, outcomes, config = config)
      tabnet_model <-  new_tabnet_fit(tabnet_model_lst, blueprint = processed$blueprint)

    } else if (!check_net_is_empty_ptr(tabnet_model) && inherits(tabnet_model, "tabnet_fit")) {
      # resume training from supervised
      if (!identical(processed$blueprint, tabnet_model$blueprint))
        runtime_error("Model dimensions don't match.")

      # model is available from tabnet_model$serialized_net
      m <- reload_model(tabnet_model$serialized_net)

      # this modifies 'tabnet_model' in-place so subsequent predicts won't
      # need to reload.
      tabnet_model$fit$network$load_state_dict(m$state_dict())
      epoch_shift <- length(tabnet_model$fit$metrics)


    } else if (inherits(tabnet_model, "tabnet_pretrain")) {
      # resume training from unsupervised

      tabnet_model_lst <- model_pretrain_to_fit(tabnet_model, predictors, outcomes, config)
      tabnet_model <-  new_tabnet_fit(tabnet_model_lst, blueprint = processed$blueprint)

    }  else if (length(tabnet_model$fit$checkpoints)) {
      # model is loaded from the last available checkpoint

      last_checkpoint <- length(tabnet_model$fit$checkpoints)

      tabnet_model$fit$network <- reload_model(tabnet_model$serialized_net)
      apply_checkpoint(tabnet_model$fit$network, tabnet_model$fit$checkpoints[[last_checkpoint]])
      epoch_shift <- last_checkpoint * tabnet_model$fit$config$checkpoint_epoch

    } else runtime_error("No model serialized weight can be found in {.var {tabnet_model}}, check the model history")

    fit_lst <- tabnet_train_supervised(tabnet_model, predictors, outcomes, config = config, epoch_shift)
    return(new_tabnet_fit(fit_lst, blueprint = processed$blueprint))

  } else if (task == "unsupervised") {

    if (!is.null(tabnet_model)) {
      warn("Using {.fn tabnet_pretrain} from a model is not currently supported.",
           "Pretraining will start from a new network initialization")
    }
    pretrain_lst <- tabnet_train_unsupervised( predictors, config = config, epoch_shift)
    return(new_tabnet_pretrain(pretrain_lst, blueprint = processed$blueprint))

  }
}


#' Predict using `tabnet`
#'
#' @param object,x A `tabnet_fit` object.
#'
#' @param new_data A data frame or matrix of new predictors.
#' @param type expected outcome type within  `c("numeric", "prob", "class")`.
#' @param epoch the epoch of an existing checkpoint to infer from.
#' 
#' @param ... Not used, but required for extensibility.
#'
#' @return
#'
#' [predict()] returns a tibble of predictions and [augment()] appends the
#' columns in `new_data`. In either case, the number of rows in the tibble is
#' guaranteed to be the same as the number of rows in `new_data`.
#'
#' For regression data, the prediction is in the column `.pred`. For
#' classification, the class predictions are in `.pred_class` and the
#' probability estimates are in columns with the pattern `.pred_{level}` where
#' `level` is the levels of the outcome factor vector.
#'
#' @examples
#' # Minimal example for quick execution
#' car_split <- rsample::initial_split(mtcars[ 1:6,   ])
#'
#' \dontrun{
#' # Fit
#' if (torch_is_installed() & interactive()) {
#'  mod <- tabnet_fit(mpg ~ cyl + log(drat), training(car_split))
#'
#'  # Predict
#'  predict(mod, testing(car_split))
#'  augment(mod, testing(car_split))
#' }
#' }
#'
#' @importFrom stats predict
#' @export
predict.tabnet_fit <- function(object, new_data, type = NULL, ..., epoch = NULL) {
  if (inherits(new_data, "Node") && !is.null(object$fit$config$ancestor)) {
    new_data_df <- node_to_df(new_data)$x
    # Enforces column order, type, column names, etc
    processed <- hardhat::forge(new_data_df, object$blueprint)
    
  } else {
    new_data_df <- new_data
    processed <- hardhat::forge(new_data, object$blueprint)
  }
  batch_size <- object$fit$config$batch_size
  out <- predict_tabnet_bridge(type, object, processed$predictors, epoch, batch_size)
  hardhat::validate_prediction_size(out, new_data_df)
  out
}

#' @export
#' @inheritParams predict.tabnet_fit
#' @rdname predict.tabnet_fit
augment.tabnet_fit <- function(x, new_data, ...) {
  res <- predict(x, new_data, ...)
  if (inherits(new_data, "Node") && !is.null(x$fit$config$ancestor)) {
    new_data_df <- node_to_df(new_data)
    # Enforces column order, type, outcomes column names, etc
    forged_truth <- hardhat::forge(cbind(new_data_df$x, new_data_df$y), x$blueprint, outcomes = TRUE)$outcomes
  } else {
    # mold XY blueprint
    # When mold() was called with a vector y, hardhat uses ".outcome" as the outcome column
    # name. forge() with outcomes = TRUE then requires new_data to contain ".outcome", which
    # won't be the case when the user passes a regular data frame.
    if (inherits(x$blueprint, "xy_blueprint") && ncol(x$blueprint$ptypes$outcomes) == 1) {
      outcome_name_col <- which(!names(new_data) %in% names(x$blueprint$ptypes$predictors))
      names(new_data)[outcome_name_col] <- ".outcome"
    } 
    forged_truth <- hardhat::forge(new_data, blueprint = x$blueprint, outcomes = TRUE)$outcomes
  }
  dplyr::bind_cols(res, forged_truth)
}


predict_tabnet_bridge <- function(type, object, predictors, epoch, batch_size) {

  type <- check_type(object$blueprint$ptypes$outcomes, type)
  is_multi_outcome <- ncol(object$blueprint$ptypes$outcomes) > 1
  outcome_nlevels <- NULL
  if (is_multi_outcome && type != "numeric") {
    outcome_nlevels <- purrr::map_dbl(object$blueprint$ptypes$outcomes, ~nlevels(.x))
  }

  if (!is.null(epoch)) {

    if (epoch > (length(object$fit$checkpoints) * object$fit$config$checkpoint_epoch))
      value_error("The model was trained for less than {.val {epoch}} epochs")

    # find closest checkpoint for that epoch
    ind <- epoch %/% object$fit$config$checkpoint_epoch

    # Ensure network is live before applying checkpoint state dict
    if (check_net_is_empty_ptr(object)) {
      m <- reload_model(object$serialized_net)
      object$fit$network$load_state_dict(m$state_dict())
    }
    apply_checkpoint(object$fit$network, object$fit$checkpoints[[ind]])

  } else if (check_net_is_empty_ptr(object)) {
    m <- reload_model(object$serialized_net)
    # this modifies 'object' in-place so subsequent predicts won't
    # need to reload.
    object$fit$network$load_state_dict(m$state_dict())
  }

  
  type_multioutcome <- paste0(type, "_", is_multi_outcome)
  switch(
    type_multioutcome,
    numeric_FALSE = predict_impl_numeric(object, predictors, batch_size),
    numeric_TRUE  = predict_impl_numeric_multiple(object, predictors, batch_size),
    prob_FALSE    = predict_impl_prob(object, predictors, batch_size),
    prob_TRUE     = predict_impl_prob_multiple(object, predictors, batch_size, outcome_nlevels),
    class_FALSE   = predict_impl_class(object, predictors, batch_size),
    class_TRUE    = predict_impl_class_multiple(object, predictors, batch_size, outcome_nlevels)
  )
}


model_pretrain_to_fit <- function(obj, x, y, config = tabnet_config()) {

  tabnet_model_lst <- tabnet_initialize(x, y, config)


  # do not restore previous metrics as loss function return non comparable
  # values, nor checkpoints
  m <- reload_model(obj$serialized_net)

  if (m$input_dim != tabnet_model_lst$network$input_dim)
    runtime_error("Model dimensions don't match.")

  # perform update of selected weights into new tabnet_model
  m_stat_dict <- m$state_dict()
  tabnet_state_dict <- tabnet_model_lst$network$state_dict()
  for (param in names(m_stat_dict)) {
    if (grepl("^encoder", param)) {
      # Convert encoder's layers name to match
      new_param <- paste0("tabnet.", param)
    } else {
      new_param <- param
    }
    if (!is.null(tabnet_state_dict[new_param])) {
      tabnet_state_dict[[new_param]] <- m_stat_dict[[param]]
    }
  }
  tabnet_model_lst$network$load_state_dict(tabnet_state_dict)
  tabnet_model_lst
}

#' Check consistency between modeling-task type and class of outcomes vars.
#'
#' infer default modeling-task type from the outcome vars class if needed.
#'
#' @param outcome_ptype shall be `model$blueprint$ptypes$outcomes` when called from
#'  a model object, or `processed$outcomes` from the result of a `mold()`
#' @param type expected outcome type within  `c("numeric", "prob", "class")`
#'
#' @return valid type within `c("numeric", "prob", "class")` for respectively regression,
#' class probabilities, or classification
#' @noRd
check_type <- function(outcome_ptype, type = NULL) {

  # outcome_ptype <- model$blueprint$ptypes$outcomes when called from model
  outcome_all_factor <- all(purrr::map_lgl(outcome_ptype, is.factor))
  outcome_all_numeric <- all(purrr::map_lgl(outcome_ptype, is.numeric))

  if (!outcome_all_numeric && !outcome_all_factor)
    not_implemented_error("Mixed multi-outcome type {.type {unique(purrr::map_chr(outcome_ptype, ~class(.x)[[1]]))}} is not supported")

  if (is.null(type)) {
    if (outcome_all_factor)
      type <- "class"
    else if (outcome_all_numeric)
      type <- "numeric"
    else if (ncol(outcome_ptype) == 1)
      type_error("Unknown outcome type {.type {class(outcome_ptype)}}")
  }

  type <- rlang::arg_match(type, c("numeric", "prob", "class"))

  if (outcome_all_factor) {
    if (!type %in% c("prob", "class"))
      type_error("Outcome is factor and the prediction type is {.type {type}}.")
  } else if (outcome_all_numeric) {
    if (type != "numeric")
      type_error("Outcome is numeric and the prediction type is {.type {type}}.")
  }

  invisible(type)
}


reload_model <- function(object) {
  con <- rawConnection(object)
  on.exit({close(con)}, add = TRUE)
  module <- torch::torch_load(con)
  module
}

# Apply a checkpoint (raw bytes) to an existing live network.
# Handles both old format (full nn_module) and new format (CPU state dict list).
apply_checkpoint <- function(network, raw_bytes) {
  con <- rawConnection(raw_bytes)
  on.exit(close(con), add = TRUE)
  loaded <- torch::torch_load(con)
  if (inherits(loaded, "nn_module")) {
    network$load_state_dict(loaded$state_dict())
  } else {
    network$load_state_dict(loaded)
  }
  invisible(network)
}

#' @export
print.tabnet_fit <- function(x, ...) {
  if (check_net_is_empty_ptr(x)) {
    print(reload_model(x$serialized_net))
  } else {
    print(x$fit$network)
  }
  invisible(x)
}
#' @export
print.tabnet_pretrain <- print.tabnet_fit

#' Prune top layer(s) of a tabnet network
#'
#' Prune `head_size` last layers of a tabnet network in order to
#'  use the pruned module as a sequential embedding module.
#' @param x nn_network to prune
#' @param head_size number of nn_layers to prune, should be less than 2
#'
#' @return a tabnet network with the top nn_layer removed
#' @rdname nn_prune_head
#' @examplesIf (torch::torch_is_installed())
#' data("ames", package = "modeldata")
#' x <- ames[,-which(names(ames) == "Sale_Price")]
#' y <- ames$Sale_Price
#' # pretrain a tabnet model on ames dataset
#' ames_pretrain <- tabnet_pretrain(x, y, epoch = 2, checkpoint_epochs = 1)
#' # prune classification head to get an embedding model
#' pruned_pretrain <- torch::nn_prune_head(ames_pretrain, 1)
#
#' @importFrom torch nn_prune_head
#' @export
nn_prune_head.tabnet_fit <- function(x, head_size) {
  if (check_net_is_empty_ptr(x)) {
    net <- reload_model(x$serialized_net)
  } else {
    net <- x$fit$network
  }
  # here we assemble nn_prune_head(x, 1) with nn_prune_head(x$tabnet, 1)
  x <- nn_prune_head(net, 1)
  x$add_module(name= "tabnet", module=nn_prune_head(net$tabnet,head_size=head_size))

}
#' @importFrom torch nn_prune_head
#' @rdname nn_prune_head
#' @export
nn_prune_head.tabnet_pretrain <- function(x, head_size) {
  if (check_net_is_empty_ptr(x)) {
    nn_prune_head(reload_model(x$serialized_net), head_size=head_size)
  } else {
    nn_prune_head(x$fit$network, head_size=head_size)
  }

}


#' Build ancestor matrix aligned with observed outcome classes
#'
#' Extracts class names from the outcome tibble (factor levels) and builds
#' the ancestor matrix only for classes that actually appear in the data.
#'
#' @param x A `data.tree::Node` object.
#' @param outcomes A tibble with factor columns (one per hierarchy level),
#'   as returned by `hardhat::mold()$outcomes`.
#' @param device Torch device ("cpu" or "cuda").
#' @return A `torch_tensor` of shape `(1, n_classes, n_classes)`.
#' @export
build_ancestor_matrix_from_outcomes <- function(x, outcomes, device = "cpu") {
  # 1. Extract all class names from factor levels (preserving order)
  #    outcomes is a tibble with one factor column per hierarchy level
  level_cols <- names(outcomes)
  all_class_names <- unlist(lapply(outcomes, levels), use.names = FALSE)
  n_classes <- length(all_class_names)
  
  if (n_classes == 0L) {
    runtime_error("No factor levels found in outcomes : {str(outcomes)}")
  }
  
  # 2. Build a lookup: class_name -> data.tree Node
  all_nodes <- data.tree::Traverse(x, traversal = "pre-order")
  all_nodes <- unname(all_nodes)
  level_lengths <- lengths(lapply(outcomes, levels))
  lvl_vector <- rep(seq_along(level_cols) + 1L, level_lengths)
  
  # 3. Resolve each class name to its Node
  class_nodes <- lapply(seq_along(all_class_names), function(k) {
    nm <- all_class_names[k]
    lvl <- lvl_vector[k]
    
    candidates <- Filter(function(n) n$level == lvl && n$name == nm, all_nodes)
    if (length(candidates) == 0) {
      runtime_error("Factor level {.var {nm}} not found at tree level {lvl} (outcomes column {.var {level_cols[lvl - 1L]}})")
    }
    candidates[[1]]
  })
  
  # 4. Create 1-based index mapping
  class_map <- setNames(seq_len(n_classes), all_class_names)
  
  # 5. Collect (descendant, ancestor) pairs by climbing up
  row_list <- vector("list", n_classes)
  col_list <- vector("list", n_classes)
  
  for (i in seq_len(n_classes)) {
    current <- class_nodes[[i]]
    anc_indices <- integer()
    
    repeat {
      idx <- class_map[current$name]
      if (!is.null(idx)) {
        anc_indices <- c(anc_indices, idx)
      }
      if (current$isRoot || is.null(current$parent)) break
      current <- current$parent
    }
    
    row_list[[i]] <- rep(i, length(anc_indices))
    col_list[[i]] <- anc_indices
  }
  
  # 6. Fill matrix
  R <- matrix(0L, nrow = n_classes, ncol = n_classes)
  rows <- unlist(row_list, use.names = FALSE)
  cols <- unlist(col_list, use.names = FALSE)
  if (length(rows) > 0) R[cbind(rows, cols)] <- 1L
  
  # 7. Convert to torch
  R_torch <- torch::torch_tensor(R, device = device)
  R_torch$unsqueeze(1)
}

test_that("returns correct shape and type for a simple 2-level hierarchy", {
  # Arbre : Root -> {A, B}, A -> {C1, C2}, B -> {D1, D2}
  tree_df <- data.frame(pathString = c(
    "Root/A/C1", "Root/A/C2", "Root/B/D1", "Root/B/D2"
  ))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(c("A", "A", "B", "B")),
    level_3 = factor(c("C1", "C2", "D1", "D2"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)
  
  # 2 level_2 classes + 4 level_3 = 6 classes
  expect_tensor(result)
  expect_tensor_shape(result, c(1, 6, 6))
  })

test_that("handles ghost classes (present in tree but absent from outcomes)", {
  # Tree with a "C" branch not in the outcomes
  tree_df <- data.frame(pathString = c(
    "Root/A/C1", "Root/A/C2", 
    "Root/B/D1", "Root/B/D2", 
    "Root/C/E1", "Root/C/E2"
  ))
  tree <- as.Node(tree_df)
  
  # Outcomes shall only contain A and B (C is a "ghost class")
  outcomes <- tibble::tibble(
    level_2 = factor(c("A", "A", "B", "B")), 
    level_3 = factor(c("C1", "C2", "D1", "D2"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)
  expect_tensor_shape(result, c(1, 6, 6))
  
  # Check of transitive loop : A(1) is ancestor of C1(3) and C2(4)
  # Order : A(1), B(2), C1(3), C2(4), D1(5), D2(6)
  expect_equal_to_r(result[1, 3, 1], 1) 
  expect_equal_to_r(result[1, 4, 1], 1)
  expect_equal_to_r(result[1, 5, 2], 1) 
  expect_equal_to_r(result[1, 6, 2], 1)
  expect_equal_to_r(result[1, 3, 3], 1) # Self-loop
})

test_that("handles non-unique names across different hierarchy levels", {
  # "Manager" may exist at level_2 (Department) and level_3 (JobRole)
  tree_df <- data.frame(pathString = c(
    "Root/Manager/Rep", 
    "Root/IT/Manager"
  ))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(c("Manager", "IT")),
    level_3 = factor(c("Rep", "Manager"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)
  
  # 2 + 2 = 4 classes
  expect_tensor_shape(result, c(1L, 4L, 4L))

  # Order : Manager_lvl2(1), IT(2), Rep(3), Manager_lvl3(4)
  # Manager_lvl2(1) is ancestor of Rep(3)
  expect_equal_to_r(result[1, 3, 1], 1)
  # IT(2) is ancestor of Manager_lvl3(4)
  expect_equal_to_r(result[1, 4, 2], 1)
})

test_that("throws an explicit error when a factor level is missing from the tree", {
  tree_df <- data.frame(pathString = c("Root/A/C1", "Root/A/C2"))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(c("A", "X")), # "X" is not in the tree
    level_3 = factor(c("C1", "C2"))
  )
  
  expect_error(
    build_ancestor_matrix_from_outcomes(tree, outcomes),
    "not found"
  )
})

test_that("throws an error when outcomes contains no factor levels", {
  tree_df <- data.frame(pathString = c("Root/A/C1"))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(character(0)),
    level_3 = factor(character(0))
  )
  
  expect_error(
    build_ancestor_matrix_from_outcomes(tree, outcomes),
    "No factor levels"
  )
})

test_that("preserves the exact class order defined in outcomes factors", {
  # Dans l'arbre, B est défini avant A
  tree_df <- data.frame(pathString = c(
    "Root/B/D1", "Root/B/D2", 
    "Root/A/C1", "Root/A/C2"
  ))
  tree <- as.Node(tree_df)
  
  # Mais dans outcomes, A est explicitement avant B
  outcomes <- tibble::tibble(
    level_2 = factor(c("A", "A", "B", "B"), levels = c("A", "B")),
    level_3 = factor(c("C1", "C2", "D1", "D2"), levels = c("C1", "C2", "D1", "D2"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)

  # Order is given by outcomes : A(1), B(2), C1(3), C2(4), D1(5), D2(6)
  # A(1) is ancestor of C1(3) et C2(4)
  expect_equal_to_r(result[1, 3, 1], 1)
  expect_equal_to_r(result[1, 4, 1], 1)
  
  # B(2) is ancestor of D1(5) et D2(6)
  expect_equal_to_r(result[1, 5, 2], 1)
  expect_equal_to_r(result[1, 6, 2], 1)
})

test_that("handles a flat single-level hierarchy correctly", {
  tree_df <- data.frame(pathString = c("Root/A", "Root/B", "Root/C"))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(c("A", "B", "C"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)
  expect_equal(result$shape, c(1L, 3L, 3L))
  
  # no hierarchy, only self-loops
  expect_equal_to_r(result$squeeze(1), diag(3))
})

test_that("computes full transitive closure for deep hierarchies (3+ levels)", {
  tree_df <- data.frame(pathString = c(
    "Root/L1_A/L2_A1/L3_A1a", 
    "Root/L1_A/L2_A1/L3_A1b", 
    "Root/L1_A/L2_A2/L3_A2a", 
    "Root/L1_B/L2_B1/L3_B1a"
  ))
  tree <- as.Node(tree_df)
  
  outcomes <- tibble::tibble(
    level_2 = factor(c("L1_A", "L1_A", "L1_A", "L1_B")),
    level_3 = factor(c("L2_A1","L2_A1", "L2_A2", "L2_B1")),
    level_4 = factor(c("L3_A1a", "L3_A1b", "L3_A2a", "L3_B1a"))
  )
  
  result <- build_ancestor_matrix_from_outcomes(tree, outcomes)
  expect_equal(result$shape, c(1L, 9L, 9L))
  
  # Order : L1_A(1), L1_B(2), L2_A1(3), L2_A2(4), L2_B1(5), 
  #         L3_A1a(6), L3_A1b(7), L3_A2a(8), L3_B1a(9)
  
  # L1_A(1) is a transitive ancestor of all the sub-tree A
  expect_equal_to_r(result[1, 3, 1], 1) # -> L2_A1
  expect_equal_to_r(result[1, 6, 1], 1) # -> L3_A1a
  expect_equal_to_r(result[1, 8, 1], 1) # -> L3_A2a
  
  # L2_A1(3) is an ancestor of all its direct children
  expect_equal_to_r(result[1, 6, 3], 1) # -> L3_A1a
  expect_equal_to_r(result[1, 7, 3], 1) # -> L3_A1b
  
  # Self-loops on the diagonal (substracting the eye don't go to negative values)
  expect_true((result$squeeze() - torch::torch_eye(9))$min()$item() >= 0)
})

test_that("pretrain accepts missing value in predictors and (unused) outcome", {

  data("attrition", package = "modeldata")
  ids <- sample(nrow(attrition), 256)

  x <- attrition[ids,-which(names(attrition) == "Attrition")]
  y <- attrition[ids,]$Attrition
  y_missing <- y
  y_missing[1] <- NA

  # numerical missing
  x_missing <- x
  x_missing[1,"Age"] <- NA

  expect_no_error(
    miss_pretrain <- tabnet_pretrain(x_missing, y, epochs = 1)
  )

  # categorical missing
  x_missing <- x
  x_missing[1,"BusinessTravel"] <- NA

  expect_no_error(
    miss_pretrain <- tabnet_pretrain(x_missing, y, epochs = 1)
  )

  # no error when missing in outcome
  expect_no_error(
    miss_pretrain <- tabnet_pretrain(x, y_missing, epochs = 1)
  )

})


test_that("fit accept missing value in predictor, not in outcome", {

  data("attrition", package = "modeldata")
  ids <- sample(nrow(attrition), 256)

  x <- attrition[ids,-which(names(attrition) == "Attrition")]
  y <- attrition[ids,]$Attrition
  y_missing <- y
  y_missing[1] <- NA

  # numerical missing
  x_missing <- x
  x_missing[1,"Age"] <- NA

  expect_no_error(
    miss_fit <- tabnet_fit(x_missing, y, epochs = 1)
  )

  # categorical missing
  x_missing <- x
  x_missing[1,"BusinessTravel"] <- NA

  expect_no_error(
    miss_fit <- tabnet_fit(x_missing, y, epochs = 1)
  )

  # missing in outcome
  expect_error(
    miss_fit <- tabnet_fit(x, y_missing, epochs = 1),
    regexp = "missing"
  )

})

test_that("fit accept missing value in `Node` predictor", {
  # fix to https://github.com/mlverse/tabnet/issues/125
  library(data.tree)
  data(starwars, package = "dplyr")

  starwars_tree <- starwars %>%
    rename(`_name` = "name", `_height` = "height") %>%
    # iconv translation ensure compatibility on macintosh
    mutate(pathString = paste("StarWars_characters", species, homeworld, `_name`, sep = "/") %>%
             iconv(from = "UTF-8", to = "ASCII//TRANSLIT")) %>%
    as.Node()

  expect_no_error(
    miss_fit <- tabnet_fit(starwars_tree, epochs = 1, cat_emb_dim = 2)
  )

  expect_no_error(
    miss_pred <- predict(miss_fit, starwars_tree)
  )

})

test_that("predict data-frame accept missing value in predictor", {

  data("attrition", package = "modeldata")
  ids <- sample(nrow(attrition), 256)

  x <- attrition[ids,-which(names(attrition) == "Attrition")]
  y <- attrition[ids,]$Attrition
  #
  fit <- tabnet_fit(x, y, epochs = 1)

  # numerical missing
  x_missing <- x
  x_missing[1,"Age"] <- NA

  # predict with numerical missing
  expect_no_error(
    predict(fit, x_missing),
  )
  # categorical missing
  x_missing <- x
  x_missing[1,"BusinessTravel"] <- NA

  # predict with categorical missing
  expect_no_error(
    predict(fit, x_missing)
  )

})

test_that("inference works with missings in the response vector", {

  data("attrition", package = "modeldata")
  ids <- sample(nrow(attrition), 256)

  rec <- recipe(EnvironmentSatisfaction ~ ., data = attrition[ids, ]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  fit <- tabnet_fit(rec, attrition, epochs = 1, valid_split = 0.25,
                    verbose = TRUE)
  # predict with empty vector
  attrition[["EnvironmentSatisfaction"]] <-NA
  expect_no_error(
    predict(fit, attrition)
  )

  # predict with wrong class
  attrition[["EnvironmentSatisfaction"]] <-NA_character_
  expect_no_error(
    predict(fit, attrition)
  )

  # predict with list column
  attrition[["EnvironmentSatisfaction"]] <- list(NA)
  expect_no_error(
    predict(fit, attrition)
  )

})

test_that("explain works with missings in predictors", {

  data("attrition", package = "modeldata")
  ids <- sample(nrow(attrition), 256)

  x <- attrition[ids,-which(names(attrition) == "Attrition")]
  y <- attrition[ids,]$Attrition
  #
  fit <- tabnet_fit(x, y, epochs = 1)

  # numerical missing
  x_missing <- x
  x_missing[1,"Age"] <- NA

  # explain with numerical missing
  expect_no_error(
    tabnet_explain(fit, x_missing)
  )
  # categorical missing
  x_missing <- x
  x_missing[1,"BusinessTravel"] <- NA

  # explain with categorical missing
  expect_no_error(
    tabnet_explain(fit, x_missing)
  )
})

test_that("transpose_metrics is not adding an unnamed entry on top of the list", {

  metrics <- list(loss = 1, loss = 2, loss = 3, loss = 4)

  expect_no_error(
    tabnet:::transpose_metrics(metrics)
  )

  expect_equal(
    tabnet:::transpose_metrics(metrics),
    list(loss = c(1, 2, 3, 4))
  )

})

test_that("Unsupervised training with default config, data.frame and formula", {

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 1)
  )
  expect_s3_class( fit, "tabnet_pretrain")
  expect_equal(length(fit), 3)
  expect_equal(names(fit), c("fit", "serialized_net", "blueprint"))
  expect_equal(length(fit$fit), 5)
  expect_equal(names(fit$fit), c("network", "metrics", "config", "checkpoints", "importances"))
  expect_equal(length(fit$fit$metrics), 1)

  expect_no_error(
    fit <- tabnet_pretrain(Sale_Price ~ ., data = ames, epochs = 1)
  )
  expect_s3_class( fit, "tabnet_pretrain")
  expect_equal(length(fit), 3)
  expect_equal(names(fit), c("fit", "serialized_net", "blueprint"))
  expect_equal(length(fit$fit), 5)
  expect_equal(names(fit$fit), c("network", "metrics", "config", "checkpoints", "importances"))
  expect_equal(length(fit$fit$metrics), 1)

})

test_that("Unsupervised training with pretraining_ratio", {

  expect_no_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, pretraining_ratio=0.2)
  )

})

test_that("Unsupervised training prevent predict with an explicit message", {

  pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, pretraining_ratio=0.2)

  expect_error(
    predict(pretrain, attrix, type = "prob"),
    regexp = "tabnet_pretrain"
  )

  expect_error(
    predict(pretrain, attrix),
    regexp = "tabnet_pretrain"
  )

})

test_that("pretraining with `tabnet_model= ` parameter raise a warning", {

  expect_warning(
    fit <- tabnet_pretrain(x, y, epochs = 1, tabnet_model = ames_pretrain)
  )
  expect_s3_class( fit, "tabnet_pretrain")
  expect_equal( length(fit), length(ames_pretrain))
  expect_equal( length(fit$fit$metrics), 1)
})

test_that("errors when using an argument that do not exist", {

  expect_error(
    pretrain <- tabnet_pretrain(x, y, pretraining_ratiosas = 1-1e5),
    regexp = "unused argument"
  )

})

test_that("works with validation split", {

  expect_no_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, valid_split = 0.2)
  )

  expect_no_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, valid_split = 0.2, verbose = TRUE)
  )

})

test_that("works with categorical embedding dimension as list", {

  config <- tabnet_config(cat_emb_dim=c(1,1,2,2,1,1,1,2,1,1,1,2,2,2))

  expect_no_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, valid_split = 0.2, config=config)
  )
})

test_that("explicit error message when categorical embedding dimension vector has wrong size", {

  config <- tabnet_config(cat_emb_dim=c(1,1,2,2))

  expect_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1, valid_split = 0.2, config=config),
    regexp = "number of categorical predictors"
  )
})

test_that("can train from a recipe", {

  rec <- recipe(Attrition ~ ., data = attrition) %>%
    step_normalize(all_numeric(), -all_outcomes())

  expect_no_error(
    pretrain <- tabnet_pretrain(rec, attrition, epochs = 1, verbose = TRUE)
  )

})

test_that("lr scheduler step works", {

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 3, lr_scheduler = "step",
                           lr_decay = 0.1, step_size = 1)
  )

  sc_fn <- function(optimizer) {
    torch::lr_step(optimizer, step_size = 1, gamma = 0.1)
  }

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 3, lr_scheduler = sc_fn,
                           lr_decay = 0.1, step_size = 1)
  )

})

test_that("lr scheduler reduce_on_plateau works", {

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 3, lr_scheduler = "reduce_on_plateau",
                           lr_decay = 0.1, step_size = 1)
  )

  sc_fn <- function(optimizer) {
    torch::lr_reduce_on_plateau(optimizer, factor = 0.1, patience = 10)
  }

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 3, lr_scheduler = sc_fn,
                           lr_decay = 0.1, step_size = 1)
  )

})

test_that("checkpoints works", {

  expect_no_error(
    pretrain <- tabnet_pretrain(x, y, epochs = 3, checkpoint_epochs = 1)
  )

  expect_length( pretrain$fit$checkpoints, 3  )

  # expect_equal(  pretrain$fit$checkpoints[[3]], pretrain$serialized_net )

})

test_that("print module works", {

  testthat::local_edition(3)
  testthat::skip_on_os("linux")
  testthat::skip_on_os("windows")

  expect_no_error(
    fit <- tabnet_pretrain(x, y, epochs = 1)
  )

  withr::with_options(new = c(cli.width = 50),
                      expect_snapshot_output(fit))

})

test_that("num_independent_decoder and num_shared_decoder change the network number of parameters", {

  expect_no_error(
    pretrain <- tabnet_pretrain(attrix, attriy, epochs = 1,
                                num_independent_decoder = 3, num_shared_decoder = 2)
  )
  expect_gt( torch:::get_parameter_count(pretrain$fit$network),
             torch:::get_parameter_count(attr_pretrained$fit$network)
  )
})

test_that("num_independent_decoder and num_shared_decoder do not change the network number of parameters for fit", {

  expect_no_error(
    config <- tabnet_config(epochs = 1,
                            num_independent_decoder = 3, num_shared_decoder = 2)
  )
  expect_no_error(
    attr_fit <- tabnet_fit(attrix, attriy, config = config)
  )
  expect_equal( torch:::get_parameter_count(attr_fit$fit$network),
                torch:::get_parameter_count(attr_fitted$fit$network)
  )
})

test_that("explain provides correct result with data.frame", {

  set.seed(2022)
  torch::torch_manual_seed(2022)

  n <- 2000
  x <- data.frame(
    x = rnorm(n),
    y = rnorm(n),
    z = rnorm(n)
  )

  y <- x$x

  fit <- tabnet_fit(x, y, epochs = 15,
                    num_steps = 1,
                    batch_size = 512,
                    attention_width = 1,
                    num_shared = 1,
                    num_independent = 1)

  expect_equal(which.max(fit$fit$importances$importance), 1)
  expect_equal(fit$fit$importances$variables, colnames(x))

  ex <- tabnet_explain(fit, x)

  expect_length(ex, 2)
  expect_length(ex[[2]], 1)
  expect_equal(nrow(ex[[1]]), nrow(x))
  expect_equal(nrow(ex[[2]][[1]]), nrow(x))

})

test_that("explain works for dataframe, formula and recipe", {

  # data.frame, regression
  expect_no_error(
    tabnet_explain(ames_pretrain_vsplit, new_data=small_ames)
  )

  expect_no_error(
    tabnet_explain(ames_fit_vsplit, new_data=small_ames)
  )

  # data.frame, classification
  expect_no_error(
    tabnet_explain(attr_pretrained_vsplit, attrix)
  )
  expect_no_error(
    tabnet_explain(attr_fitted_vsplit, attrix)
  )


  # formula
  tabnet_pretrain <- tabnet_pretrain(Sale_Price ~., data=small_ames, epochs = 3, valid_split=0.2,
                                     num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)
  expect_no_error(
    tabnet_explain(tabnet_pretrain, new_data=small_ames)
  )

  tabnet_fit <- tabnet_fit(Sale_Price ~., data=small_ames, tabnet_model=tabnet_pretrain, epochs = 3,
                           num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)
  expect_no_error(
    tabnet_explain(tabnet_fit, new_data=small_ames)
  )

  # recipe
  rec <- recipe(Sale_Price ~., data = small_ames) %>%
    step_zv(all_predictors()) %>%
    step_normalize(all_numeric_predictors())

  tabnet_pretrain <- tabnet_pretrain(rec, data=small_ames, epochs = 3, valid_split=0.2,
                                     num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)
  expect_no_error(
    tabnet_explain(tabnet_pretrain, new_data=small_ames)
  )

  tabnet_fit <- tabnet_fit(rec, data=small_ames, tabnet_model=tabnet_pretrain, epochs = 3,
                           num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)
  expect_no_error(
    tabnet_explain(tabnet_fit, new_data=small_ames)
  )
})

test_that("support for vip on tabnet_fit and tabnet_pretrain", {

  skip_if_not_installed("vip")

  n <- 1000
  x <- data.frame(
    x = runif(n),
    y = runif(n),
    z = runif(n)
  )

  y <- x$x

  pretrain <- tabnet_pretrain(x, y, epochs = 1,
                    num_steps = 1,
                    batch_size = 512,
                    attention_width = 1,
                    num_shared = 1,
                    num_independent = 1)

  fit <- tabnet_fit(x, y, epochs = 1,
                    num_steps = 1,
                    batch_size = 512,
                    attention_width = 1,
                    num_shared = 1,
                    num_independent = 1)

  expect_no_error(vip::vip(pretrain))
  expect_no_error(vip::vip(fit))

})


test_that("Importance is skipped if skip_importance flag is used", {

  set.seed(2022)
  torch::torch_manual_seed(2022)

  n <- 1000
  x <- data.frame(
    x = rnorm(n),
    y = rnorm(n),
    z = rnorm(n)
  )

  y <- x$x

  fit <- tabnet_fit(x, y, epochs = 15,
                    num_steps = 1,
                    batch_size = 512,
                    attention_width = 1,
                    num_shared = 1,
                    num_independent = 1,
                    skip_importance = TRUE)

  expect_equal(fit$fit$importances, NULL)

  fit <- tabnet_fit(x, y, epochs = 15,
                    num_steps = 1,
                    batch_size = 512,
                    attention_width = 1,
                    num_shared = 1,
                    num_independent = 1,
                    skip_importance = FALSE)


  expect_equal(which.max(fit$fit$importances$importance), 1)
  expect_equal(fit$fit$importances$variables, colnames(x))

})

test_that("explain works for parsnip model", {

  model <- tabnet() %>%
    parsnip::set_mode("regression") %>%
    parsnip::set_engine("torch")
  fit <- model %>%
    parsnip::fit(Sale_Price ~ ., data = small_ames)

  expect_no_error(
    tabnet_explain(fit, new_data = small_ames),
  )

})

test_that("explain works for multi-outcome classification model", {

  fit <- tabnet_fit(x, data.frame(y = y, z = y + 1), epochs = 1)

  expect_no_error(tabnet_explain(fit, new_data = x))

})

test_that("Training regression for data.frame and formula", {

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 1)
  )

  expect_no_error(
    fit <- tabnet_fit(Sale_Price ~ ., data = ames, epochs = 1)
  )

  expect_no_error(
    predict(fit, x)
  )
  expect_no_error(
    augment(fit, ames)
  )

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 2, verbose = TRUE)
  )
})

test_that("Training classification for data.frame", {

  expect_no_error(
    fit <- tabnet_fit(attrix, attriy, epochs = 1)
  )

  expect_no_error(
    predict(fit, attrix, type = "prob")
  )

  expect_no_error(
    predict(fit, attrix)
  )

  expect_no_error(
    augment(fit, attrition[ids,] , type = "prob")
  )

  expect_no_error(
    augment(fit, attrition[ids,] )
  )

})

test_that("works with validation split", {

  expect_no_error(
    fit <- tabnet_fit(attrix, attriy, epochs = 1, valid_split = 0.5)
  )

  expect_no_error(
    fit <- tabnet_fit(attrix, attriy, epochs = 1, valid_split = 0.5, verbose = TRUE)
  )

})


test_that("can train from a recipe", {

  rec <- recipe(Attrition ~ ., data = attrition) %>%
    step_normalize(all_numeric(), -all_outcomes())

  expect_no_error(
    fit <- tabnet_fit(rec, attrition[1:256,], epochs = 1, valid_split = 0.25,
                    verbose = TRUE)
  )

  expect_no_error(
    predict(fit, attrition)
  )

})

test_that("serialization with saveRDS just works", {

  predictions <-  predict(ames_fit, ames)

  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_fit, tmp))

  # rm(fit)
  gc()

  fit2 <- readRDS(tmp)

  expect_equal(
    predictions,
    predict(fit2, ames)
  )

  expect_equal(as.numeric(fit2$fit$network$.check), 1)

})

test_that("checkpoints works for inference", {

  expect_no_error(
    fit <- tabnet_fit(x, y, epochs = 3, checkpoint_epochs = 1)
  )

  expect_no_error(
    p1 <- predict(fit, x, epoch = 1)
  )

  expect_no_error(
    p2 <- predict(fit, x, epoch = 2)
  )

  expect_no_error(
    p3 <- predict(fit, x, epoch = 3)
  )

  expect_equal(p3, predict(fit, x))

  expect_no_error(augment(fit, ames[ids,], epoch = 1))
})

test_that("print module works even after a reload from disk", {

  testthat::skip_on_os("linux")
  testthat::skip_on_os("windows")

  withr::with_options(new = c(cli.width = 50),
                      expect_snapshot_output(ames_fit))

  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_fit, tmp))
  fit2 <- readRDS(tmp)

  withr::with_options(new = c(cli.width = 50),
                      expect_snapshot_output(fit2))

})


test_that("num_workers works for pretrain, fit an predict", {

  expect_no_error(
    tabnet_pretrain(x, y, epochs = 1, num_workers=1L,
                    batch_size=128, virtual_batch_size=64)
  )
  expect_no_error(
    tabnet_pretrain(x, y, epochs = 1, num_workers=1L, valid_split=0.2,
                    batch_size=128, virtual_batch_size=64)
  )

  expect_no_error(
    tabnet_fit(x, y, epochs = 1, num_workers=1L,
               batch_size=128, virtual_batch_size=64)
  )

  expect_no_error(
    tabnet_fit(x, y, epochs = 1, num_workers=1L, valid_split=0.2,
               batch_size=128, virtual_batch_size=64)
  )

  expect_no_error(
    predict(ames_fit, x, num_workers=1L,
            batch_size=128, virtual_batch_size=64)
  )

})


test_that("we can prune head of tabnet pretrain and tabnet fit models", {

  expect_no_error(pruned_pretrain <- torch::nn_prune_head(ames_pretrain, 1))
  test_that("decoder has been removed from the list of modules", {
    expect_equal(all(stringr::str_detect("decoder", names(pruned_pretrain$children))),FALSE)
  })


  expect_no_error(pruned_fit <- torch::nn_prune_head(ames_fit, 1))
  test_that("decoder has been removed from the list of modules", {
    expect_equal(all(stringr::str_detect("final_mapping", names(pruned_pretrain$children))),FALSE)
  })


})

test_that("we can prune head of restored models from disk", {
  testthat::skip_on_os("linux")
  testthat::skip_on_os("windows")

  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_pretrain, tmp))
  ames_pretrain2 <- readRDS(tmp)
  expect_no_error(pruned_pretrain <- torch::nn_prune_head(ames_pretrain2, 1))
  test_that("decoder has been removed from the list of modules", {
    expect_equal(all(stringr::str_detect("decoder", names(pruned_pretrain$children))),FALSE)
  })


  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_fit, tmp))
  ames_fit2 <- readRDS(tmp)
  expect_no_error(pruned_fit <- torch::nn_prune_head(ames_fit2, 1))
  test_that("decoder has been removed from the list of modules", {
    expect_equal(all(stringr::str_detect("final_mapping", names(pruned_pretrain$children))),FALSE)
  })

})

test_that("Autoplot with unsupervised training, w and wo valid_split", {

  expect_no_error(
    print(autoplot(attr_pretrained))
  )

  expect_no_error(
    print(autoplot(attr_pretrained_vsplit))
  )

})

test_that("Autoplot with supervised training, w and wo valid_split", {

  expect_no_error(
    print(autoplot(attr_fitted))
  )

  expect_no_error(
    print(autoplot(attr_fitted_vsplit))
  )

})

test_that("Autoplot a model without checkpoint", {

  tabnet_pretrain <- tabnet_pretrain(attrix, attriy, epochs = 3)
  expect_no_error(
    print(autoplot(tabnet_pretrain))
  )

  tabnet_pretrain <- tabnet_pretrain(attrix, attriy, epochs = 3, valid_split=0.3)
  expect_no_error(
    print(autoplot(tabnet_pretrain))
  )

  tabnet_fit <- tabnet_fit(attrix, attriy, epochs = 3)
  expect_no_error(
    print(autoplot(tabnet_fit))
  )

  tabnet_fit <- tabnet_fit(attrix, attriy, epochs = 3, valid_split = 0.3)
  expect_no_error(
    print(autoplot(tabnet_fit))
  )

})

test_that("Autoplot of pretrain then fit scenario, pretrain without checkpoints, fit without valid", {

  tabnet_fit <- tabnet_fit(attrix, attriy, tabnet_model = attr_pretrained_vsplit, epochs = 12)

  expect_no_error(
    print(autoplot(tabnet_fit))
  )

  fit_no_checkpoint <- tabnet_fit(Sale_Price ~., data = small_ames, epochs = 2, valid_split = 0.2, checkpoint_epoch = 3, batch_size = 64)
  expect_no_error(
    print(autoplot(fit_no_checkpoint))
  )
  fit_with_checkpoint <- tabnet_fit(Sale_Price ~., data = small_ames, tabnet_model = fit_no_checkpoint, epochs = 2, checkpoint_epoch = 1)
  expect_warning(
    print(autoplot(fit_with_checkpoint)),
    "Removed 2 rows containing missing values"
  )

})

test_that("Autoplot of tabnet_explain works for pretrain and fitted model", {

  explain_pretrain <- tabnet_explain(attr_pretrained_vsplit, attrix)
  explain_fit <- tabnet_explain(attr_fitted_vsplit, attrix)

  expect_no_error(
    print(autoplot(explain_pretrain))
  )

  expect_no_error(
    print(autoplot(explain_pretrain, type = "steps"))
  )

  expect_no_error(
    print(autoplot(explain_pretrain, type = "steps", quantile = 0.99)),

  )

  expect_no_error(
    print(autoplot(explain_fit))
  )

  expect_no_error(
    print(autoplot(explain_fit, type = "steps"))
  )

  expect_no_error(
    print(autoplot(explain_fit, type = "steps", quantile = 0.99))
  )

})

test_that("Autoplot of multi-outcome regression explainer", {

  x <- small_ames[,-which(names(ames) %in% c("Sale_Price", "Pool_Area"))]
  y <- small_ames[, c("Sale_Price", "Pool_Area")]
  ames_fit <- tabnet_fit(x, y, epochs = 5, verbose=TRUE)
  ames_explain <- tabnet_explain(ames_fit, ames)

  expect_no_error(
    print(autoplot(ames_explain))
  )

  expect_no_error(
    print(autoplot(ames_explain, type = "steps", quantile = 0.99))
  )

})

test_that("Training multi-output regression from data.frame", {

  expect_no_error(
    fit <- tabnet_fit(x, data.frame(y = y, z = y + 1), epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, x)
  )
  expect_equal(ncol(result), 2)

  expect_no_error(
    augmented <- augment(fit, cbind(x, data.frame(y = y, z = y + 1)))
  )
  expect_equal(ncol(augmented), 4)
  
})

test_that("Training multi-output regression from formula", {

  expect_no_error(
    fit <- tabnet_fit(Sale_Price + Latitude + Longitude ~ ., small_ames, epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, ames)
  )
  expect_equal(ncol(result), 3)

  expect_no_error(
    augmented <- augment(fit, ames)
  )
  expect_equal(ncol(augmented), 6)

})

test_that("Training multi-output regression from recipe", {

  rec <- recipe(Sale_Price + Latitude + Longitude ~ ., data = small_ames) %>%
    step_zv(all_predictors()) %>%
    step_normalize(all_numeric(), -all_outcomes())

  expect_no_error(
    fit <- tabnet_fit(rec, small_ames, epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, ames)
  )
  expect_equal(ncol(result), 3)

  expect_no_error(
    augmented <- augment(fit, ames)
  )
  expect_equal(ncol(augmented), 6)
  
})

test_that("Training multilabel classification from data.frame", {

  expect_no_error(
    fit <- tabnet_fit(attri_mult_x, data.frame(y = attriy, z = attriy, sat = attrix$JobSatisfaction),
                      epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, attri_mult_x, type = "prob")
  )

  expect_equal(ncol(result), 3)
  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  # we get back outcomes vars with a `.pred_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_"), names(outcome_nlevels))

  # result columns are tibbles of resp 2, 2, 4 columns
  expect_true(all(purrr::map_lgl(result, tibble::is_tibble)))
  expect_equal(purrr::map_dbl(result, ncol), outcome_nlevels, ignore_attr = TRUE)

  expect_no_error(
    result <- predict(fit, attri_mult_x)
  )
  expect_equal(ncol(result), 3)
  expect_equal(stringr::str_remove(names(result), ".pred_class_"), names(outcome_nlevels))

})

test_that("Training multilabel classification from formula", {

  expect_no_error(
    fit <- tabnet_fit(Attrition + JobSatisfaction ~ ., data = attrition[ids,],
                      epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, attri_mult_x, type = "prob")
  )

  expect_equal(ncol(result), 2)
  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  # we get back outcomes vars with a `.pred_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_"), names(outcome_nlevels))

  # result columns are tibbles of resp 2, 2, 4 columns
  expect_true(all(purrr::map_lgl(result, tibble::is_tibble)))
  expect_equal(purrr::map_dbl(result, ncol), outcome_nlevels, ignore_attr = TRUE)

})

test_that("Training multilabel classification from recipe", {

  rec <- recipe(Attrition + JobSatisfaction ~ ., data = attrition[ids,]) %>%
    step_zv(all_predictors()) %>%
    step_normalize(all_numeric(), -all_outcomes())

  expect_no_error(
    fit <- tabnet_fit(rec, attrition[ids,], epochs = 1, valid_split = 0.25,
                      verbose = TRUE)
  )

  expect_no_error(
    result <- predict(fit, attrition)
  )
  expect_equal(ncol(result), 2)

  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  expect_equal(stringr::str_remove(names(result), ".pred_class_"), names(outcome_nlevels))

})

test_that("Training multilabel classification from data.frame with validation split", {

  expect_no_error(
    fit <- tabnet_fit(attri_mult_x, data.frame(y=attriy, z=attriy, sat=attrix$JobSatisfaction),
                      valid_split = 0.2, epochs = 1)
  )

  expect_no_error(
    result <- predict(fit, attri_mult_x, type = "prob")
  )

  expect_equal(ncol(result), 3)

  outcome_nlevels <- purrr::map_dbl(fit$blueprint$ptypes$outcomes, ~nlevels(.x))
  # we get back outcomes vars with a `.pred_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_"), names(outcome_nlevels))

  # result columns are tibbles of resp 2, 2, 4 columns
  expect_true(all(purrr::map_lgl(result, tibble::is_tibble)))
  expect_equal(purrr::map_dbl(result, ncol), outcome_nlevels, ignore_attr = TRUE)

  expect_no_error(
    result <- predict(fit, attri_mult_x)
  )
  expect_equal(ncol(result), 3)

  # we get back outcomes vars with a `.pred_class_` prefix
  expect_equal(stringr::str_remove(names(result), ".pred_class_"), names(fit$blueprint$ptypes$outcomes))
})


test_that("Training multilabel mixed output fails with explicit error", {

  attri_multi_x <- attrix[-which(names(attrix) == "PercentSalaryHike")]
  expect_error(
    fit <- tabnet_fit(attri_multi_x, data.frame(y = attriy, hik = attrix$PercentSalaryHike), epochs = 1),
    "Mixed multi-outcome type"
  )
  expect_error(
    fit <- tabnet_fit(Attrition + PercentSalaryHike ~ ., data = attrition[ids,], epochs = 1),
    "Mixed multi-outcome type"
  )
  rec <- recipe(Attrition + PercentSalaryHike ~ ., data = attrition[ids,]) %>%
    step_normalize(all_numeric(), -all_outcomes())
  expect_error(
    fit <- tabnet_fit(rec, data = attrition[ids,], epochs = 1),
    "Mixed multi-outcome type"
  )
})

test_that("Training multi-output regression fails for matrix", {

  expect_error(
    fit <- tabnet_fit(x, matrix(rnorm( 2 * length(y)), ncol = 2), epochs = 1),
    "All columns of `y` must have unique names"
  )

  expect_error(
    fit <- tabnet_fit(x, matrix(factor(runif( 2 * length(y)) < 0.5) , ncol = 2), epochs = 1),
    "All columns of `y` must have unique names"
  )

})

test_that("early stopping message get translated in french", {
  # skip on linux on ci due to missing language in image
  testthat::skip_if((testthat:::on_ci() && testthat:::system_os() == "linux"))
  testthat::skip_on_cran()
  withr::with_language(lang = "fr",
                       expect_error(
                         tabnet_fit(attrix, attriy, epochs = 200, verbose=TRUE,
                                    early_stopping_monitor="cross_validation_loss",
                                    early_stopping_tolerance=1e-7, early_stopping_patience=3, learn_rate = 0.2),
                         regexp = "n'est pas une m"
                       )
  )
})

test_that("scheduler message translated in french", {
  # skip on linux on ci due to missing language in image
  testthat::skip_if((testthat:::on_ci() && testthat:::system_os() == "linux"))
  testthat::skip_on_cran()
  withr::with_language(lang = "fr",
                       expect_error(
                         fit <- tabnet_pretrain(x, y, epochs = 3, lr_scheduler = "multiplicative",
                                           lr_decay = 0.1, step_size = 1),
                       regexp = "Seule les planifications \"step\" et"
                       )
  )
})

test_that("Supervised training can continue with a additional fit, with or wo from_epoch=", {

  fit_2 <- tabnet_fit(x, y, tabnet_model = ames_fit, epochs = 1)

  expect_equal(fit_2$fit$config$epoch, 1)
  expect_length(fit_2$fit$metrics, 6)
  expect_identical(ames_fit$fit$metrics[[1]]$train, fit_2$fit$metrics[[1]]$train)
  expect_identical(ames_fit$fit$metrics[[5]]$train, fit_2$fit$metrics[[5]]$train)

  expect_no_error(
    fit_3 <- tabnet_fit(x, y, tabnet_model = ames_fit, from_epoch = 2, epoch = 1 )
  )
  expect_equal(fit_3$fit$config$epoch, 1)
  expect_length(fit_3$fit$metrics, 3)
  expect_identical(ames_fit$fit$metrics[[1]]$train, fit_2$fit$metrics[[1]]$train)
  expect_identical(ames_fit$fit$metrics[[2]]$train, fit_2$fit$metrics[[2]]$train)

})

test_that("we can change the tabnet_options between training epoch", {

  fit_2 <- tabnet_fit(x, y, ames_fit, epochs = 1, penalty = 0.003, learn_rate = 0.002)

  expect_equal(fit_2$fit$config$epoch, 1)
  expect_length(fit_2$fit$metrics, 6)
  expect_equal(fit_2$fit$config$learn_rate, 0.002)

})

test_that("epoch counter is valid for retraining from a checkpoint", {

  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_fit, tmp))

  fit1 <- readRDS(tmp)
  fit_2 <- tabnet_fit(x, y, ames_fit, epochs = 12, verbose=T)

  expect_equal(fit_2$fit$config$epoch, 12)
  expect_length(fit_2$fit$metrics, 17)
  expect_lte(mean(fit_2$fit$metrics[[17]]$train), mean(fit_2$fit$metrics[[1]]$train))

})

test_that("training loss keeps decreasing across a checkpoint boundary", {
  # Default checkpoint_epochs = 10; training to epoch 12 crosses the first checkpoint.
  # A broken device round-trip during checkpointing caused the optimizer to lose its
  # parameter references, freezing the loss at an identical floating-point value for
  # every epoch after the first checkpoint.
  fit <- tabnet_fit(Attrition ~ ., data = attrition, epochs = 12, learn_rate = 1e-2)

  expect_length(fit$fit$metrics, 12)
  train_losses <- purrr::map_dbl(fit$fit$metrics, ~mean(.x$train))

  # Loss must still be changing after the checkpoint at epoch 10: a frozen
  # optimizer repeats an exact floating-point value every epoch
  expect_false(train_losses[11] == train_losses[12])
})

test_that("training loss keeps decreasing when resuming from a disk-saved model", {
  # attr_fitted has 12 epochs with checkpoint_epochs = 10 (one checkpoint saved).
  # After disk restore, check_net_is_empty_ptr = TRUE and the code restores weights
  # from serialized_net + apply_checkpoint, then resumes training with epoch_shift = 10.
  # Before the fix, this code path could also freeze the optimizer.
  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(attr_fitted, tmp))

  fit_from_disk <- readRDS(tmp)
  fit2 <- tabnet_fit(attrix, attriy, tabnet_model = fit_from_disk, epochs = 2)

  train_losses <- purrr::map_dbl(fit2$fit$metrics, ~mean(.x$train))
  n <- length(train_losses)

  # After disk restore, the optimizer must still be updating parameters:
  # a frozen optimizer would produce identical loss values every epoch
  expect_false(train_losses[n - 1] == train_losses[n])
  # Overall, loss after resumed training must be lower than at the very start
  expect_lte(train_losses[n], train_losses[1])
})

test_that("trying to continue training with different dataset raise error", {

  pretrain_1 <- tabnet_pretrain(x, y, epochs = 1)

  expect_error(
    pretrain_2 <- tabnet_fit(attrix, y, tabnet_model=pretrain_1, epochs = 1),
    regexp = "Model dimensions"
  )

  fit_1 <- tabnet_fit(x, y, epochs = 1)

  expect_error(
    fit_2 <- tabnet_fit(attrix, y, tabnet_model=fit_1, epochs = 1),
    regexp = "Model dimensions"
  )

  expect_error(
    fit_2 <- tabnet_fit(x, attriy, tabnet_model=fit_1, epochs = 1),
    regexp = "Model dimensions"
  )

})

test_that("Supervised training can continue unsupervised training, with or wo from_epoch=", {

  expect_no_error(
    tabnet_fit(x, y, tabnet_model = ames_pretrain, epoch = 1)
  )

  expect_no_error(
    tabnet_fit(Attrition ~ ., data = attrition, tabnet_model = attr_pretrained, epochs = 1)
  )

  expect_no_error(
    tabnet_fit(x, y, tabnet_model = ames_pretrain, from_epoch = 1, epoch = 1 )
  )

})

test_that("Supervised training can continue unsupervised training, with a Libtorch optimizer", {
  testthat::skip_if(!torch_has_optim_ignite())

  expect_no_error(
    tabnet_pretrain(x, y, epoch = 1, config = tabnet_config(
      optimizer = torch::optim_ignite_adamw)
      )
  )

  expect_no_error(
    tabnet_fit(Attrition ~ ., data = attrition, tabnet_model = attr_pretrained, epochs = 1, 
      optimizer = torch::optim_ignite_adamw
      )
  )
})


test_that("serialization of tabnet_pretrain with saveRDS just works", {

  fit <- tabnet_fit(x, y, ames_pretrain, epoch = 1, learn_rate = 1e-12)

  tmp <- tempfile("model", fileext = "rds")
  withr::local_file(saveRDS(ames_pretrain, tmp))

  pretrain2 <- readRDS(tmp)
  fit2 <- tabnet_fit(x, y, pretrain2, epoch = 1, learn_rate = 1e-12)

  expect_equal(
    predict(fit, ames),
    predict(fit2, ames),
    tolerance = 20
  )

  expect_equal(as.numeric(fit2$fit$network$.check), 1)

})

#' Transforms input data into a list of_tensors and parameters for model input
#'
#' The 3 torch tensors being
#' $x , $x_na_mask, $y
#'  and parameters being
#' cat_idx the vector of x categorical predictor index
#' cat_dims the vector of number of levels of each x categorical predictor
#' input_dim  the number of col in `x`
#' output_dim the `ncol(y)` in case of (multi-outcome) regression or
#'            the `nlevels(y)` in case of classification or
#'            the vector of `nlevels(y)` in case of multi-outcome classification
#'
#' @param x a data frame
#' @param y a response vector
#' @noRd
resolve_data <- function(x, y) {
  cat_idx <- which(sapply(x, is.factor))
  cat_dims <- sapply(cat_idx, function(i) nlevels(x[[i]]))
  # convert factors into integers
  if (length(cat_idx) > 0) {
    x[,cat_idx] <- sapply(cat_idx, function(i) as.integer(x[[i]]))
  } else {
    # prevent empty cat idx
    cat_idx <- 0L
    cat_dims <- 0L
  }
  x_tensor <- torch::torch_tensor(as.matrix(x), dtype = torch::torch_float())
  x_na_mask <- x %>% is.na %>% as.matrix %>% torch::torch_tensor(dtype = torch::torch_bool())

  # convert factors to integers, based on the class of target first column
  # TODO do not assume but assert type-consistency of all y cols
  # and record output_dim
  if (is.factor(y[[1]])) {
    y_tensor <- torch::torch_tensor(sapply(y, function(i) as.integer(i)), dtype = torch::torch_long())
    if (is.atomic(y)) {
      output_dim <- nlevels(y)
    } else {
      output_dim <- sapply(y, function(i) nlevels(i))
    }
  } else {
    y_tensor <- torch::torch_tensor(as.matrix(y), dtype = torch::torch_float())
    output_dim <- ncol(y)
  }

  input_dim <- ncol(x)

  list(x = x_tensor, x_na_mask = x_na_mask, y = y_tensor,
       cat_idx = cat_idx,
       output_dim = output_dim,
       input_dim = input_dim, cat_dims = cat_dims)
}

#' Configuration for TabNet models
#'
#' @param batch_size (int) Number of examples per batch, large batch sizes are
#'   recommended. (default: 1024^2)
#' @param penalty This is the extra sparsity loss coefficient as proposed
#'   in the original paper. The bigger this coefficient is, the sparser your model
#'   will be in terms of feature selection. Depending on the difficulty of your
#'   problem, reducing this value could help (default 1e-3).
#' @param clip_value If a num is given this will clip the gradient at
#'   clip_value. Pass `NULL` to not clip.
#' @param loss (character or function) Loss function for training (default to mse
#'   for regression and cross entropy for classification)
#' @param epochs (int) Number of training epochs.
#' @param drop_last (logical) Whether to drop last batch if not complete during
#'   training
#' @param decision_width (int) Width of the decision prediction layer. Bigger values gives
#'   more capacity to the model with the risk of overfitting. Values typically
#'   range from 8 to 64.
#' @param attention_width (int) Width of the attention embedding for each mask. According to
#'   the paper n_d = n_a is usually a good choice. (default=8)
#' @param num_steps (int) Number of steps in the architecture
#'   (usually between 3 and 10)
#' @param feature_reusage (num) This is the coefficient for feature reusage in the masks.
#'   A value close to 1 will make mask selection least correlated between layers.
#'   Values range from 1 to 2.
#' @param mask_type (character) Final layer of feature selector in the attentive_transformer
#'   block, either `"sparsemax"`, `"entmax"` or `"entmax15"`.Defaults to `"sparsemax"`.
#' @param mask_topk (int) mask sparsity top-k for `sparsemax15` and `entmax15.` See [entmax15()] for detail.
#' @param virtual_batch_size (int) Size of the mini batches used for
#'   "Ghost Batch Normalization" (default=256^2)
#' @param learn_rate initial learning rate for the optimizer.
#' @param optimizer the optimization method. currently only `"adam"` is supported,
#'   you can also pass any torch optimizer function.
#' @param valid_split In \[0, 1). The fraction of the dataset used for validation.
#'   (default = 0 means no split)
#' @param num_independent Number of independent Gated Linear Units layers at each step of the encoder.
#'   Usual values range from 1 to 5.
#' @param num_shared Number of shared Gated Linear Units at each step of the encoder. Usual values
#'    at each step of the decoder. range from 1 to 5
#' @param num_independent_decoder For pretraining, number of independent Gated Linear Units layers
#'   Usual values range from 1 to 5.
#' @param num_shared_decoder For pretraining, number of shared Gated Linear Units at each step of the
#'    decoder. Usual values range from 1 to 5.
#' @param verbose (logical) Whether to print progress and loss values during
#'   training.
#' @param lr_scheduler if `NULL`, no learning rate decay is used. If "step"
#'   decays the learning rate by `lr_decay` every `step_size` epochs. If "reduce_on_plateau"
#'   decays the learning rate by `lr_decay` when no improvement after `step_size` epochs.
#'   It can also be a [torch::lr_scheduler] function that only takes the optimizer
#'   as parameter. The `step` method is called once per epoch.
#' @param lr_decay multiplies the initial learning rate by `lr_decay` every
#'   `step_size` epochs. Unused if `lr_scheduler` is a `torch::lr_scheduler`
#'   or `NULL`.
#' @param step_size the learning rate scheduler step size. Unused if
#'   `lr_scheduler` is a `torch::lr_scheduler` or `NULL`.
#' @param cat_emb_dim Size of the embedding of categorical features. If int, all categorical
#'   features will have same embedding size, if list of int, every corresponding feature will have
#'   specific embedding size.
#' @param momentum Momentum for batch normalization, typically ranges from 0.01
#'   to 0.4 (default=0.02)
#' @param pretraining_ratio Ratio of features to mask for reconstruction during
#'   pretraining.  Ranges from 0 to 1 (default=0.5)
#' @param checkpoint_epochs checkpoint model weights and architecture every
#'   `checkpoint_epochs`. (default is 10). This may cause large memory usage.
#'   Use `0` to disable checkpoints.
#' @param device the device to use for training. "cpu" or "cuda". The default ("auto")
#'   uses  to "cuda" if it's available, otherwise uses "cpu".
#' @param importance_sample_size sample of the dataset to compute importance metrics.
#'   If the dataset is larger than 1e5 obs we will use a sample of size 1e5 and
#'   display a warning.
#' @param early_stopping_monitor Metric to monitor for early_stopping. One of "valid_loss", "train_loss" or "auto" (defaults to "auto").
#' @param  early_stopping_tolerance Minimum relative improvement to reset the patience counter.
#'  0.01 for 1% tolerance (default 0)
#' @param early_stopping_patience Number of epochs without improving until stopping training. (default=5)
#' @param num_workers (int, optional): how many subprocesses to use for data
#'   loading. 0 means that the data will be loaded in the main process.
#'   (default: `0`)
#' @param skip_importance if feature importance calculation should be skipped (default: `FALSE`)
#' @return A named list with all hyperparameters of the TabNet implementation.
#'
#' @examplesIf (torch::torch_is_installed() && require("modeldata"))
#' data("ames", package = "modeldata")
#'
#' # change the model config for an faster ignite optimizer
#' config <- tabnet_config(optimizer = torch::optim_ignite_adamw)
#'
#' ## Single-outcome regression using formula specification
#' fit <- tabnet_fit(Sale_Price ~ ., data = ames, epochs = 1, config = config)
#'
#' @export
tabnet_config <- function(batch_size = 1024^2,
                          penalty = 1e-3,
                          clip_value = NULL,
                          loss = "auto",
                          epochs = 5,
                          drop_last = FALSE,
                          decision_width = NULL,
                          attention_width = NULL,
                          num_steps = 3,
                          feature_reusage = 1.3,
                          mask_type = "sparsemax",
                          mask_topk = NULL,
                          virtual_batch_size = 256^2,
                          valid_split = 0,
                          learn_rate = 2e-2,
                          optimizer = "adam",
                          lr_scheduler = NULL,
                          lr_decay = 0.1,
                          step_size = 30,
                          checkpoint_epochs = 10,
                          cat_emb_dim = 1,
                          num_independent = 2,
                          num_shared = 2,
                          num_independent_decoder = 1,
                          num_shared_decoder = 1,
                          momentum = 0.02,
                          pretraining_ratio = 0.5,
                          verbose = FALSE,
                          device = "auto",
                          importance_sample_size = NULL,
                          early_stopping_monitor = "auto",
                          early_stopping_tolerance = 0,
                          early_stopping_patience = 0L,
                          num_workers=0L,
                          skip_importance = FALSE
                          ) {
  if (is.null(decision_width) && is.null(attention_width)) {
    decision_width <- 8 # default is 8
  }

  if (is.null(attention_width))
    attention_width <- decision_width

  if (is.null(decision_width))
    decision_width <- attention_width

  list(
    batch_size = batch_size,
    lambda_sparse = penalty,
    clip_value = clip_value,
    loss = loss,
    epochs = epochs,
    drop_last = drop_last,
    n_d = decision_width,
    n_a = attention_width,
    n_steps = num_steps,
    gamma = feature_reusage,
    mask_type = mask_type,
    mask_topk = mask_topk,
    virtual_batch_size = virtual_batch_size,
    valid_split = valid_split,
    learn_rate = learn_rate,
    optimizer = resolve_optimizer(optimizer),
    lr_scheduler = lr_scheduler,
    lr_decay = lr_decay,
    step_size = step_size,
    checkpoint_epochs = checkpoint_epochs,
    cat_emb_dim = cat_emb_dim,
    n_independent = num_independent,
    n_shared = num_shared,
    n_independent_decoder = num_independent_decoder,
    n_shared_decoder = num_shared_decoder,
    momentum = momentum,
    pretraining_ratio = pretraining_ratio,
    verbose = verbose,
    device = device,
    importance_sample_size = importance_sample_size,
    early_stopping_monitor = resolve_early_stop_monitor(early_stopping_monitor, valid_split),
    early_stopping_tolerance = early_stopping_tolerance,
    early_stopping_patience = early_stopping_patience,
    early_stopping = !(early_stopping_tolerance == 0 || early_stopping_patience == 0),
    num_workers = num_workers,
    skip_importance = skip_importance
  )
}

resolve_loss <- function(config, dtype) {
  loss <- config$loss

  if (is_loss_generator(loss))
    loss_fn <- loss()
  else if (is.function(loss))
    loss_fn <- loss
  else if (loss %in% c("mse", "auto") && !dtype == torch::torch_long())
    loss_fn <- torch::nn_mse_loss()
  else if (!is.null(config$ancestor))
    loss_fn <- nn_mc_loss(R = config$ancestor)
  else if ((loss %in% c("bce", "cross_entropy", "auto") && dtype == torch::torch_long()))
    # cross entropy loss is required
    loss_fn <- torch::nn_cross_entropy_loss()
  else
    value_error("{.val {loss}} is not a valid loss for outcome of type {.type {dtype}}")

  loss_fn
}

resolve_early_stop_monitor <- function(early_stopping_monitor, valid_split) {
  if (early_stopping_monitor %in% c("valid_loss", "auto") && valid_split > 0)
    early_stopping_monitor <- "valid_loss"
  else if (early_stopping_monitor %in% c("train_loss", "auto"))
    early_stopping_monitor <- "train_loss"
  else
    value_error("{.val {early_stopping_monitor}} is not a valid early-stopping metric to monitor with {.val valid_split = {valid_split}}")

  early_stopping_monitor
}

train_batch <- function(network, optimizer, batch, config) {
  # NULL-ing values to avoid a R-CMD Check Note "No visible binding for global variable"
  out <- M_loss <- NULL
  # forward pass
  c(out, M_loss) %<-% network(batch$x, batch$x_na_mask)

  # if target is multi-outcome but not max_constraint loss, loss has to be applied to each label-group
  if (max(batch$output_dim$shape) > 1 && is.null(config$ancestor)) {
    # standard multi-outcome
    outcome_nlevels <- as.numeric(batch$output_dim$to(device="cpu"))

    # use `resolved_loss`
    loss <- torch::torch_sum(torch::torch_stack(purrr::pmap(
      list(
        torch::torch_split(out, outcome_nlevels, dim = 2),
        torch::torch_split(batch$y, rep(1, length(outcome_nlevels)), dim = 2)
      ),
      ~config$loss_fn(.x, .y$squeeze(2))
    )),
    dim = 1)
  } else if (!is.null(config$ancestor)) {
    # multi-outcome max_constraint loss ned one-hot encoding of targets
    loss <- config$loss_fn(out, nnf_multilabel_one_hot(
      y = batch$y,
      outcomes = config$outcomes,
      device = out$device
    ))
  } else if (batch$y$dtype == torch::torch_long()) {
    # classifier needs a squeeze for bce loss
    loss <- config$loss_fn(out, batch$y$squeeze(2))
  } else {
    loss <- config$loss_fn(out, batch$y)
  }
  # Add the overall sparsity loss
  loss <- loss - config$lambda_sparse * M_loss

  # step of the optimization
  optimizer$zero_grad()
  loss$backward()
  if (!is.null(config$clip_value)) {
    torch::nn_utils_clip_grad_norm_(network$parameters, config$clip_value)
  }
  optimizer$step()

  list(
    loss = loss$item()
  )
}

valid_batch <- function(network, batch, config) {
  # NULLing values to avoid a R-CMD Check Note "No visible binding for global variable"
  out <- M_loss <- NULL
  # forward pass
  c(out, M_loss) %<-% network(batch$x, batch$x_na_mask)
  # loss has to be applied to each label-group when output_dim is a vector
  if (max(batch$output_dim$shape) > 1 && is.null(config$ancestor)) {
    # standard multi-outcome
    outcome_nlevels <- as.numeric(batch$output_dim$to(device="cpu"))
    # use `resolved_loss`
    loss <- torch::torch_sum(torch::torch_stack(purrr::pmap(
      list(
        torch::torch_split(out, outcome_nlevels, dim = 2),
        torch::torch_split(batch$y, rep(1, length(outcome_nlevels)), dim = 2)
      ),
      ~config$loss_fn(.x, .y$squeeze(2))
    )),
    dim = 1)
    
  } else if (!is.null(config$ancestor)) {
    # multi-outcome max_constraint loss ned one-hot encoding of targets
    loss <- config$loss_fn(out, nnf_multilabel_one_hot(
      y = batch$y,
      outcomes = config$outcomes,
      device = out$device
    ))
  } else {
    if (batch$y$dtype == torch::torch_long()) {
      # classifier needs a squeeze for bce loss
      loss <- config$loss_fn(out, batch$y$squeeze(2))
    } else {
      loss <- config$loss_fn(out, batch$y)
    }
  }
  # Add the overall sparsity loss
  loss <- loss - config$lambda_sparse * M_loss

  list(
    loss = loss$item()
  )
}

get_device_from_config <- function(config) {
  if (config$device == "auto") {
    if (torch::cuda_is_available()){
      device <- "cuda"
    } else if (torch::backends_mps_is_available()) {
      device <- "mps"
    } else {
      device <- "cpu"
    }
  } else {
    device <- config$device
  }
  device
}

tabnet_initialize <- function(x, y, config = tabnet_config()) {

  torch::torch_manual_seed(sample.int(1e6, 1))
  has_valid <- config$valid_split > 0

  device <- get_device_from_config(config)

  if (has_valid) {
    n <- nrow(x)
    valid_idx <- sample.int(n, n * config$valid_split)
    valid_x <- x[valid_idx, ]
    valid_y <- y[valid_idx, ]
    train_y <- y[-valid_idx, ]
    valid_ds <-   torch::dataset(
      initialize = function() {},
      .getbatch = function(batch) {resolve_data(valid_x[batch,], valid_y[batch, ])},
      .length = function() {nrow(valid_x)}
    )()
    x <- x[-valid_idx, ]
    y <- train_y
  }

  # training dataset
  train_ds <-   torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {resolve_data(x[batch, ], y[batch, ])},
    .length = function() {nrow(x)}
  )()
  # we can get training_set parameters from the 2 first samples
  train <- train_ds$.getbatch(batch = c(1:2))

  # resolve loss
  config$loss_fn <- resolve_loss(config, train$y$dtype)

  # create network
  network <- tabnet_nn(
    input_dim = train$input_dim,
    output_dim = train$output_dim,
    cat_idxs = train$cat_idx,
    cat_dims = train$cat_dims,
    n_d = config$n_d,
    n_a = config$n_a,
    n_steps = config$n_steps,
    gamma = config$gamma,
    virtual_batch_size = config$virtual_batch_size,
    cat_emb_dim = config$cat_emb_dim,
    n_independent = config$n_independent,
    n_shared = config$n_shared,
    momentum = config$momentum,
    mask_type = config$mask_type,
    mask_topk = config$mask_topk
  )

  # main loop
  metrics <- list()
  checkpoints <- list()


  importances <- tibble::tibble(
    variables = colnames(x),
    importance = NA
  )

  list(
    network = network,
    metrics = metrics,
    config = config,
    checkpoints = checkpoints,
    importances = importances
  )
}

tabnet_train_supervised <- function(obj, x, y, config = tabnet_config(), epoch_shift = 0L) {
  stopifnot("tabnet_model shall be initialised or pretrained" = (length(obj$fit$network) > 0))
  torch::torch_manual_seed(sample.int(1e6, 1))

  device <- get_device_from_config(config)

  # validation dataset & dataloaders
  has_valid <- config$valid_split > 0
  if (has_valid) {
    n <- nrow(x)
    valid_idx <- sample.int(n, n * config$valid_split)
    valid_x <- x[valid_idx, ]
    valid_y <- y[valid_idx, ]
    train_y <- y[-valid_idx, ]

    valid_ds <-   torch::dataset(
      initialize = function() {},
      .getbatch = function(batch) {resolve_data(valid_x[batch,], valid_y[batch,])},
      .length = function() {nrow(valid_x)}
    )()

    valid_dl <- torch::dataloader(
      valid_ds,
      batch_size = config$batch_size,
      shuffle = FALSE,
      num_workers = config$num_workers
    )

    x <- x[-valid_idx, ]
    y <- train_y
  }

  # training dataset & dataloader
  train_ds <-   torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {resolve_data(x[batch,], y[batch,])},
    .length = function() {nrow(x)}
  )()

  train_dl <- torch::dataloader(
    train_ds,
    batch_size = config$batch_size,
    drop_last = config$drop_last,
    shuffle = TRUE ,
    num_workers = config$num_workers
  )

  # move ancestor matrix to training device before building the loss function
  if (!is.null(config$ancestor)) {
    config$ancestor <- config$ancestor$to(device = device)
  }

  # resolve loss
  config$loss_fn <- resolve_loss(config, train_ds$.getbatch(batch = c(1:2))$y$dtype)

  # restore network from model and send it to device
  network <- obj$fit$network

  network$to(device = device)

  # provide ancestor to torch tensor in case of hierarchical classification
  if (!is.null(config$ancestor)) {
    if (!inherits(config$ancestor, "torch_tensor")) {
      # config is expected to carry the tensor
      runtime_error("ancestor was configured. Expecting a tensor but got {.cls {class(config$ancestor)}}")
    }
  }

  # instantiate optimizer
  if (is_optim_generator(config$optimizer)) {
    optimizer <- config$optimizer(network$parameters, config$learn_rate)
  } else {
    type_error("{.var optimizer} must be resolved into a torch optimizer generator.")
  }

  # define scheduler
  if (is.null(config$lr_scheduler)) {
    scheduler <- list(step = function() {})
  } else if (rlang::is_function(config$lr_scheduler)) {
    scheduler <- config$lr_scheduler(optimizer)
  } else if (config$lr_scheduler == "reduce_on_plateau") {
    scheduler <- torch::lr_reduce_on_plateau(optimizer, factor = config$lr_decay, patience = config$step_size)
  } else if (config$lr_scheduler == "step") {
    scheduler <- torch::lr_step(optimizer, config$step_size, config$lr_decay)
  } else {
    not_implemented_error("Currently only the {.str step} and {.str reduce_on_plateau} scheduler are supported.", call. = FALSE)
  }

  # restore previous metrics & checkpoints
  metrics <- obj$fit$metrics
  checkpoints <- obj$fit$checkpoints
  patience_counter <- 0L

  # main loop
  for (epoch in seq_len(config$epochs) + epoch_shift) {

    metrics[[epoch]] <- list()
    train_metrics <- c()
    valid_metrics <- c()

    network$train()

    if (config$verbose)
      pb <- progress::progress_bar$new(
        total = length(train_dl),
        format = "[:bar] loss= :loss"
      )

    coro::loop(for (batch in train_dl) {
      m <- train_batch(network, optimizer, to_device(batch, device), config)
      if (config$verbose) pb$tick(tokens = m)
      train_metrics <- c(train_metrics, m)
    })
    metrics[[epoch]][["train"]] <- transpose_metrics(train_metrics)$loss

    if (config$checkpoint_epochs > 0 && epoch %% config$checkpoint_epochs == 0) {
      checkpoints[[length(checkpoints) + 1]] <- state_to_raw(network)
      metrics[[epoch]][["checkpoint"]] <- TRUE
    }

    network$eval()
    if (has_valid) {
      coro::loop(for (batch in valid_dl) {
        m <- valid_batch(network, to_device(batch, device), config)
        valid_metrics <- c(valid_metrics, m)
      })
      metrics[[epoch]][["valid"]] <- transpose_metrics(valid_metrics)$loss
    }

    if (config$verbose && !has_valid)
      message(gettextf("[Epoch %03d] Loss: %3f", epoch, mean(metrics[[epoch]]$train)))
    if (config$verbose && has_valid)
      message(gettextf("[Epoch %03d] Loss: %3f, Valid loss: %3f", epoch, mean(metrics[[epoch]]$train), mean(metrics[[epoch]]$valid)))


    # Early-stopping checks
    if (config$early_stopping && config$early_stopping_monitor=="valid_loss"){
      current_loss <- mean(metrics[[epoch]]$valid)
    } else {
      current_loss <- mean(metrics[[epoch]]$train)
    }
    if (config$early_stopping && epoch > 1+epoch_shift) {
      # compute relative change, and compare to best_metric
      change <- (current_loss - best_metric) / current_loss
      if (change > config$early_stopping_tolerance){
        patience_counter <- patience_counter + 1
        if (patience_counter >= config$early_stopping_patience){
          if (config$verbose)
            cli::cli_alert_success(gettextf("Early-stopping at epoch {.val epoch}"))
          break
        }
      } else {
        # reset the patience counter
        best_metric <- current_loss
        patience_counter <- 0L
      }
    }
    if (config$early_stopping && epoch == 1+epoch_shift) {
      # initialise best_metric
      best_metric <- current_loss
    }

    if ("metrics" %in% names(formals(scheduler$step))) {
      scheduler$step(current_loss)
    } else {
      scheduler$step()
    }
  }

  network$to(device = "cpu")
  if(!config$skip_importance) {
    importance_sample_size <- config$importance_sample_size
    if (is.null(config$importance_sample_size) && train_ds$.length() > 1e5) {
      warn("Computing importances for a dataset with size {.val {train_ds$.length()}}. 
           This can consume too much memory. We are going to use a sample of size 1e5. 
           You can disable this message by using the `importance_sample_size` argument.")
      importance_sample_size <- 1e5
    }
    indexes <- as.numeric(torch::torch_randint(
      1, train_ds$.length(), min(importance_sample_size, train_ds$.length()),
      dtype = torch::torch_long()
    ))
    importances <- tibble::tibble(
      variables = colnames(x),
      importance = compute_feature_importance(
        network,
        train_ds$.getbatch(batch =indexes)$x$to(device = "cpu"),
        train_ds$.getbatch(batch =indexes)$x_na_mask$to(device = "cpu"))
    )
  } else {
    importances <- NULL
  }
  
  list(
    network = network,
    metrics = metrics,
    config = config,
    checkpoints = checkpoints,
    importances = importances
  )
}

predict_impl <- function(obj, x, batch_size = 1e5) {
  # prediction dataset
  device = obj$fit$config$device
  predict_ds <-   torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {resolve_data(x[batch,], matrix(1L, nrow = length(batch)))},
    .length = function() {nrow(x)}
  )()

  network <- obj$fit$network
  num_workers <- obj$fit$config$num_workers
  yhat <- c()
  network$eval()

  predict_dl <- torch::dataloader(
    predict_ds,
    batch_size = batch_size,
    drop_last = FALSE,
    shuffle = FALSE ,
    # num_workers = num_workers
    num_workers = 0L
  )
  coro::loop(for (batch in predict_dl) {
    batch <- to_device(batch, device)
    yhat <- c(yhat, network(batch$x, batch$x_na_mask)[[1]])
  })
  # bind rows of the batches
  torch::torch_cat(yhat)
}

predict_impl_numeric <- function(obj, x, batch_size) {
  p <- as.matrix(predict_impl(obj, x, batch_size))
  hardhat::spruce_numeric(as.numeric(p))
}

predict_impl_numeric_multiple <- function(obj, x, batch_size) {
  p <- as.matrix(predict_impl(obj, x, batch_size))
  # TODO use a cleaner function to turn matrix into vectors
  hardhat::spruce_numeric_multiple(!!!purrr::map(seq_len(ncol(p)), ~p[,.x]))
}

#' single-outcome level blueprint
#'
#' @param obj : a tabnet object
#'
#' @return : outcome levels
#' @noRd
get_blueprint_levels <- function(obj) {
  levels(obj$blueprint$ptypes$outcomes[[1]])
}

#' multi-outcome levels blueprint
#'
#' @param obj : a tabnet object
#'
#' @return : a list of levels vectors for each outcome
#' @noRd
get_blueprint_levels_multiple <- function(obj) {
  purrr::map(obj$blueprint$ptypes$outcomes, levels) %>%
    rlang::set_names(names(obj$blueprint$ptypes$outcomes))
}

predict_impl_prob <- function(obj, x, batch_size) {
  p <- predict_impl(obj, x, batch_size)
  p <- torch::nnf_softmax(p, dim = 2)
  p <- as.matrix(p)
  hardhat::spruce_prob(get_blueprint_levels(obj), p)
}

predict_impl_prob_multiple <- function(obj, x, batch_size, outcome_nlevels) {
  p <- predict_impl(obj, x, batch_size)
  p <- torch::nnf_softmax(p, dim = 2)
  p <- as.matrix(p)
  # TODO use a cleaner function to turn matrix into vectors
  p_blueprint <- get_blueprint_levels_multiple(obj)
  p_probs <- purrr::map(torch::torch_split(p, outcome_nlevels, dim = 2),
                        as.matrix)
  hardhat::spruce_prob_multiple(!!!purrr::pmap(
    list(p_blueprint, p_probs),
    # TODO BUG each element of `...` must be a tibble, not a list.
    ~hardhat::spruce_prob(.x, .y)) %>%
      rlang::set_names(names(p_blueprint))
    )
}

predict_impl_class <- function(obj, x, batch_size) {
  p <- predict_impl(obj, x, batch_size)
  p_idx <- as.integer(torch::torch_max(p, dim = 2)[[2]])
  p_idx <- get_blueprint_levels(obj)[p_idx]
  p <- factor(p_idx, levels = get_blueprint_levels(obj))
  hardhat::spruce_class(p)
}

predict_impl_class_multiple <- function(obj, x, batch_size, outcome_nlevels) {
  p <- predict_impl(obj, x, batch_size)
  p_levels <- get_blueprint_levels_multiple(obj)
  p_idx <- purrr::map(
    torch::torch_split(p, outcome_nlevels, dim = 2),
    ~as.integer(torch::torch_max(.x, dim = 2)[[2]])
    ) %>% rlang::set_names(names(p_levels))
  p_factor_lst <- purrr::pmap(
    list(p_idx, p_levels),
    ~factor(.y[.x], levels = .y)
  )
  hardhat::spruce_class_multiple(!!!p_factor_lst)
}

if (torch::cuda_is_available()) {
  device <- "cuda"
} else {
  device <- "cpu"
}


test_that("resolve_data works through a dataloader", {
  data("ames", package = "modeldata")

  x <- ames[-which(names(ames) == "Sale_Price")]
  y <- ames[,"Sale_Price"]
  # dataset are R6 class and shall be instantiated
  train_ds <- torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {tabnet:::resolve_data(x[batch,], y[batch,])},
    .length = function() {nrow(x)}
  )()
  expect_no_error(
    train_ds$.getbatch(batch = 1:2)
  )
  # dataloader
  train_dl <- torch::dataloader(
    train_ds,
    batch_size = 2000 ,
    drop_last = TRUE,
    shuffle = FALSE #,
    # num_workers = 0L
  )
  expect_no_error(
    coro::loop(for (batch in train_dl) {
      expect_tensor_shape(batch$x, c(2000, 73))
      expect_true(batch$x$dtype == torch::torch_float())
      expect_tensor_shape(batch$x_na_mask, c(2000, 73))
      expect_true(batch$x_na_mask$dtype == torch::torch_bool())
      expect_tensor_shape(batch$y, c(2000, 1))
      expect_true(batch$y$dtype == torch::torch_float())
      expect_tensor_shape(batch$cat_idx, 40)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_equal_to_r(batch$output_dim, 1L)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_tensor_shape(batch$input_dim, 1)
      expect_true(batch$input_dim$dtype == torch::torch_long())
      expect_tensor_shape(batch$cat_dims, 40)
      expect_true(batch$cat_dims$dtype == torch::torch_long())

    })
  )

})

test_that("resolve_data works through a dataloader without nominal variables", {
  n <- 1000
  x <- data.frame(
    x = rnorm(n),
    y = rnorm(n),
    z = rnorm(n)
  )

  y <- x[,"x", drop = FALSE]
  # dataset are R6 class and shall be instanciated
  train_ds <- torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {tabnet:::resolve_data(x[batch,], y[batch,])},
    .length = function() {nrow(x)}
  )()
  expect_no_error(
    train_ds$.getbatch(batch = 1:2)
  )

  # dataloader
  train_dl <- torch::dataloader(
    train_ds,
    batch_size = 2000 ,
    drop_last = TRUE,
    shuffle = FALSE #,
    # num_workers = 0L
  )
  expect_no_error(
    coro::loop(for (batch in train_dl) {
      expect_tensor_shape(batch$x, c(2000, 3))
      expect_true(batch$x$dtype == torch::torch_float())
      expect_tensor_shape(batch$x_na_mask, c(2000, 3))
      expect_true(batch$x_na_mask$dtype == torch::torch_bool())
      expect_tensor_shape(batch$y, c(2000, 1))
      expect_true(batch$y$dtype == torch::torch_float())
      expect_tensor_shape(batch$cat_idx, 0)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_equal_to_r(batch$output_dim, 1L)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_tensor_shape(batch$input_dim, 1)
      expect_true(batch$input_dim$dtype == torch::torch_long())
      expect_tensor_shape(batch$cat_dims, 0)
      expect_true(batch$cat_dims$dtype == torch::torch_long())

    })
  )

})

test_that("resolve_data works for multioutput regression", {
  data("ames", package = "modeldata")

  x <- ames[-which(names(ames) %in% c("Sale_Price", "Lot_Area"))]
  y <- ames[,c("Sale_Price", "Lot_Area")]
  # dataset are R6 class and shall be instantiated
  train_ds <- torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {tabnet:::resolve_data(x[batch,], y[batch,])},
    .length = function() {nrow(x)}
  )()
  expect_error(
    train_ds$.getbatch(batch = 1:2),
    NA
  )
  # dataloader
  train_dl <- torch::dataloader(
    train_ds,
    batch_size = 2000 ,
    drop_last = TRUE,
    shuffle = FALSE #,
    # num_workers = 0L
  )
  expect_no_error(
    coro::loop(for (batch in train_dl) {
      expect_tensor_shape(batch$x, c(2000, 72))
      expect_true(batch$x$dtype == torch::torch_float())
      expect_tensor_shape(batch$x_na_mask, c(2000, 72))
      expect_true(batch$x_na_mask$dtype == torch::torch_bool())
      expect_tensor_shape(batch$y, c(2000, 2))
      expect_true(batch$y$dtype == torch::torch_float())
      expect_tensor_shape(batch$cat_idx, 40)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_equal_to_r(batch$output_dim, 2L)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_tensor_shape(batch$input_dim, 1)
      expect_true(batch$input_dim$dtype == torch::torch_long())
      expect_tensor_shape(batch$cat_dims, 40)
      expect_true(batch$cat_dims$dtype == torch::torch_long())

    })
  )

})
test_that("resolve_data works for multioutput classification", {

  x <- attrix[-which(names(attrix) == "JobSatisfaction")]
  y <- data.frame(y = attriy, z = attriy, sat = attrix$JobSatisfaction)
  # dataset are R6 class and shall be instantiated
  train_ds <- torch::dataset(
    initialize = function() {},
    .getbatch = function(batch) {tabnet:::resolve_data(x[batch,], y[batch,])},
    .length = function() {nrow(x)}
  )()
  expect_error(
    train_ds$.getbatch(batch = 1:2),
    NA
  )
  # dataloader
  train_dl <- torch::dataloader(
    train_ds,
    batch_size = 2000 ,
    drop_last = TRUE,
    shuffle = FALSE #,
    # num_workers = 0L
  )
  expect_no_error(
    coro::loop(for (batch in train_dl) {
      expect_tensor_shape(batch$x, c(2000, 72))
      expect_true(batch$x$dtype == torch::torch_float())
      expect_tensor_shape(batch$x_na_mask, c(2000, 72))
      expect_true(batch$x_na_mask$dtype == torch::torch_bool())
      expect_tensor_shape(batch$y, c(2000, 2))
      expect_true(batch$y$dtype == torch::torch_float())
      expect_tensor_shape(batch$cat_idx, 40)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_equal_to_r(batch$output_dim, 2L)
      expect_true(batch$cat_idx$dtype == torch::torch_long())
      expect_tensor_shape(batch$input_dim, 1)
      expect_true(batch$input_dim$dtype == torch::torch_long())
      expect_tensor_shape(batch$cat_dims, 40)
      expect_true(batch$cat_dims$dtype == torch::torch_long())

    })
  )

})

#' Interpretation metrics from a TabNet model
#'
#' @param object a TabNet fit object
#' @param new_data a data.frame to obtain interpretation metrics.
#'
#' @return
#'
#' Returns a list with
#'
#' * `M_explain`: the aggregated feature importance masks as detailed in
#'   TabNet's paper.
#' * `masks` a list containing the masks for each step.
#'
#' @examplesIf torch::torch_is_installed()
#'
#' set.seed(2021)
#'
#' n <- 256
#' x <- data.frame(
#'   x = rnorm(n),
#'   y = rnorm(n),
#'   z = rnorm(n)
#' )
#'
#' y <- x$x
#'
#' fit <- tabnet_fit(x, y, epochs = 10,
#'                   num_steps = 1,
#'                   batch_size = 512,
#'                   attention_width = 1,
#'                   num_shared = 1,
#'                   num_independent = 1)
#'
#'
#'  ex <- tabnet_explain(fit, x)
#'
#'
#' @export
tabnet_explain <- function(object, new_data) {
  UseMethod("tabnet_explain")
}

#' @export
#' @rdname tabnet_explain
tabnet_explain.default <- function(object, new_data) {
  type_error("{.fn tabnet_explain} is not defined for a {.type {class(object)[1]}}.")
}

#' @export
#' @rdname tabnet_explain
tabnet_explain.tabnet_fit <- function(object, new_data) {
  if (inherits(new_data, "Node")) {
    new_data_df <- node_to_df(new_data)$x
  } else {
    new_data_df <- new_data
  }
  # Enforces column order, type, column names, etc
  processed <- hardhat::forge(new_data_df, object$blueprint, outcomes = FALSE)
  data <- resolve_data(processed$predictors, y = rep(1, nrow(processed$predictors)))
  device <- get_device_from_config(object$fit$config)
  data <- to_device(data, device)
  output <- explain_impl(object$fit$network, data$x, data$x_na_mask)

  # convert stuff to matrix with colnames
  nms <- colnames(processed$predictors)
  output$M_explain <- convert_to_df(output$M_explain, nms)
  output$masks <- lapply(output$masks, convert_to_df, nms = nms)
  class(output) <- "tabnet_explain"
  output
}

#' @export
#' @rdname tabnet_explain
tabnet_explain.tabnet_pretrain <- tabnet_explain.tabnet_fit

#' @export
#' @rdname tabnet_explain
tabnet_explain.model_fit <- function(object, new_data) {
  tabnet_explain(parsnip::extract_fit_engine(object), new_data)
}

convert_to_df <- function(x, nms) {
  x <- as.data.frame(as.matrix(x$to(device = "cpu")$detach()))
  colnames(x) <- nms
  tibble::as_tibble(x)
}

explain_impl <- function(network, x, x_na_mask) {
  curr_device <- network$.check$device
  withr::defer({
    network$to(device = curr_device)
  })
  network$to(device=x$device)
  # NULLing values to avoid a R-CMD Check Note
  M_explain_emb_dim <- masks_emb_dim <- NULL
  c(M_explain_emb_dim, masks_emb_dim) %<-% network$forward_masks(x, x_na_mask)

  # summarize the categorical embeddedings into 1 column
  # per variable
  M_explain <- sum_embedding_masks(
    mask = M_explain_emb_dim,
    input_dim = network$input_dim,
    cat_idx = network$cat_idxs,
    cat_emb_dim = network$cat_emb_dim
  )

  masks <- lapply(
    masks_emb_dim,
    FUN = sum_embedding_masks,
    input_dim = network$input_dim,
    cat_idx = network$cat_idxs,
    cat_emb_dim = network$cat_emb_dim
  )

  list(M_explain = M_explain$to(device="cpu"), masks = to_device(masks, "cpu"))
}

compute_feature_importance <- function(network, x, x_na_mask) {
  out <- explain_impl(network, x, x_na_mask)
  m <- as.numeric(out$M_explain$sum(dim = 1)$detach()$to(device = "cpu"))
  m/sum(m)
}

# sum embeddings taking their sizes into account.
sum_embedding_masks <- function(mask, input_dim, cat_idx, cat_emb_dim) {
  sizes <- rep(1, input_dim)
  sizes[cat_idx] <- cat_emb_dim

  splits <- mask$split_with_sizes(sizes, dim = 2)
  splits <- lapply(splits, torch::torch_sum, dim = 2, keepdim = TRUE)

  torch::torch_cat(splits, dim = 2)
}

vi_model.tabnet_fit <- function(object, ...) {
  tib <- object$fit$importances
  names(tib) <- c("Variable", "Importance")
  tib
}

vi_model.tabnet_pretrain <- vi_model.tabnet_fit

# Run before any test
suppressPackageStartupMessages(library(recipes))
suppressPackageStartupMessages(library(ggplot2))
suppressPackageStartupMessages(library(data.tree))


# ames small data
utils::data("ames", package = "modeldata")
ids <- sample(nrow(ames), 256)
small_ames <- ames[ids,]
x <- ames[ids,-which(names(ames) == "Sale_Price")]
y <- ames[ids,]$Sale_Price

# ames common models
ames_pretrain <- tabnet_pretrain(x, y, epoch = 2, checkpoint_epochs = 1)
ames_pretrain_vsplit <- tabnet_pretrain(x, y, epochs = 3, valid_split=0.2,
                                        num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)
ames_fit <- tabnet_fit(x, y, epochs = 5 , checkpoint_epochs = 2)
ames_fit_vsplit <- tabnet_fit(x, y, tabnet_model=ames_pretrain_vsplit, epochs = 3,
                              num_steps = 1, attention_width = 1, num_shared = 1, num_independent = 1)

# attrition small data
utils::data("attrition", package = "modeldata")
ids <- sample(nrow(attrition), 256)

# attrition common models
attrix <- attrition[ids,-which(names(attrition) == "Attrition")]
attri_mult_x <- attrix[-which(names(attrix) == "JobSatisfaction")]

attriy <- attrition[ids,]$Attrition

attr_pretrained <- tabnet_pretrain(attrix, attriy, epochs = 12)
attr_pretrained_vsplit <- tabnet_pretrain(attrix, attriy, epochs = 12, valid_split=0.3)
attr_fitted <- tabnet_fit(attrix, attriy, epochs = 12)
attr_fitted_vsplit <- tabnet_fit(attrix, attriy, epochs = 12, valid_split=0.3)

# data.tree Node dataset
utils::data("acme", package = "data.tree")
acme_df <-  data.tree::ToDataFrameTypeCol(acme, acme$attributesAll) %>%
  select(-starts_with("level_"))
# acme2 <- acme$clone()
# acme2$RemoveAttribute("level_3")

attrition_tree <- attrition %>%
  # ensure pure tree
  filter(!(Department %in% c("Sales", "Research_Development") & JobRole == "Manager")) %>%
  tibble::rowid_to_column() %>%
  mutate(pathString = paste("attrition", Department, JobRole, rowid, sep = "/")) %>%
  select(-Department, -JobRole, -rowid) %>%
  data.tree::as.Node()

# Run after all tests
withr::defer(testthat::teardown_env())

#' Plot tabnet_fit model loss along epochs
#'
#' @param object A `tabnet_fit` or `tabnet_pretrain` object as a result of
#' [tabnet_fit()] or [tabnet_pretrain()].
#' @param ...  not used.
#' @return A `ggplot` object.
#' @details
#'  Plot the training loss along epochs, and validation loss along epochs if any.
#'  A dot is added on epochs where model snapshot is available, helping
#'  the choice of `from_epoch` value for later model training resume.
#'
#' @examplesIf (torch::torch_is_installed() && require("modeldata"))
#'  \dontrun{
#' library(ggplot2)
#' data("attrition", package = "modeldata")
#' attrition_fit <- tabnet_fit(Attrition ~. , data=attrition, valid_split=0.2, epoch=11)
#'
#' # Plot the model loss over epochs
#' autoplot(attrition_fit)
#'}
#' @importFrom rlang .data
#'
autoplot.tabnet_fit <- function(object, ...) {

  collect_metrics <- tibble::enframe(object$fit$metrics, name = "epoch") %>%
    tidyr::unnest_wider(value) %>%
    dplyr::mutate_if(is.list, ~purrr::map_dbl(.x, mean)) %>%
    dplyr::select_if(function(x) {!all(is.na(x))} ) %>%
    tidyr::pivot_longer(cols = !dplyr::matches("epoch|checkpoint"),
                        names_to = "dataset", values_to = "loss")


  p <- ggplot2::ggplot(collect_metrics, ggplot2::aes(x = epoch, y = loss, color = dataset)) +
    ggplot2::geom_line() +
    ggplot2::scale_y_log10() +
    ggplot2::guides(colour = ggplot2::guide_legend("Dataset", order=1, override.aes = list(size = 1.7, shape = " ")),
                    size = ggplot2::guide_legend("has checkpoint", order = 2, override.aes = list(size = 3, color = "#F8766D"),
                                                 label.theme = ggplot2::element_text(colour = "#FFFFFF"))) +
    ggplot2::theme(legend.position = "bottom") +
    ggplot2::labs(y="Mean loss (log scale)")

  if ("checkpoint" %in% names(collect_metrics)) {
    checkpoints <- collect_metrics %>%
      dplyr::filter(checkpoint, dataset == "train") %>%
      dplyr::select(-checkpoint) %>%
      dplyr::mutate(size = 2)
    p +
      ggplot2::geom_point(data = checkpoints, ggplot2::aes(x = epoch, y = loss, color = dataset, size = .data$size ))
  } else {
    p
  }
}

#' @rdname autoplot.tabnet_fit
autoplot.tabnet_pretrain <- autoplot.tabnet_fit

#' Plot tabnet_explain mask importance heatmap
#'
#' @param object A `tabnet_explain` object as a result of [tabnet_explain()].
#' @param type a character value. Either `"mask_agg"` the default, for a single
#'  heatmap of aggregated mask importance per predictor along the dataset,
#'   or `"steps"` for one heatmap at each mask step.
#' @param quantile numerical value between 0 and 1. Provides quantile clipping of the
#'  mask values
#' @param ...  not used.
#' @return A `ggplot` object.
#' @details
#'  Plot the `tabnet_explain` object mask importance per variable along the predicted dataset.
#'  `type="mask_agg"` output a single heatmap of mask aggregated values,
#'  `type="steps"` provides a plot faceted along the `n_steps` mask present in the model.
#'  `quantile=.995` may be used for strong outlier clipping, in order to better highlight
#'  low values. `quantile=1`, the default, do not clip any values.
#'
#' @examplesIf (torch::torch_is_installed() && require("modeldata"))
#'  \dontrun{
#' library(ggplot2)
#' data("attrition", package = "modeldata")
#'
#' ## Single-outcome binary classification of `Attrition` in `attrition` dataset
#' attrition_fit <- tabnet_fit(Attrition ~. , data=attrition, epoch=11)
#' attrition_explain <- tabnet_explain(attrition_fit, attrition)
#' # Plot the model aggregated mask interpretation heatmap
#' autoplot(attrition_explain)
#'
#' ## Multi-outcome regression on `Sale_Price` and `Pool_Area` in `ames` dataset,
#' data("ames", package = "modeldata")
#' x <- ames[,-which(names(ames) %in% c("Sale_Price", "Pool_Area"))]
#' y <- ames[, c("Sale_Price", "Pool_Area")]
#' ames_fit <- tabnet_fit(x, y, epochs = 1, verbose=TRUE)
#' ames_explain <- tabnet_explain(ames_fit, x)
#' autoplot(ames_explain, quantile = 0.99)
#' }
autoplot.tabnet_explain <- function(object, type = c("mask_agg", "steps"), quantile = 1, ...) {
  type <- match.arg(type)

  if (type == "steps") {
    .data <- object$masks %>%
      purrr::imap_dfr(~dplyr::mutate(
        .x,
        step = sprintf("Step %d", .y),
        rowname = dplyr::row_number()
      )) %>%
      tidyr::pivot_longer(-c(rowname, step), names_to = "variable", values_to = "mask_agg") %>%
      dplyr::group_by(step) %>%
      dplyr::mutate(mask_agg = quantile_clip(mask_agg, probs=quantile)) %>%
      dplyr::ungroup()
  } else {

  .data <- object$M_explain %>%
    dplyr::mutate(rowname = dplyr::row_number()) %>%
    tidyr::pivot_longer(-rowname, names_to = "variable", values_to = "mask_agg") %>%
    dplyr::mutate(mask_agg = quantile_clip(mask_agg, probs=quantile),
                  step = "mask_aggregate")
  }

  p <- ggplot2::ggplot(.data, ggplot2::aes(x = rowname, y = variable, fill = mask_agg)) +
    ggplot2::geom_tile() +
    ggplot2::scale_fill_viridis_c() +
    ggplot2::facet_wrap(~step) +
    ggplot2::theme_minimal()
  p
}

quantile_clip <- function(x, probs) {
  quantile <- quantile(x, probs = probs)
  purrr::map_dbl(x, ~min(.x, quantile))
}

initialize_non_glu <- function(module, input_dim, output_dim) {
  gain_value <- sqrt((input_dim + output_dim)/sqrt(4*input_dim))
  torch::nn_init_xavier_normal_(module$weight, gain = gain_value)
}

initialize_glu <- function(module, input_dim, output_dim) {
  gain_value <- sqrt((input_dim + output_dim)/sqrt(input_dim))
  torch::nn_init_xavier_normal_(module$weight, gain = gain_value)
}

# Ghost Batch Normalization
# https://arxiv.org/abs/1705.08741
#
gbn <- torch::nn_module(
  "gbn",
  initialize = function(input_dim, virtual_batch_size=128, momentum = 0.02) {
    self$virtual_batch_size <- virtual_batch_size
    self$bn = torch::nn_batch_norm1d(input_dim, momentum = momentum)
  },
  forward = function(x) {
    chunks <- x$chunk(as.integer(ceiling(x$shape[1] / self$virtual_batch_size)), 1)
    res <- lapply(chunks, self$bn)
    torch::torch_cat(res, dim = 1)
  }
)

# Defines main part of the TabNet network without the embedding layers.
#
#
tabnet_encoder <- torch::nn_module(
  "tabnet_encoder",
  initialize = function(input_dim, output_dim,
                        n_d=8, n_a=8,
                        n_steps=3, gamma=1.3,
                        n_independent=2, n_shared=2, epsilon=1e-15,
                        virtual_batch_size=128, momentum = 0.02,
                        mask_type="sparsemax", mask_topk=NULL) {

    self$n_d <- n_d
    self$n_steps <- n_steps
    self$gamma <- gamma
    self$epsilon <- epsilon
    self$mask_topk <- mask_topk # debug purpose
    self$initial_bn <- torch::nn_batch_norm1d(input_dim, momentum = momentum)

    if (n_shared > 0) {
      shared_feat_transform <- torch::nn_module_list(purrr::map(1:n_shared, 
        ~ if (.x == 1) {
          torch::nn_linear(input_dim, 2*(n_d + n_a), bias = FALSE)
        } else {
          torch::nn_linear(n_d + n_a, 2*(n_d + n_a), bias = FALSE)
        }
      ))
    } else {
      shared_feat_transform <- NULL
    }
    
    self$initial_splitter <- feat_transformer(
      input_dim, n_d + n_a, shared_feat_transform,
      n_glu_independent = n_independent,
      virtual_batch_size = virtual_batch_size,
      momentum = momentum
    )
    # stack as many steps as needed of feature_transformer and attentive_transformer
    self$feat_transformers <- torch::nn_module_list(purrr::map(1:n_steps, 
       ~ feat_transformer(input_dim, n_d + n_a, shared_feat_transform,
                        n_glu_independent = n_independent,
                        virtual_batch_size = virtual_batch_size,
                        momentum = momentum)                                                        
    ))
    self$att_transformers <- torch::nn_module_list(purrr::map(1:n_steps, 
      ~ attentive_transformer(n_a, input_dim,
                              virtual_batch_size = virtual_batch_size,
                              momentum = momentum,
                              mask_type = mask_type, 
                              mask_topk = mask_topk)
    ))
  },

  forward = function(x, prior) {
    res <- torch::torch_tensor(0, device = x$device)
    x <- self$initial_bn(x)

    #prior <- torch::torch_ones(size = x$shape, device = x$device)
    M_loss <- 0
    att <- self$initial_splitter(x)[, (self$n_d + 1):N]

    steps_output <- list()
    for (step in seq_len(self$n_steps)) {

      M <- self$att_transformers[[step]](prior, att)
      M_loss <- M_loss + torch::torch_mean(torch::torch_sum(
        torch::torch_mul(M, torch::torch_log(M + self$epsilon)),
        dim = 2
      ))

      # update prior
      prior <- torch::torch_mul(self$gamma - M, prior)

      # output
      masked_x <- torch::torch_mul(M, x)
      out <- self$feat_transformers[[step]](masked_x)
      d <- torch::nnf_relu(out[.., 1:(self$n_d)])
      steps_output[[step]] <- d
      res <- torch::torch_add(res, d)
      # update attention
      att <- out[, (self$n_d + 1):N]

    }

    M_loss <- M_loss / self$n_steps

    list(res, M_loss, steps_output)
  },
  forward_masks = function(x, x_na_mask) {

    x <- self$initial_bn(x)

    prior <- x_na_mask$logical_not()
    M_explain <- torch::torch_zeros(x$shape, device = x$device)
    att <- self$initial_splitter(x)[, (self$n_d+1):N]
    masks <- list()

    for (step in seq_len(self$n_steps)) {

      M <- self$att_transformers[[step]](prior, att)
      masks[[step]] <- M

      # update prior
      prior <- torch::torch_mul(self$gamma - M, prior)

      # output
      masked_x <- torch::torch_mul(M, x)
      out <- self$feat_transformers[[step]](masked_x)
      d <- torch::nnf_relu(out[.., 1:(self$n_d)])

      # explain
      step_importance <- torch::torch_sum(d, dim=2)
      M_explain <- M_explain + torch::torch_mul(M, step_importance$unsqueeze(dim=2))

      # update attention
      att <- out[, (self$n_d+1):N]

    }

    list(M_explain, masks)
  }
)

tabnet_decoder <- torch::nn_module(
  "tabnet_decoder",
  initialize = function(input_dim, n_d=8,
                        n_steps=3, n_independent=2, n_shared=2,
                        virtual_batch_size = 128, momentum = 0.02) {

    self$input_dim <- input_dim
    self$n_d <- n_d
    self$n_steps <- n_steps
    self$n_independent <- n_independent
    self$n_shared <- n_shared
    self$virtual_batch_size <- virtual_batch_size

    self$feat_transformers <- torch::nn_module_list()
    self$reconstruction_layers <- torch::nn_module_list()

    if (self$n_shared > 0) {
      shared_feat_transform <- torch::nn_module_list()

      for (i in seq_len(self$n_shared)) {
        if (i == 1) {
          shared_feat_transform$append(torch::nn_linear(
            n_d, 2 * n_d, bias = FALSE)
          )
        } else {
          shared_feat_transform$append(torch::nn_linear(
            n_d, 2 * n_d, bias = FALSE)
          )
        }
      }

    } else {
      shared_feat_transform <- NULL
    }

    for (step in seq_len(n_steps)) {

      transformer <- feat_transformer(n_d, n_d, shared_feat_transform,
                                      n_glu_independent=self$n_independent,
                                      virtual_batch_size=self$virtual_batch_size,
                                      momentum = momentum)

      self$feat_transformers$append(transformer)
      reconstruction_layer <- torch::nn_linear(n_d, self$input_dim, bias = FALSE)
      initialize_non_glu(reconstruction_layer, n_d, self$input_dim)
      self$reconstruction_layers$append(reconstruction_layer)

    }

  },
  forward = function(steps_output) {
    res <- torch::torch_tensor(0, device = steps_output[[1]]$device)
    for (step_nb in seq_along(steps_output)) {

      x <- self$feat_transformers[[step_nb]](steps_output[[step_nb]])
      x <- self$reconstruction_layers[[step_nb]](steps_output[[step_nb]])
      res <- torch::torch_add(res, x)

    }

    res
  },
)

tabnet_pretrainer <- torch::nn_module(
  "tabnet_pretrainer",
  initialize = function(input_dim, pretraining_ratio = 0.2,
                        n_d = 8, n_a = 8,
                        n_steps = 3, gamma = 1.3,
                        cat_idxs = c(), cat_dims = c(),
                        cat_emb_dim = 1, n_independent = 2,
                        n_shared = 2, n_independent_decoder = 1,
                        n_shared_decoder = 1, epsilon = 1e-15,
                        virtual_batch_size = 128, momentum = 0.02,
                        mask_type = "sparsemax", mask_topk=NULL) {

    # a check par, just to easily find out when we need to
    # reload the model
    self$.check <- torch::nn_parameter(torch::torch_tensor(1, requires_grad = TRUE))

    # 3 vars reused for sum_embedding_masks() within compute_feature_importance()
    self$input_dim <- input_dim
    self$cat_idxs <- cat_idxs
    self$cat_emb_dim <- cat_emb_dim
    self$mask_topk <- mask_topk # debug purpose
    self$initial_bn <- torch::nn_batch_norm1d(input_dim, momentum = momentum)

    if (n_steps <= 0)
      value_error("{.var n_steps} should be a positive integer.")
    if (n_independent == 0 && n_shared == 0)
      value_error("{.var n_shared} and {.var n_independant} can't be both zero.")

    # self$virtual_batch_size <- virtual_batch_size
    self$embedder <- embedding_generator(input_dim, cat_dims, cat_idxs, cat_emb_dim)
    self$embedder_na <- na_embedding_generator(input_dim, cat_dims, cat_idxs, cat_emb_dim)
    self$post_embed_dim <- self$embedder$post_embed_dim
    self$masker = random_obfuscator(pretraining_ratio)
    self$encoder = tabnet_encoder(
      input_dim = self$post_embed_dim,
      output_dim = self$post_embed_dim,
      n_d = n_d,
      n_a = n_a,
      n_steps = n_steps,
      gamma = gamma,
      n_independent = n_independent,
      n_shared = n_shared,
      epsilon = epsilon,
      virtual_batch_size = virtual_batch_size,
      momentum = momentum,
      mask_type = mask_type,
      mask_topk = mask_topk
    )
    self$decoder = tabnet_decoder(
      self$post_embed_dim,
      n_d = n_d,
      n_steps = n_steps,
      n_independent = n_independent_decoder,
      n_shared = n_shared_decoder,
      virtual_batch_size = virtual_batch_size,
      momentum = momentum
    )

  },
  forward = function(x, x_na_mask) {
    embedded_x <- self$embedder(x)
    embedded_x_na_mask <- self$embedder_na(x_na_mask)

    if (self$training) {
      c(masked_x, obfuscated_vars) %<-% self$masker(embedded_x, embedded_x_na_mask)
      # set prior of encoder as !obf_mask
      prior <- obfuscated_vars$logical_not()
      steps_out <- self$encoder(masked_x, prior)[[3]]
      res <- self$decoder(steps_out)
      list(res,
           embedded_x,
           obfuscated_vars)
    } else {
      prior <- embedded_x_na_mask$logical_not()
      steps_out <- self$encoder(embedded_x, prior)[[3]]
      res <- self$decoder(steps_out)
      list(res,
           embedded_x,
           embedded_x_na_mask)
    }
  },
  forward_masks = function(x, x_na_mask) {
    embedded_x <- self$embedder(x)
    embedded_x_na_mask <- self$embedder_na(x_na_mask)
    self$encoder$forward_masks(embedded_x, embedded_x_na_mask)
  }

)

tabnet_no_embedding <- torch::nn_module(
  "tabnet_no_embedding",
  initialize = function(input_dim, output_dim,
                        n_d=8, n_a=8,
                        n_steps=3, gamma=1.3,
                        n_independent=2, n_shared=2, epsilon=1e-15,
                        virtual_batch_size=128, momentum = 0.02,
                        mask_type="sparsemax", mask_topk=NULL) {

    self$is_multi_outcome <- !is.atomic(output_dim)

    self$mask_topk <- mask_topk # debug purpose
    self$initial_bn <- torch::nn_batch_norm1d(input_dim, momentum = momentum)

    self$encoder <- tabnet_encoder(
      input_dim = input_dim,
      output_dim = output_dim,
      n_d = n_d,
      n_a = n_a,
      n_steps = n_steps,
      gamma = gamma,
      n_independent = n_independent,
      n_shared = n_shared,
      epsilon = epsilon,
      virtual_batch_size = virtual_batch_size,
      momentum = momentum,
      mask_type = mask_type,
      mask_topk = mask_topk
    )
    if (self$is_multi_outcome) {
      self$multi_outcome_mapping <- torch::nn_module_list()
      for (task_dim in output_dim) {
        task_mapping <- torch::nn_linear(n_d, task_dim, bias = FALSE)
        initialize_non_glu(task_mapping, n_d, task_dim)
        self$multi_outcome_mapping$append(task_mapping)
      }
    } else {
      self$final_mapping <- torch::nn_linear(n_d, sum(output_dim), bias = FALSE)
      initialize_non_glu(self$final_mapping, n_d, sum(output_dim))
    }

  },
  forward = function(x, x_na_mask) {
    prior <- x_na_mask$logical_not()
    c(res, M_loss, steps_output) %<-% self$encoder(x, prior)
    if (self$is_multi_outcome) {
      out <- torch::torch_stack(purrr::map(self$multi_outcome_mapping, exec, !!!res), dim = 2)$squeeze(3)
    } else {
      out <- self$final_mapping(res)
    }

    list(out, M_loss)
  },
  forward_masks = function(x, x_na_mask) {
    self$encoder$forward_masks(x, x_na_mask)
    }
)


#' TabNet Model Architecture
#'
#' This is a `nn_module` representing the TabNet architecture from
#' [Attentive Interpretable Tabular Deep Learning](https://arxiv.org/abs/1908.07442).
#'
#' @param input_dim Initial number of features.
#' @param output_dim Dimension of network output. Examples : one for regression, 2 for
#'   binary classification etc.. Vector of those dimensions in case of multi-output.
#' @param n_d Dimension of the prediction layer (usually between 4 and 64).
#' @param n_a Dimension of the attention layer (usually between 4 and 64).
#' @param n_steps Number of successive steps in the network (usually between 3 and 10).
#' @param gamma Scaling factor for attention updates (usually between 1 and 2).
#' @param cat_idxs Index of each categorical column in the dataset.
#' @param cat_dims Number of categories in each categorical column.
#' @param cat_emb_dim Size of the embedding of categorical features if int, all categorical
#'   features will have same embedding size if list of int, every corresponding feature will have
#'   specific size.
#' @param n_independent Number of independent GLU layer in each GLU block of the encoder.
#' @param n_shared Number of shared GLU layer in each GLU block of the encoder.
#' @param epsilon Avoid log(0), this should be kept very low.
#' @param virtual_batch_size Batch size for Ghost Batch Normalization.
#' @param momentum  Numerical value between 0 and 1 which will be used for momentum in all batch norm.
#' @param mask_type Either "sparsemax", "entmax" or "entmax15": the sparse masking function to use.
#' @param mask_topk the mask top-k value for k-sparsity selection in the mask for `sparsemax` and `entmax15`.
#'  defaults to 1/4 of last `input_dim` if `NULL`. See [entmax15] for details.
#' @export
tabnet_nn <- torch::nn_module(
  "tabnet",
  initialize = function(input_dim, output_dim, n_d=8, n_a=8,
                        n_steps=3, gamma=1.3, cat_idxs=c(), cat_dims=c(), cat_emb_dim=1,
                        n_independent=2, n_shared=2, epsilon=1e-15,
                        virtual_batch_size = 128, momentum = 0.02,
                        mask_type="sparsemax", mask_topk=NULL) {

    # a check par, just to easily find out when we need to reload the model
    self$.check <- torch::nn_parameter(torch::torch_tensor(1, requires_grad = TRUE))
    
    # 3 vars reused for sum_embedding_masks()
    self$input_dim <- input_dim
    self$cat_idxs <- cat_idxs
    self$cat_emb_dim <- cat_emb_dim
    
    if (n_steps <= 0)
      value_error("{.var n_steps} should be a positive integer.")
    if (n_independent == 0 && n_shared == 0)
      value_error("{.var n_shared} and {.var n_independant} can't be both zero.")
    
    self$virtual_batch_size <- virtual_batch_size
    self$embedder <- embedding_generator(input_dim, cat_dims, cat_idxs, cat_emb_dim)
    self$embedder_na <- na_embedding_generator(input_dim, cat_dims, cat_idxs, cat_emb_dim)
    self$post_embed_dim <- self$embedder$post_embed_dim
    self$tabnet <- tabnet_no_embedding(self$post_embed_dim, output_dim, n_d, n_a, n_steps,
                                     gamma, n_independent, n_shared, epsilon,
                                     virtual_batch_size, momentum, mask_type, mask_topk)

  },
  forward = function(x, x_na_mask) {
    embedded_x <- self$embedder(x)
    embedded_x_na_mask <- self$embedder_na(x_na_mask)
    self$tabnet(embedded_x, embedded_x_na_mask)
  },
  forward_masks = function(x, x_na_mask) {
    embedded_x <- self$embedder(x)
    embedded_x_na_mask <- self$embedder_na(x_na_mask)
    self$tabnet$forward_masks(embedded_x, embedded_x_na_mask)
  }
)

attentive_transformer <- torch::nn_module(
  "attentive_transformer",
  initialize = function(input_dim, output_dim,
                        virtual_batch_size = 128,
                        momentum = 0.02,
                        mask_type="sparsemax", 
                        mask_topk = NULL) {
    self$mask_topk <- mask_topk # debug purpose
    self$fc <- torch::nn_linear(input_dim, sum(output_dim), bias=FALSE)
    initialize_non_glu(self$fc, input_dim, sum(output_dim))
    self$bn <- gbn(sum(output_dim), virtual_batch_size=virtual_batch_size,
                  momentum = momentum)


    if (mask_type == "sparsemax15") {
      if (is.null(mask_topk)) {
        mask_topk <- round(input_dim[-1L] / 4)
      } else {
        mask_topk <- as.integer(mask_topk)
      }
      self$selector <- sparsemax15(dim = -1L, k = self$mask_topk)
    } else if (mask_type == "entmax15") {
      if (is.null(mask_topk)) {
        mask_topk <- round(input_dim[-1L] / 4)
      } else {
        mask_topk <- as.integer(mask_topk)
      }
      self$selector <- entmax15(dim = -1L, k = self$mask_topk)
    } else if (mask_type == "entmax")
      self$selector <- entmax(dim = -1L)
    else if (mask_type == "sparsemax")
      self$selector <- sparsemax(dim = -1L)
    else
      value_error("Please choose either {.val sparsemax}, {.val sparsemax15}, {.val entmax} or {.val entmax15} as {.var mask_type}")
  },
  forward = function(priors, processed_feat) {
    x <- self$fc(processed_feat)
    x <- self$bn(x)
    x <- torch::torch_mul(x, priors)
    x <- self$selector(x)
    x
  }
)

feat_transformer <- torch::nn_module(
  "feat_transformer",
  initialize = function(input_dim, output_dim, shared_layers, n_glu_independent,
                        virtual_batch_size=128, momentum = 0.02) {

    params <- list(
      'n_glu'= n_glu_independent,
      'virtual_batch_size'= virtual_batch_size,
      'momentum'= momentum
    )

    if (is.null(shared_layers)) {
      self$shared <- torch::nn_identity()
      is_first <- TRUE
    } else {
      self$shared <- glu_block(input_dim, output_dim,
                               first=TRUE,
                               shared_layers=shared_layers,
                               n_glu=length(shared_layers),
                               virtual_batch_size=virtual_batch_size,
                               momentum = momentum)
      is_first <- FALSE
    }

    if (n_glu_independent == 0) {
      self$specifics <- torch::nn_identity()
    } else {
      if (is_first)
        spec_input_dim <- input_dim
      else
        spec_input_dim <- output_dim
      self$specifics <- do.call(glu_block, append(
        list(spec_input_dim, output_dim, first = is_first),
        params
      ))
    }
  },
  forward = function(x) {
    x <- self$shared(x)
    x <- self$specifics(x)
    x
  }
)

glu_block <- torch::nn_module(
  "glu_block",
  initialize = function(input_dim, output_dim, n_glu=2, first=FALSE,
                        shared_layers=NULL,
                        virtual_batch_size=128, momentum = 0.02) {

    self$first <- first
    self$shared_layers <- shared_layers
    self$n_glu <- n_glu
    self$glu_layers <- torch::nn_module_list()

    params = list(
      'virtual_batch_size' = virtual_batch_size,
      'momentum' = momentum
    )

    if (length(shared_layers) > 0)
      fc <- shared_layers[[1]]
    else
      fc <- NULL

    self$glu_layers$append(do.call(
      glu_layer,
      append(list(input_dim, output_dim, fc = fc), params)
    ))

    if (self$n_glu >= 2) {
      for (glu_id in 2:(self$n_glu)) {

        if (length(shared_layers) > 0)
          fc <- shared_layers[[glu_id]]
        else
          fc <- NULL

        self$glu_layers$append(do.call(
          glu_layer,
          append(list(output_dim, output_dim, fc = fc), params)
        ))

      }
    }

  },
  forward = function(x) {
    scale <- torch::torch_sqrt(torch::torch_tensor(0.5, device = x$device))

    if (self$first) {
      x <- self$glu_layers[[1]](x)
      layers_left <- seq_len(self$n_glu)[-1]
    } else {
      layers_left <- seq_len(self$n_glu)
    }

    for (glu_id in layers_left) {
      x <- torch::torch_add(x, self$glu_layers[[glu_id]](x))
      x <- x*scale
    }

    x
  }
)

glu_layer <- torch::nn_module(
  "glu_layer",
  initialize = function(input_dim, output_dim, fc=NULL,
                        virtual_batch_size=128, momentum = 0.02) {
    self$output_dim <- sum(output_dim)

    if (!is.null(fc))
      self$fc <- fc
    else
      self$fc <- torch::nn_linear(input_dim, 2 * self$output_dim, bias = FALSE)

    initialize_glu(self$fc, input_dim, 2 * self$output_dim)

    self$bn <- gbn(2 * self$output_dim, virtual_batch_size = virtual_batch_size,
                  momentum = momentum)
  },
  forward = function(x) {
    x <- self$fc(x)
    x <- self$bn(x)
    out <- torch::torch_mul(
      x[, 1:self$output_dim],
      torch::torch_sigmoid(x[, (self$output_dim + 1):N])
    )
    out
  }
)

embedding_generator <- torch::nn_module(
  "embedding_generator",
  initialize = function(input_dim, cat_dims, cat_idxs, cat_emb_dim) {

    if (length(cat_dims) == 0 || length(cat_idxs) == 0) {
      self$skip_embedding <- TRUE
      self$post_embed_dim <- input_dim
      return(invisible(NULL))
    }

    self$skip_embedding <- FALSE
    self$cat_dims <- cat_dims

    if (length(cat_emb_dim) == 1)
      self$cat_emb_dims <- rep(cat_emb_dim, length(cat_idxs))
    else
      self$cat_emb_dims <- cat_emb_dim

    # check that all embeddings dimensions are provided
    if (length(self$cat_emb_dims) != length(cat_dims)){
      value_error("{.var cat_emb_dim} length must be 1 or the number of categorical predictors, 
                  got length {.val {length(self$cat_emb_dims)}} for {.val {length(cat_dims)}} 
                  categorical predictors")
    }

    self$post_embed_dim <- as.integer(input_dim + sum(self$cat_emb_dims) - length(self$cat_emb_dims))
    self$embeddings <- torch::nn_module_list()

    # Sort dims by cat_idx
    sorted_idx <- order(cat_idxs)
    cat_dims <- cat_dims[sorted_idx]
    self$cat_emb_dims <- self$cat_emb_dims[sorted_idx]

    for (i in seq_along(cat_dims)) {
      self$embeddings$append(
        torch::nn_embedding(
          cat_dims[i],
          self$cat_emb_dims[i]
        )
      )
    }

    # record continuous indices
    self$continuous_idx <- rep(TRUE, input_dim)
    self$continuous_idx[cat_idxs] <- FALSE

  },
  forward = function(x) {

    if (self$skip_embedding) {
      # no embeddings required
      return(x)
    }

    cols <- list()
    cat_feat_counter <- 1L

    for (i in seq_along(self$continuous_idx)) {

      if (self$continuous_idx[i]) {
        # impute nan with 0s
        cols[[i]] <- x[,i]$nan_to_num(0)$to(dtype = torch::torch_float())$view(c(-1, 1))
      } else {
        # nan mask
        mask <- x[, i]$ge(1)$bitwise_and(x[, i]$le(self$cat_dims[cat_feat_counter]))$to(dtype = torch::torch_long())
        # impute nan with 1s (categorical vars are 1-indexed)
        # cols[[i]] <- self$embeddings[[cat_feat_counter]](x[, i]$nan_to_num(1)$to(dtype = torch::torch_long()))
        # obsolete : requires a masking + imputation to impute all nans with torch 0.11
        imputed_xi <- (x[, i]$mul(mask) + (1 - mask))$nan_to_num(1)
        cols[[i]] <- self$embeddings[[cat_feat_counter]](imputed_xi$to(dtype = torch::torch_long()))
        cat_feat_counter <- cat_feat_counter + 1L
      }

    }

    # concat
    post_embeddings <- torch::torch_cat(cols, dim=2)
    post_embeddings
  }
)

na_embedding_generator <- torch::nn_module(
  "na_embedding_generator",
  initialize = function(input_dim, cat_dims, cat_idxs, cat_emb_dim) {

    if (length(cat_dims) == 0 || length(cat_idxs) == 0) {
      self$skip_embedding <- TRUE
      return(invisible(NULL))
    }

    self$skip_embedding <- FALSE

    if (length(cat_emb_dim) == 1)
      self$cat_emb_dims <- rep(cat_emb_dim, length(cat_idxs))
    else
      self$cat_emb_dims <- cat_emb_dim

    # Sort dims by cat_idx
    sorted_idx <- order(cat_idxs)
    self$cat_emb_dims <- self$cat_emb_dims[sorted_idx]

    # record continuous indices
    self$continuous_idx <- rep(TRUE, input_dim)
    self$continuous_idx[cat_idxs] <- FALSE

  },
  forward = function(x) {

    if (self$skip_embedding) {
      # no embeddings required
      return(x)
    }

    cols <- list()
    cat_feat_counter <- 1

    for (i in seq_along(self$continuous_idx)) {

      if (self$continuous_idx[i]) {
        cols[[i]] <- x[,i]$to(dtype = torch::torch_bool())$view(c(-1, 1))
      } else {
        # extend the vector to match the dimention of the embedding_x via matmul
        # TODO basic tensor broadcasting function could be more efficient and have lower footprint
        cols[[i]] <- x[,i:i]$logical_and(torch::torch_ones(1,self$cat_emb_dims[cat_feat_counter], device = x$device))
        cat_feat_counter <- cat_feat_counter + 1
      }

    }

    # concat
    post_embeddings <- torch::torch_cat(cols, dim=2)
    post_embeddings
  }
)

random_obfuscator <- torch::nn_module(
  "random_obfuscator",
  initialize = function(pretraining_ratio) {

    if (pretraining_ratio <= 0 || pretraining_ratio >= 1) {
      pretraining_ratio <- 0.5
    }

    self$pretraining_ratio <- pretraining_ratio

  },
  forward = function(x, x_na_mask) {
    # workaround while torch_bernoulli is not available in CUDA
    ones <- torch::torch_ones_like(x, device="cpu")
    obfuscated_vars <- torch::torch_bernoulli(self$pretraining_ratio * ones)$to(device=x$device)
    masked_input = torch::torch_mul(obfuscated_vars$logical_or(x_na_mask)$logical_not(), x)

    list(masked_input, obfuscated_vars)

  }
)

# as per https://github.com/mlverse/torch/blob/main/tests/testthat/helper-tensor.R
Sys.setenv(KMP_DUPLICATE_LIB_OK = TRUE)
# torch_zeros(1, names="hello") # trigger warning about named tensors

skip_if_not_test_examples <- function() {
  if (Sys.getenv("TEST_EXAMPLES", unset = "0") != "1") {
    skip("Not testing examples/readme. Set the env var TEST_EXAMPLES = 1.")
  }
}

skip_if_cuda_not_available <- function() {
  if (!cuda_is_available()) {
    skip("A GPU is not available for testing.")
  }
}

skip_if_not_m1_mac <- function() {
  if (!grepl("darwin", R.version$os)) {
    skip("Not on MacOS")
  }

  if (R.version$arch != "aarch64") {
    skip("Not an M1 Mac")
  }
}

expect_equal_to_tensor <- function(object, expected, ...) {
  expect_equal(torch::as_array(object), torch::as_array(expected), ...)
}

expect_not_equal_to_tensor <- function(object, expected) {
  expect_false(isTRUE(all.equal(torch::as_array(object), torch::as_array(expected))))
}

expect_no_error <- function(object, ...) {
  expect_error(object, NA, ...)
}

expect_tensor <- function(object) {
  expect_true(torch:::is_torch_tensor(object))
  expect_no_error(torch::as_array(object$to_dense()$to(device = "cpu")))
}

expect_equal_to_r <- function(object, expected, ...) {
  expect_equal(torch::as_array(object$cpu()), expected, ...)
}

expect_tensor_shape <- function(object, expected) {
  expect_tensor(object)
  expect_equal(object$shape, expected)
}


expect_tensor_dtype <- function(object, expected_dtype) {
  expect_tensor(object)
  expect_true(object$dtype == expected_dtype)
}


expect_undefined_tensor <- function(object) {
  # TODO
}

expect_identical_modules <- function(object, expected) {
  expect_identical(
    attr(object, "module"),
    attr(expected, "module")
  )
}

#!/bin/bash

# stop if error
set -e

read -p 'Release version: ' version
echo ${version} | grep v && echo "Version should be x.y.z (for example, 1.1.1, 2.0.0, ...)" && exit -1

localDir=`readlink -f .`
releaseDir="${localDir}/release-${version}"
rm -rf ${releaseDir}
mkdir ${releaseDir}
cd $releaseDir

echo "Cloning repo into tabnet"
git clone -q git@github.com:dreamquark-ai/tabnet.git tabnet

cd tabnet
# Create release branch and push it
git checkout -b release/${version}
# Change version of package
docker run --rm -v ${PWD}:/work -w /work tabnet:latest poetry version ${version}
# Generate docs
make install doc
# Add modified files
git add pyproject.toml docs/
# Commit release
git commit -m "chore: release v${version}"
# Create tag for changelog generation
git tag v${version}
docker run -v ${PWD}:/work -w /work --entrypoint "" release-changelog:latest git config --global --add safe.directory /work &&\
 conventional-changelog -p angular -i CHANGELOG.md -s -r 0 && \
 chmod 777 CHANGELOG.md
# Removing 4 first line of the file
echo "$(tail -n +4 CHANGELOG.md)" > CHANGELOG.md
# Deleting tag
git tag -d v${version}
# Adding CHANGELOG to commit
git add CHANGELOG.md
git commit --amend --no-edit
# Push release branch
git push origin release/${version}

cd ${localDir}
sudo rm -rf ${releaseDir}

#!/bin/bash

# stop if error
set -e

read -p 'Release version: ' version
read -p 'You personal access token for Github: ' token
read -p 'Your username for pipy: ' pipyUser
read -p 'Your password for pipy: ' pipyPassword

echo ${version} | grep v && echo "Version should be x.y.z (for example, 1.1.1, 2.0.0, ...)" && exit -1

localDir=`readlink -f .`
releaseDir="${localDir}/release-${version}"
rm -rf ${releaseDir}
mkdir ${releaseDir}
cd $releaseDir

echo "Cloning repo into tabnet"
git clone -q git@github.com:dreamquark-ai/tabnet.git tabnet

cd tabnet
# Create release branch and push it
git checkout release/${version}

# Tagging proper version
echo "Tagging proper version"
git tag v${version}

# Build release
echo "Building latest build"
docker run --rm -v ${PWD}:/work -w /work tabnet:latest poetry build

echo "Merging into develop and master"
git checkout master
git merge --no-ff origin/release/${version} -m "chore: release v${version} (merge)"
git checkout develop
git merge --no-ff origin/release/${version} -m "chore: release v${version} (merge)"

echo "Pushing branch"
git push origin develop
git push origin master
echo "Pushing tag"
git push origin --tags

echo "Making github release"
docker run -v ${PWD}:/work -w /work --entrypoint "" release-changelog:latest conventional-github-releaser -p angular --token ${token}

# Build release
echo "Building latest build"
docker run --rm -v ${PWD}:/work -w /work tabnet:latest poetry build
# Build release
echo "Publishing latest build"
docker run --rm -v ${PWD}:/work -w /work tabnet:latest poetry publish -u ${pipyUser} -p ${pipyPassword}

echo "Deleting release branch"
git checkout develop
git push origin :release/${version}

cd ${localDir}
rm -rf ${releaseDir}

#!/bin/bash
set -e

SCRIPT_DIR=$(dirname ${BASH_SOURCE[0]})
DOC_FOLDER_NAME="source/generated_docs"
DOC_FOLDER="${SCRIPT_DIR}/${DOC_FOLDER_NAME}"
TARGET_FOLDER="../pytorch_tabnet"

rm -rf ${DOC_FOLDER}
mkdir -p ${DOC_FOLDER}

MODULES_LIST=$(find ${SCRIPT_DIR}/${TARGET_FOLDER} -type f 2>/dev/null | grep -v \.cache | grep -v "/tests" |  grep -e '.py$' | sed -e 's/^[\.+/]*\///g' | sed -e 's/\.py$//g' | sed -e 's/\//./g') 
PACKAGE_LIST=$(echo ${MODULES_LIST} | tr ' ' '\n' | sed -e 's/^\(.*\)\.\(.*\)/\1/g' | sort | uniq)

#echo $(seq -s= $(($(echo pytorch_tabnet.sparsemax|wc -c) + 9))|tr -d '[:digit:]')

echo ${PACKAGE_LIST} | tr ' ' '\n' | xargs -n1 -I {} echo "cp ${SCRIPT_DIR}/package_template.tpl ${DOC_FOLDER}/{}.rst" | sh
echo ${PACKAGE_LIST} | tr ' ' '\n' | xargs -n1 -I {} echo "sed -si \"s/#PACKAGE_NAME#/{} package\n\$(seq -s= \$((\$(echo {}|wc -c) + 8))|tr -d '[:digit:]')/g\" ${DOC_FOLDER}/{}.rst" | sh

# module
echo ${MODULES_LIST} |\
 tr ' ' '\n' |\
 xargs -n1 -I {} echo "cat ${SCRIPT_DIR}/module_template.tpl | sed -s \"s/#MODULE_NAME_TITLE#/{} module\n\$(seq -s. \$((\$(echo {}|wc -c) + 9))|tr -d '[:digit:]')/g\" | sed -s \"s/#MODULE_NAME#/{}/g\" >> \$(echo ${DOC_FOLDER}/{} | sed -e 's/^\(.*\)\.\(.*\)/\1.rst/g')" | sh
cp "${SCRIPT_DIR}/../README.md" "${DOC_FOLDER}/README.md"

---
title: "Interpretation tools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Interpretation tools}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = torch::torch_is_installed(),
  out.width = "100%", 
  out.height = "300px", 
  fig.width = 14
)
```

```{r setup}
library(tabnet)
library(tidyverse, warn.conflicts = FALSE)
set.seed(1)
# You may want to add this for complete reproducibility
# torch::torch_manual_seed(1)
```

TabNet is one of the first deep-learning interpretable model. Thanks to the underlying 
neural-network architecture, TabNet uses feature-selection masks that can help identify
which predictor is used at each step.

The paper also defines an aggregate measure that combines the masks of
each step into a single measure.

Note that the selection of important features in the masks is done
instance-wise. Thus, you can identify, for each observation, which
predictors were considered relevant.

## Experiments

To show how to use the interpretation tools of `tabnet`, we are going to
perform 2 experiments using synthetic datasets very similar to those
that were used in the paper.

### Datasets

First, let's define the functions that we will use to generate data:

-   `make_syn2` will generate a dataset with 10 columns, but only
    columns 3-6 are used to calculate the `y` response vector. This is
    similar to *Syn2* in the paper.

-   `make_syn4` will generate a dataset with 10 columns too. The
    response vector depends on column 10: if the value is greater than
    0, we use columns 1-2 to compute the logits, otherwise we use
    columns 5-6.

```{r}
logit_to_y <- function(logits) {
  p <- exp(logits)/(1 + exp(logits))
  y <- factor(ifelse(p > 0.5, "yes", "no"), levels = c("yes", "no"))
  y
}

make_random_x <- function(n) {
  x <- as.data.frame(lapply(1:10, function(x) rnorm(n)))
  names(x) <- sprintf("V%02d", 1:10)
  x
}

make_syn2 <- function(n = 5000) {
  x <- make_random_x(n)
  logits <- rowSums(x[,3:6])
  x$y <- logit_to_y(logits)
  x
}

make_syn4 <- function(n = 5000) {
  x <- make_random_x(n)
  logits <- ifelse(
    x[,10] > 0,
    rowSums(x[,1:2]),
    rowSums(x[,5:6])
  )
  
  x$y <- logit_to_y(logits)
  x
}
```

Now let's generate the datasets:

```{r}
syn2 <- make_syn2()
syn4 <- make_syn4()
```

### Syn2

Let's fit a TabNet model to the `syn2` dataset and analyze the
interpretation metrics.

```{r}
fit_syn2 <- tabnet_fit(y ~ ., syn2, epochs = 45, learn_rate = 0.06, device = "cpu")
```

In the feature importance plot we can see that, as expected, features
`V03-V06` are by far the most important ones.

```{r}
#| fig.alt: "A variable importance plot of the fitted model on syn2 dataset showing V03 then V06, V04, v10 and V5 as the 5 most important features, in that order."
vip::vip(fit_syn2)
```

Now let's visualize the aggregated-masks plot. In this figure we see
each observation on the x axis and each variable on the y axis. The
colors represent the importance of the feature in predicting the value
for each observation.

```{r}
#| fig.alt: "A tabnet explaination plot of the fitted model on syn2 dataset. The plot shows numerous important observations in V03 then V06 and V04. the other variables are shown with low importance points or sparse observations with importance"
library(tidyverse)
ex_syn2 <- tabnet_explain(fit_syn2, syn2)

autoplot(ex_syn2, quantile = 0.99)
```

We can see that the region between V03 and V06 concentrates most of the
higher intensity colors, and the other variables are close to 0. This is
expected because those are the variables that we considered when
building the dataset.

Next, we can visualize the attention masks for each step in the
architecture.

```{r}
#| fig.alt: "3 tabnet explaination plots, one for each step of the fitted model on syn2 dataset. The Step 1 plot shows numerous important observations with V02 and V03 having high importance, step 2 plot highlight the importance of V03 and V06. Third step plot highlight V03 and V10 as important variables"
autoplot(ex_syn2, type="steps")
```

We see that the first step captures a lot of noise, but the other 2
steps focus specifically on the important features.

## Syn 4

Now let's analyze the results for the Syn4 dataset. This dataset is a
little more complicated for TabNet because there's a strong interaction
between the variables. Depending on V10, different variables are used to
create the response variable and we expect to see this in the masks.

First we fit the model for 10 epochs.

```{r}
fit_syn4 <- tabnet_fit(y ~ ., syn4, epochs = 50, device = "cpu", learn_rate = 0.08)
```

In the feature importance plot we have, as expected, strong importance
for `V10`, and the other features that are used conditionally - either
`V01-V02` or `V05-V06`.

```{r}
#| fig.alt: "A variable importance plot of the fitted model on syn4 dataset"
vip::vip(fit_syn4)
```

Now let's visualize the attention masks. Notice that we arranged the
dataset by `V10` so we can easily visualize the interaction effects.

We also trimmed to the 98th percentile so the colors shows the
importance even if there are strong outliers.

```{r}
#| fig.alt: "A tabnet explaination plot of the fitted model on syn4 dataset. The plot shows numerous important observations in V06 for low values of V10, and importance of V01 and V02 for high values of V10."
ex_syn4 <- tabnet_explain(fit_syn4, arrange(syn4, V10))

autoplot(ex_syn4, quantile=.98)
```

From the figure we see that V10 is important for all observations. We
also see that for the first half of the dataset `V05` and `V06` is the
most important feature, while for the other half, `V01` and `V02` are
the important ones.

We can also visualize the masks at each step in the architecture.

```{r}
#| fig.alt: "3 tabnet explaination plots, one for each step of the fitted model on syn4 dataset."
autoplot(ex_syn4, type="steps", quantile=.995)
```

We see that step 1 and 3 both focus on `V10`, but on different
additional features, depending on `V10`. Step 2 seems to have found some
noise in `V08`, but also focuses strongly on `V01-V02` and `V05-V06`.

---
name: Bug report
about: Create a report to help us improve
title: ''
labels: bug
assignees: eduardocarvp, Hartorn, j-abi, Optimox

---

<!-- Please don't delete this template or we'll close your issue -->
**Describe the bug**
<!-- A clear and concise description of what the bug is. -->

**What is the current behavior?**

**If the current behavior is a bug, please provide the steps to reproduce.**
<!--
1. Go to '...'
2. Click on '....'
3. Scroll down to '....'
4. See error
-->

**Expected behavior**
<!-- "It should work" is not a good explanation -->
<!-- Explain how exactly you expecting it to behave -->

**Screenshots**
<!-- If applicable, add screenshots to help explain your problem. -->

**Other relevant information:**
poetry version:  
python version: 
Operating System: 
Additional tools:

**Additional context**
<!-- Add any other context about the problem here. -->

---
title: "Self-supervised training and fine-tuning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Self-supervised training and fine-tuning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(tabnet)
library(tidymodels)
library(modeldata)
library(ggplot2)
```

In this vignette we show how to 
 - pretrain TabNet model with a self-supervised task on unlabeled data
 - fine-tune the pretrained TabNet model with labeled data

We are going to use the `lending_club` dataset available
in the `modeldata` package, using 90 % of it as unlabeled data.

First, let's split our dataset into `unlabeled` and `labeled` datasets, so we can later on train the supervised step of the model:

```{r}
set.seed(123)
data("lending_club", package = "modeldata")
split <- initial_split(lending_club, strata = Class, prop = 9/10)
unlabeled <- training(split) %>% mutate(Class=factor(NA))
labeled  <- testing(split)
```

Then we proceed with the usual random split of the labeled dataset into `train` and `test` so that we can evaluate performance of our model:

```{r}
set.seed(123)
labeled_split <- initial_split(labeled, strata = Class)
train <- training(labeled_split) 
test  <- testing(labeled_split)
```

## Data preprocessing

We now define our pre-processing steps in a recipe.  
The `recipe()` here is defined on the whole dataset in order to capture the full variability of numerical variables, as well as all categories present in nominal variables.  
Note that tabnet handles categorical variables, so we don't need to do any kind of transformation to them. Normalizing the numeric variables is a good idea though.

```{r}
rec <- recipe(Class ~ ., lending_club) %>%
  step_normalize(all_numeric())
unlabeled_baked_df <- rec %>% prep %>% bake(new_data=unlabeled)
```

We now have the normalized dataset ready for self-supervised training. 

## Self-supervised training step  

Next, we pre-train our model with a [autoassociative self-supervised learning](https://en.wikipedia.org/wiki/Self-supervised_learning#Autoassociative_self-supervised_learning) task. This step will gives us a `tabnet_pretrain` object that will contain a representation of the dataset variables and their interactions in its latent space.  
We are going to train for 50 epochs with a batch size of 5000 i.e. half of the dataset because it is small enough to fit into memory. There are other hyperparameters available, but we are going to use the default values here.

```{r}
mod <- tabnet_pretrain(rec, unlabeled, epochs = 50, valid_split = 0.2, batch_size = 5000, verbose = TRUE)
```
```
[Epoch 001] Loss: 4.195504 Valid loss: 1.713710                                                                                    
[Epoch 002] Loss: 4.589761 Valid loss: 1.474615                                                                                    
[Epoch 003] Loss: 1.812671 Valid loss: 1.410668                                                                                    
[Epoch 004] Loss: 1.553153 Valid loss: 1.359872                                                                                    
[Epoch 005] Loss: 2.345665 Valid loss: 1.299493                                                                                    
[Epoch 006] Loss: 2.719557 Valid loss: 1.258007                                                                                    
[Epoch 007] Loss: 1.285357 Valid loss: 1.233827                                                                                    
[Epoch 008] Loss: 1.283360 Valid loss: 1.210789                                                                                    
[Epoch 009] Loss: 1.452972 Valid loss: 1.194631                                                                                    
[Epoch 010] Loss: 1.256993 Valid loss: 1.181094                                                                                    
[Epoch 011] Loss: 1.327342 Valid loss: 1.158956                                                                                    
[Epoch 012] Loss: 1.258828 Valid loss: 1.145682                                                                                    
[Epoch 013] Loss: 1.130475 Valid loss: 1.130623
...
[Epoch 041] Loss: 1.002896 Valid loss: 0.950189                                                                                    
[Epoch 042] Loss: 1.142027 Valid loss: 0.944114                                                                                    
[Epoch 043] Loss: 0.940986 Valid loss: 0.940836                                                                                    
[Epoch 044] Loss: 1.032234 Valid loss: 0.939989                                                                                    
[Epoch 045] Loss: 0.947644 Valid loss: 0.937613                                                                                    
[Epoch 046] Loss: 1.007923 Valid loss: 0.935926                                                                                    
[Epoch 047] Loss: 1.721710 Valid loss: 0.938697                                                                                    
[Epoch 048] Loss: 0.984387 Valid loss: 0.941066                                                                                    
[Epoch 049] Loss: 1.092131 Valid loss: 0.944751                                                                                    
[Epoch 050] Loss: 1.175343 Valid loss: 0.947859 
```

After a few minutes we can get the results:

```{r}
autoplot(mod)
```
![pretraining task loss plot](pretraining_loss.png)
We may continue the training for few epoch but the validation loss seems to plateau so it is like the model would start to overfit...
Here we can identify on the plot that the model epoch providing the lowest loss on the validation-set is around epoch 35. We will reuse it  through the `from_epoch` option for training continuation, instead of the default being last epoch.   
But historical model weights are only available for checkpoint-ed epoch , so in our case, checkpoint at epoch 40 seems a good compromise, and we would then use `from_epoch=40`.

## Continuing training with supervised task

Now, we reuse our pre-processing steps recipe and feed it directly in a supervised fitting task on top of our pre-trained model. 

```{r}
model_fit <- tabnet_fit(rec, train , tabnet_model = mod, from_epoch=40, valid_split = 0.2, epochs = 50, verbose=TRUE)
```
```
[Epoch 041] Loss: 0.260110 Valid loss: 0.250054                                                                                    
[Epoch 042] Loss: 0.213628 Valid loss: 0.219784                                                                                    
[Epoch 043] Loss: 0.154529 Valid loss: 0.201305                                                                                    
[Epoch 044] Loss: 0.187461 Valid loss: 0.202042                                                                                    
[Epoch 045] Loss: 0.144463 Valid loss: 0.196665                                                                                    
[Epoch 046] Loss: 0.191665 Valid loss: 0.196853                                                                                    
[Epoch 047] Loss: 0.186484 Valid loss: 0.190732                                                                                    
[Epoch 048] Loss: 0.166680 Valid loss: 0.188041                                                                                    
[Epoch 049] Loss: 0.134771 Valid loss: 0.180121                                                                                    
[Epoch 050] Loss: 0.127854 Valid loss: 0.168102   
...
[Epoch 082] Loss: 0.101062 Valid loss: 0.262795                                                                                    
[Epoch 083] Loss: 0.097425 Valid loss: 0.280869                                                                                    
[Epoch 084] Loss: 0.113473 Valid loss: 0.296119                                                                                    
[Epoch 085] Loss: 0.094036 Valid loss: 0.303786                                                                                    
[Epoch 086] Loss: 0.092469 Valid loss: 0.309182                                                                                    
[Epoch 087] Loss: 0.117400 Valid loss: 0.316657                                                                                    
[Epoch 088] Loss: 0.097304 Valid loss: 0.338525                                                                                    
[Epoch 089] Loss: 0.091212 Valid loss: 0.341918                                                                                    
[Epoch 090] Loss: 0.087092 Valid loss: 0.361795 
```
The training continue starting at epoch 41 as expected and ending at epoch 90.  
Now let's diagnose the model training with a new `autoplot()`:

```{r}
autoplot(model_fit)
```

![training and validation loss plot for pretrained tabnet](pretrained_model.png)
We can see that model starts to overfit after epoch 52 as the validation-set loss reaches its minimum. If we consider epoch 54 to be the epoch to move to production, we can redo the training from checkpoint 50 for 4 epochs : 


```{r}
model_fit <- tabnet_fit(rec, train , tabnet_model = model_fit, from_epoch=50, epochs = 4, valid_split = 0.2, verbose=TRUE)
```

Now, `predict()` and `tabnet_explain()` functions will use the model from last epoch, being here epoch 54.  
Finally, we can measure the results against our test set:

```{r}
augment(model_fit, test, type = "prob") %>% 
  roc_auc(Class, .pred_bad)
```

```
# A tibble: 1 × 3
  .metric .estimator .estimate
  <chr>   <chr>          <dbl>
1 roc_auc binary         0.653
```

## Comparing against a model without pretraining

The question now is "what if we did not pretrain the model ?" We can build a vanilla tabnet model on the `train` dataset for comparison : 

```{r}
vanilla_model_fit <- tabnet_fit(rec, train , valid_split = 0.2, epochs = 50, verbose=TRUE)

```

```
[Epoch 001] Loss: 0.725771 Valid loss: 0.491390                                                                                    
[Epoch 002] Loss: 0.328243 Valid loss: 0.268964                                                                                    
[Epoch 003] Loss: 0.241829 Valid loss: 0.181927                                                                                    
[Epoch 004] Loss: 0.205124 Valid loss: 0.166245                                                                                    
[Epoch 005] Loss: 0.218578 Valid loss: 0.161050                                                                                    
[Epoch 006] Loss: 0.192196 Valid loss: 0.150333                                                                                    
[Epoch 007] Loss: 0.235362 Valid loss: 0.148007                                                                                    
[Epoch 008] Loss: 0.167538 Valid loss: 0.151358                                                                                    
[Epoch 009] Loss: 0.146265 Valid loss: 0.155682                                                                                    
[Epoch 010] Loss: 0.168639 Valid loss: 0.154740                                                                                    
[Epoch 011] Loss: 0.164683 Valid loss: 0.153092                                                                                    
[Epoch 012] Loss: 0.166895 Valid loss: 0.152931                                                                                    
[Epoch 013] Loss: 0.153897 Valid loss: 0.155470                                                                                    
...
[Epoch 040] Loss: 0.125999 Valid loss: 0.184985                                                                                    
[Epoch 041] Loss: 0.116995 Valid loss: 0.173780                                                                                    
[Epoch 042] Loss: 0.161852 Valid loss: 0.166246                                                                                    
[Epoch 043] Loss: 0.144460 Valid loss: 0.171052                                                                                    
[Epoch 044] Loss: 0.146622 Valid loss: 0.170186                                                                                    
[Epoch 045] Loss: 0.138121 Valid loss: 0.165647                                                                                    
[Epoch 046] Loss: 0.130105 Valid loss: 0.154928                                                                                    
[Epoch 047] Loss: 0.123869 Valid loss: 0.158995                                                                                    
[Epoch 048] Loss: 0.139626 Valid loss: 0.159415                                                                                    
[Epoch 049] Loss: 0.125219 Valid loss: 0.151668                                                                                    
[Epoch 050] Loss: 0.129994 Valid loss: 0.154256                                          

```

```{r}
autoplot(vanilla_model_fit)
```
![Training and validation loss plot for vanilla tabnet](vanillia_model.png)
The model overfits around epoch 20, we will restore it at checkpoint epoch 20 and retrain for 1 epoch , and proceed to prediction : 
```{r}
vanilla_model_fit <- tabnet_fit(rec, train , tabnet_model= vanilla_model_fit, from_epoch=20, valid_split = 0.2, epochs = 1, verbose=TRUE)
augment(vanilla_model_fit, test, type = "prob") %>% 
  roc_auc(Class, .pred_good)

```

```
# A tibble: 1 x 3
  .metric .estimator .estimate
  <chr>   <chr>          <dbl>
1 roc_auc binary         0.557
```

We can see here a much lower ROC-AUC compared with using the pretraining step. 

---
title: "Hierarchical Classification"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Hierarchical Classification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
requireNamespace("vip", quietly = TRUE)
```

```{r setup}
library(tabnet)
library(dplyr)
library(data.tree)
library(ggplot2)
library(rsample)
library(tibble)
set.seed(202307)
```

## Data format

The supported data format for hierarchical classification is the `Node` object format from package `{data.tree}`.

This is a general purpose format that fits generic hierarchical tree encoding needs. Each node of the tree is associated with predictor values through the `attributes` in the data `Node` object.

|{tabnet} concept| {data.tree} concept |see command|
|---|---|---|
|dataset predictor| Node `attributesAll` |  acme example |
|dataset multi-label target| Node hierarchy | print(acme) |


A very basic example is the `acme` dataset to show you how the two predictors values `cost` and `p` are associates attributes of each node in the hierarchy : 
```{r}
data(acme, package = "data.tree")
acme$attributesAll
print(acme, "cost", "p" , limit = 8)
```
So printing Node objects reverse the usual ordering, as target is printed first in column `levelName`, and predictors printed right of it. 

As you can see, only leaf nodes of the tree gets predictors value. {tabnet} will take this into account via an `ancestor` square sparse tensor registering all possible parent-child relation among the target labels.


## Data preparation

Multiple manual or programmatic methods are available to create or update predictors. They are detailled in the `vignette("data.tree", package = "data.tree")`.

- a lot of native hierarchical data-format conversion from files to `Node` are covered by the`{data.tree}` package. You can find them in the "Create tree from a file" section of the same vignette. If needed, the `{ape}` package covers a lot of conversion format to the `philo` format. Thus you can reach the `Node` format in maybe two transformation steps...

- A quick way to achieve `Node` format from a data frame with columns being the different levels of the hierarchy consist in pasting the columns into a single string with `"/"` separator into a `pathString` column.
This will be turn into the expected hierarchy by the `data.tree::as.Node()` command.

Let's do it with `starwars` dataset as a toy example : 
```{r}
data(starwars, package = "dplyr")
head(starwars, 4)

# erroneous Node construction
starwars_tree <- starwars %>% 
  mutate(pathString = paste("StarWars_characters", species, sex, `name`, sep = "/")) %>%
  as.Node()
print(starwars_tree, "name","height", "mass", "eye_color", limit = 8)

```

You may have noticed that `name` and `height` have unexpected values according to the original data: `Human` is not part of the `name` in orginal dataset, and `height` values have been changed into the local height of the tree. This is due to some rules we will have to follow to create the Node from data frame.

## `Node` preparation rules for {tabnet} models

### Avoid `factor` predictors

As `as.Node()` will only consider the as.numeric() values of a factor(), you should turn them into characters before applying the `as.Node()` function in order for {tabnet} to properly embed them.  

### Avoid column name collision with reserved {data.tree} names

`name` and `height` are both part of the `NODE_RESERVED_NAMES_CONST` reserved list of names for `Node` attributes. So they must **not be used** as predictor names, or the `as.Node()` function will silently discard them.

### Avoid column named `level_*` to avoid collision with output data.tree names

Your dataset hierarchy will be turn internally into multi-outcomes named `level_1` to `level_n`, n beeing the depth of your tree. Thus column names starting with `level_` should be avoided.

### Ensure the last hierarchy of the tree is the **observation id**

The tree only keeps a single row of attributes per tree leaf. Thus in order to transfer your complete predictors dataset into the Node object, you must keep the last level of the hierarchy to be a unique observation identifier (last resort beeing `rowid_to_column()` to achieve it).

The classification will be done **removing the last level of hierarchy** in any case.

### Ensure there is a **root level** in the hierarchy

The tree should have a single root for all nodes to be consistent. Thus you have to use a constant prefix to all `pathString`.

The classification will be done **removing the first level of hierarchy** in any case.

### Ensure there is no **missing values** in the hierarchical classes

Missing values should be replaced. Turning them  by an explicit "Unknown_something" is a good approach.


Now let's have all those rules applied to the `starwars_tree` : 

```{r}
# demonstration of reserved column modification in Node construction
starwars_tree <- starwars %>% 
  rename(`_name` = "name", `_height` = "height") %>% 
  mutate(
    species = coalesce(species, "Unknown_Species"),
    sex     = coalesce(sex, "Unknown_Sex"),
    pathString = paste("StarWars_characters", species, sex, `_name`, sep = "/")) %>%
  as.Node()
print(starwars_tree, "name", "_name","_height", "mass", "eye_color", limit = 8)
```
 We can see that the reserved `name` column contains slightly different content that the original `_name` column.
 
## Model building

### Data set split

The  `starwars` dataset contains list columns, hosting some variability in the predictor values. Thus we decide here to `unnest_longer` every list column to each of its values. This will triple the size of the `starwars` dataset.  
The dataset split here will be done upfront of the transformation into `as.Node()`.
We will use `rsample::initial_split()` to split with a stratification on the parent category of the first level of our hierarchy which is `species`.
```{r}
starw_split <- starwars %>% 
  tidyr::unnest_longer(films) %>% 
  tidyr::unnest_longer(vehicles, keep_empty = TRUE) %>% 
  tidyr::unnest_longer(starships, keep_empty = TRUE) %>% 
  initial_split( prop = .8, strata = "species")
```


In order to train a model properly, we should prevent the outcomes to be part of the predictor columns. For the sake of demonstration, the `_name` column was present in `starwars_tree` but must now be dropped.
```{r}
# correct Node construction for hierarchical modeling
starwars_train_tree <- starw_split %>% 
  training() %>% 
  # avoid reserved column names
  rename(`_name` = "name", `_height` = "height") %>% 
  rowid_to_column() %>% 
  mutate(pathString = paste("StarWars_characters", species, sex, rowid, sep = "/")) %>%
  # remove outcomes labels from predictors
  select(-species, -sex, -`_name`, -rowid) %>% 
  # turn it as hierarchical Node
  as.Node()

starwars_test_tree <- starw_split %>% 
  testing() %>% 
  rename(`_name` = "name", `_height` = "height") %>% 
  rowid_to_column() %>% 
  mutate(pathString = paste("StarWars_characters", species, sex, rowid, sep = "/")) %>%
  select(-species, -sex, -`_name`, -rowid) %>% 
  as.Node()

starwars_train_tree$attributesAll
```

Now we can see that none of the predictor leaks the outcome hierarchy information.

## Model building

This `starwars_tree` can now be used as an input for `tabnet_fit()` : 
```{r model training}
config <- tabnet_config(decision_width = 8, attention_width = 8, num_steps = 3, penalty = .003, cat_emb_dim = 2, valid_split = 0.2, learn_rate = 1e-3, lr_scheduler = "reduce_on_plateau", early_stopping_monitor = "valid_loss", early_stopping_patience = 4, verbose = FALSE)

starw_model <- tabnet_fit(starwars_train_tree, config = config, epoch = 170, checkpoint_epochs = 25)

```

## Model diagnostic

We have avoid the verbose output of the model, thus very first diagnostic is the check for model over-fitting though the training loss plot.  

```{r}
autoplot(starw_model)
```

Then global feature importance gives us a clue of model quality
```{r}
vip::vip(starw_model)
```

## Model inference

We can infer on the test-set
```{r}
starwars_hat <- augment(starw_model, starwars_test_tree)
tail(starwars_hat, n = 5)

```

We can see in the Warnings that the dataset is a challenge as many new levels are found in a lot of predictors in the test set.  
The model also here is very poor on the `level_2` ( `species` ) and on `level_3` ( ` sex` ) as this is definitively not a model-intended dataset. The reason is that the input dataset not collecting large samples of distinctive observation per leaf, but rather a very diverse but limited number of characters compatible with watching a movie saga.

Despite the performance, we do have local feature importance on the complete dataset here :  
```{r}
starwars_explain <- tabnet_explain(starw_model, starwars_test_tree)
autoplot(starwars_explain)
autoplot(starwars_explain, type = "steps")
```
Hopefully your own hierarchical outcome will have a better success than the one here with `starwars` dataset. But in this journey, you have learned a lot in the data format constraints and solutions, and you now have a new performing solution in your toolbox.

<!-- Thanks for submitting a pull request! Please provide enough information so that others can review your pull request. -->
<!-- Explain the **motivation** for making this change. What existing problem does the pull request solve? -->
**IMPORTANT: Please do not create a Pull Request without creating an issue first.**

*Any change needs to be discussed before proceeding. Failure to do so may result in the rejection of the pull request.*

<!-- Thanks for submitting a pull request! Please provide enough information so that others can review your pull request. -->
<!-- You can skip this if you're fixing a typo or adding an app to the Showcase.  -->
<!-- Explain the **motivation** for making this change. What existing problem does the pull request solve? -->


<!-- In addition to that please answer these questions: -->

**What kind of change does this PR introduce?**

<!-- E.g. a bugfix, feature, refactoring, build related change, etc… -->

**Does this PR introduce a breaking change?**

<!-- If this PR introduces a breaking change, please describe the impact and a migration path for existing applications. -->

**What needs to be documented once your changes are merged?**

<!-- List all the information that needs to be added to the documentation after merge -->
<!-- When your changes are merged you will be asked to contribute this to the documentation -->

**Closing issues**

Put `closes #XXXX` in your comment to auto-close the issue that your PR fixes (if such).

---
name: Feature request
about: Suggest an idea for this project
title: ''
labels: enhancement
assignees: eduardocarvp, Hartorn, j-abi, Optimox

---

<!-- Please don't delete this template or we'll close your issue -->

## Feature request


**What is the expected behavior?**


**What is motivation or use case for adding/changing the behavior?**


**How should this be implemented in your opinion?**


**Are you willing to work on this yourself?**
yes

---
title: "Using ROC AUM loss for imbalanced binary classification"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Using ROC AUM loss for imbalanced binary classification}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
editor_options: 
  markdown: 
    fig-width: 9
    fig-height: 6
    fig-cap-location: "top"
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = torch::torch_is_installed()
)
```

```{r setup}
library(tabnet)
suppressPackageStartupMessages(library(tidymodels))
library(modeldata)
data("lending_club", package = "modeldata")
set.seed(20250809)
```

::: callout-note
This vignette is a continuation of `vignette("tidymodels-interface")`. So we highly encourage you to start with it to be up to speed with this vignette.
:::

## Introduction

The previously used `lending_club` dataset is highly imbalanced, leading to challenging result in the binary classification task. Despite we got fairly good accuracy with default model design, the `roc_auc()` metric was poor, mainly due to this imbalanced problem.

Here, we will see how tabnet features allow improved performance on such family of classification problems.

## How imbalance is my problem ?

The target variable `Class` imbalance can be evaluated through the class imbalance Ratio :

```{r}
class_ratio <- lending_club |> 
  summarise(sum( Class == "good") / sum( Class == "bad")) |> 
  pull() 

class_ratio
```

With a class_ratio of 18.1, the target variable is seriously imbalanced, making the minority class much harder to model.

## Solutions to improve imbalanced classification models

First, usual solution to such problem is over-sampling of the minority class, and/or down-sampling the majority class in the training data. We won't cover this here.

The second solution is case weighting. As {tidymodels} offers the framework to manage such case weighting, we'll first use it to compare two model families - XGBoost and Tabnet - with that feature.

Last, we would like to also **optimize** the model according to the metric we are looking at. As the metric of choice for imbalanced dataset are `roc_auc()` or `roc_pr()`, we definitively want a **loss function** that is a proxy of those. This loss is available in {tabnet} with the `nn_aum_loss()` from [Optimizing ROC Curves with a Sort-Based Surrogate Loss for Binary Classification and Changepoint Detection (J Hillman, TD Hocking)](https://jmlr.org/papers/v24/21-0751.html).

## Using the AUC metric and `pr_curve()` plots

Measuring the ROC_AUC or AUC_PR can't be separated from plotting the `pr_curve()`.

Let's baseline our models on two different workflows, one for tabnet, the other for XGBoost. This is a big chunk of code, but it is mainly a copy of the previous vignette.

```{r}
lending_club <- lending_club |>
  mutate(
    case_wts = if_else(Class == "bad", class_ratio, 1),
    case_wts = importance_weights(case_wts)
  )

split <- initial_split(lending_club, strata = Class)
train <- training(split)
test  <- testing(split)

tab_rec <- train |>
  recipe() |>
  update_role(Class, new_role = "outcome") |>
  update_role(-has_role(c("outcome", "id", "case_weights")), new_role = "predictor")

xgb_rec <- tab_rec |> 
  step_dummy(term, sub_grade, addr_state, verification_status, emp_length)

tab_mod <- tabnet(epochs = 100) |> 
  set_engine("torch", device = "cpu") |> 
  set_mode("classification")

xgb_mod <- boost_tree(trees = 100) |> 
  set_engine("xgboost") |> 
  set_mode("classification")

tab_wf <- workflow() |> 
  add_model(tab_mod) |> 
  add_recipe(tab_rec) |> 
  add_case_weights(case_wts)

xgb_wf <- workflow() |> 
  add_model(xgb_mod) |> 
  add_recipe(xgb_rec) |> 
  add_case_weights(case_wts)
```

Few details to be noticed

-   we compute and tag as `importance_weight()` a new column `case_wts` in the `lending_club` dataset.

-   this column is excluded from the recipe() predictors role.

-   we explicitly mention this column role in each `workflow()` via `add_case_weights()`.

We can now `fit()` each model and plot the precision-recall curve on the test-set :

```{r}
#| label: "vanilia_models_fitting"
#| layout.ncol: 2
#| fig.cap: 
#|   - "Tabnet, no case-weight, default loss"
#|   - "XGBoost, no case-weight"
#| 
tab_fit <- tab_wf |> fit(train)
xgb_fit <- xgb_wf |> fit(train)

tab_test <- tab_fit |> augment(test)
xgb_test <- xgb_fit |> augment(test)

tab_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot()

xgb_test |>
  pr_curve(Class, .pred_good) |>
  autoplot()

```

Both models are returning poor results.

## Case-weight

Weighting each observation by the importance weight of the class is made available in {tabnet} through

-   marking one variable as importance weight variable via `workflow::add_case_weights()`

-   using the case_weight variable as such at inference time through the `case_weights =` parameter in functions that allows it.

Let's proceed

```{r}
#| label: "case-weights_prediction"
#| layout.ncol: 2
#| fig.cap: 
#|   - "Tabnet, with case-weight, default loss"
#|   - "XGBoost, with case-weight"
#| 
tab_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 

xgb_test |>
  pr_curve(Class, .pred_good, case_weights = case_wts) |>
  autoplot() 
```

The boost on the `pr_curve()` is impressive for both models, Tabnet remains behind XGBoost here[^1].

[^1]: Or may become leader if you change the initial random seed.

## ROC_AUM loss

{tabnet} implement the ROC AUM loss that will drive the torch optimizer to the best possible AUC. Let's use it to compare to previous models :

```{r}
#| label: "AUM_based_model_fit"
# configure the AUM loss
tab_aum_mod <- tabnet(epochs = 100, loss = tabnet::nn_aum_loss, learn_rate = 0.02) |> 
  set_engine("torch", device = "cpu") |> 
  set_mode("classification")

# derive a workflow
tab_aum_wf <- workflow() |> 
  add_model(tab_aum_mod) |> 
  add_recipe(tab_rec) |> 
  add_case_weights(case_wts)

# fit and augment the test dataset with prediction
tab_aum_fit <- tab_aum_wf |> fit(train)
tab_aum_test <- tab_aum_fit |> augment(test)
```

Now let's compare the result on the PR curve with the default loss side by side:

```{r}
#| label: "AUM_model_pr_curve"
#| layout.ncol: 2
#| fig.cap: 
#|   - "Tabnet, no case-weight, default loss"
#|   - "Tabnet, no case-weight, ROC_AUM loss"
#|   
tab_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot() 

tab_aum_test |> 
  pr_curve(Class, .pred_good) |> 
  autoplot() 
```

We can see a real[^2] improvement with the AUM loss, compared to the default `nn_bce_loss()` but globally still a poor recall. 

[^2]: With improvement level being sensitive to the random seed.

## All together

Nothing prevent us to use both features, as they are independent. That is what we do here. Moreover, it is here without additional computation, as it is done post inference.

```{r}
#| label: "AUM_and_case-weights_prediction"
#| layout.ncol: 2
#| fig.cap: 
#|   - "Tabnet, with case-weight, default loss"
#|   - "Tabnet, with case-weight, ROC_AUM loss"
#| 
tab_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 


tab_aum_test |> 
  pr_curve(Class, .pred_good, case_weights = case_wts) |> 
  autoplot() 
```

Here the boost in recall is impressive, making Tabnet model far above any experimented challenger model[^3].

[^3]: Within an educational and reproducible intend only.

---
title: "Interpretation tools"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Interpretation tools}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = torch::torch_is_installed(),
  out.width = "100%", 
  out.height = "300px", 
  fig.width = 14
)
requireNamespace("vip", quietly = TRUE)
```

```{r setup}
library(tabnet)
library(tidyverse, warn.conflicts = FALSE)
set.seed(1)
# You may want to add this for complete reproducibility
# torch::torch_manual_seed(1)
```

TabNet is one of the first deep-learning interpretable model. Thanks to the underlying 
neural-network architecture, TabNet uses feature-selection masks that can help identify
which predictor is used at each step.

The paper also defines an aggregate measure that combines the masks of
each step into a single measure.

Note that the selection of important features in the masks is done
instance-wise. Thus, you can identify, for each observation, which
predictors were considered relevant.

## Experiments

To show how to use the interpretation tools of `tabnet`, we are going to
perform 2 experiments using synthetic datasets very similar to those
that were used in the paper.

### Datasets

First, let's define the functions that we will use to generate data:

-   `make_syn2` will generate a dataset with 10 columns, but only
    columns 3-6 are used to calculate the `y` response vector. This is
    similar to *Syn2* in the paper.

-   `make_syn4` will generate a dataset with 10 columns too. The
    response vector depends on column 10: if the value is greater than
    0, we use columns 1-2 to compute the logits, otherwise we use
    columns 5-6.

```{r}
logit_to_y <- function(logits) {
  p <- exp(logits)/(1 + exp(logits))
  y <- factor(ifelse(p > 0.5, "yes", "no"), levels = c("yes", "no"))
  y
}

make_random_x <- function(n) {
  x <- as.data.frame(lapply(1:10, function(x) rnorm(n)))
  names(x) <- sprintf("V%02d", 1:10)
  x
}

make_syn2 <- function(n = 5000) {
  x <- make_random_x(n)
  logits <- rowSums(x[,3:6])
  x$y <- logit_to_y(logits)
  x
}

make_syn4 <- function(n = 5000) {
  x <- make_random_x(n)
  logits <- ifelse(
    x[,10] > 0,
    rowSums(x[,1:2]),
    rowSums(x[,5:6])
  )
  
  x$y <- logit_to_y(logits)
  x
}
```

Now let's generate the datasets:

```{r}
syn2 <- make_syn2()
syn4 <- make_syn4()
```

### Syn2

Let's fit a TabNet model to the `syn2` dataset and analyze the
interpretation metrics.

```{r}
fit_syn2 <- tabnet_fit(y ~ ., syn2, epochs = 45, learn_rate = 0.06, device = "cpu")
```

In the feature importance plot we can see that, as expected, features
`V03-V06` are by far the most important ones.

```{r}
#| fig.alt: "A variable importance plot of the fitted model on syn2 dataset showing V03 then V06, V04, v10 and V5 as the 5 most important features, in that order."
vip::vip(fit_syn2)
```

Now let's visualize the aggregated-masks plot. In this figure we see
each observation on the x axis and each variable on the y axis. The
colors represent the importance of the feature in predicting the value
for each observation.

```{r}
#| fig.alt: "A tabnet explaination plot of the fitted model on syn2 dataset. The plot shows numerous important observations in V03 then V06 and V04. the other variables are shown with low importance points or sparse observations with importance"
library(tidyverse)
ex_syn2 <- tabnet_explain(fit_syn2, syn2)

autoplot(ex_syn2, quantile = 0.99)
```

We can see that the region between V03 and V06 concentrates most of the
higher intensity colors, and the other variables are close to 0. This is
expected because those are the variables that we considered when
building the dataset.

Next, we can visualize the attention masks for each step in the
architecture.

```{r}
#| fig.alt: "3 tabnet explaination plots, one for each step of the fitted model on syn2 dataset. The Step 1 plot shows numerous important observations with V02 and V03 having high importance, step 2 plot highlight the importance of V03 and V06. Third step plot highlight V03 and V10 as important variables"
autoplot(ex_syn2, type="steps")
```

We see that the first step captures a lot of noise, but the other 2
steps focus specifically on the important features.

## Syn 4

Now let's analyze the results for the Syn4 dataset. This dataset is a
little more complicated for TabNet because there's a strong interaction
between the variables. Depending on V10, different variables are used to
create the response variable and we expect to see this in the masks.

First we fit the model for 10 epochs.

```{r}
fit_syn4 <- tabnet_fit(y ~ ., syn4, epochs = 50, device = "cpu", learn_rate = 0.08)
```

In the feature importance plot we have, as expected, strong importance
for `V10`, and the other features that are used conditionally - either
`V01-V02` or `V05-V06`.

```{r}
#| fig.alt: "A variable importance plot of the fitted model on syn4 dataset"
vip::vip(fit_syn4)
```

Now let's visualize the attention masks. Notice that we arranged the
dataset by `V10` so we can easily visualize the interaction effects.

We also trimmed to the 98th percentile so the colors shows the
importance even if there are strong outliers.

```{r}
#| fig.alt: "A tabnet explaination plot of the fitted model on syn4 dataset. The plot shows numerous important observations in V06 for low values of V10, and importance of V01 and V02 for high values of V10."
ex_syn4 <- tabnet_explain(fit_syn4, arrange(syn4, V10))

autoplot(ex_syn4, quantile=.98)
```

From the figure we see that V10 is important for all observations. We
also see that for the first half of the dataset `V05` and `V06` is the
most important feature, while for the other half, `V01` and `V02` are
the important ones.

We can also visualize the masks at each step in the architecture.

```{r}
#| fig.alt: "3 tabnet explaination plots, one for each step of the fitted model on syn4 dataset."
autoplot(ex_syn4, type="steps", quantile=.995)
```

We see that step 1 and 3 both focus on `V10`, but on different
additional features, depending on `V10`. Step 2 seems to have found some
noise in `V08`, but also focuses strongly on `V01-V02` and `V05-V06`.

---
title: "Fitting tabnet with tidymodels"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Fitting tabnet with tidymodels}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

```{r setup}
library(tabnet)
library(tidymodels)
library(modeldata)
```

In this vignette, we show how to create a TabNet model using the tidymodels interface.

We are going to use the `lending_club` dataset available
in the `modeldata` package.

First let's split our dataset into training and testing so we can later access performance of our model:

```{r}
set.seed(123)
data("lending_club", package = "modeldata")
split <- initial_split(lending_club, strata = Class)
train <- training(split)
test  <- testing(split)
```

We now define our pre-processing steps. Note that TabNet handles categorical variables, so we don't need to do any kind of transformation to them. Normalizing the numeric variables is a good idea though.

```{r}
rec <- recipe(Class ~ ., train) %>%
  step_normalize(all_numeric())
```

Next, we define our model. We are going to train for 50 epochs with a batch size of 128. There are other hyperparameters but, we are going to use the defaults.

```{r}
mod <- tabnet(epochs = 50) %>%
  set_engine("torch", verbose = TRUE) %>%
  set_mode("classification")
```

We also define our `workflow` object:

```{r}
wf <- workflow() %>%
  add_model(mod) %>%
  add_recipe(rec)
```

We can now define our cross-validation strategy:

```{r}
folds <- vfold_cv(train, v = 5)
```

And finally, fit the model:

```{r}
fit_rs <- wf %>% fit_resamples(folds)
```

After a few minutes we can get the results:

```{r}
collect_metrics(fit_rs)
```

```
# A tibble: 3 × 6
  .metric     .estimator   mean     n  std_err .config             
  <chr>       <chr>       <dbl> <int>    <dbl> <chr>               
1 accuracy    binary     0.945      5 0.000869 Preprocessor1_Model1
2 brier_class binary     0.0535     5 0.00122  Preprocessor1_Model1
3 roc_auc     binary     0.611      5 0.0153   Preprocessor1_Model1
```

And finally, we can verify the results in our test set:

```{r}
model <- wf %>% fit(train)
model %>%
  augment( test) %>% 
  roc_auc(Class, .pred_good, event_level = "second")
```

```
# A tibble: 1 x 3
  .metric .estimator .estimate
  <chr>   <chr>          <dbl>
1 roc_auc binary         0.710
```

---
title: "investigating issue 104"
output: html_notebook
tldr: see https://github.com/mlverse/tabnet/issues/114
---


# SYSTEM SETTING
## Import Packages
```{r}
#| label: setup
#| warning: false
#| message: false
set.seed(seed = 2023)

library(tidyverse)
library(here)
library(tidymodels)
library(torch)
# library(tabnet)
devtools::load_all()
library(data.tree)
# library(tidyverse)
```


```{r }
#| eval: false
#may take 3 mins
source(here("tools/setup.R"))
```
NODE_RESERVED_NAMES_CONST

## replace reserved_names

```{r}
#| eval: false
check_compliant_node(attrition_tree)
check_compliant_node(starwars_tree)
```
## fit a multi-outcome classifier

```{r}
#| eval: false
#| label: Reprex for https://github.com/mlverse/tabnet/issues/125
library(tabnet)
library(recipes)
starwars_rec <-  recipe(species + homeworld ~ ., data = starwars %>%
                          select(-where(is.list)) |> 
                          mutate(species = coalesce(species, "Unknown_Species"),
           homeworld     = coalesce(homeworld, "Unknown_homeworld"))) %>%
    step_zv(all_predictors()) %>%
    step_impute_mode(all_outcomes()) %>%
    step_string2factor(all_string_predictors(), all_outcomes())

fit_multi <- tabnet_fit(starwars_rec, starwars, cat_emb_dim = 2)
```

## prepare Node objects
```{r}
#| eval: true
starw_split <- starwars %>% 
  tidyr::unnest_longer(films) %>% 
  tidyr::unnest_longer(vehicles, keep_empty = TRUE) %>% 
  tidyr::unnest_longer(starships, keep_empty = TRUE) %>% 
  initial_split( prop = .8, strata = "species")

starwars_train_tree <- starw_split %>% 
  training() %>% 
  # avoid reserved column names
  rename(`_name` = "name", `_height` = "height") %>% 
  rowid_to_column() %>% 
    mutate(species = coalesce(species, "Unknown_Species"),
           sex     = coalesce(sex, "Unknown_Sex"),
           pathString = paste("StarWars_characters", species, sex, `_name`, sep = "/")) %>%
  # remove outcomes labels from predictors
  select(-species, -sex, -`_name`, -rowid) %>% 
  # turn it as hierarchical Node
  as.Node()

starwars_test_tree <- starw_split %>% 
  testing() %>% 
  rename(`_name` = "name", `_height` = "height") %>% 
  rowid_to_column() %>% 
  mutate(pathString = paste("StarWars_characters", species, sex, rowid, sep = "/")) %>%
  select(-species, -sex, -`_name`, -rowid) %>% 
  as.Node() # [77 entries, 11 attributes]

starwars_test_tree$attributesAll

```

```{r}
#| eval: true

attrition_tree <- attrition %>% 
  rowid_to_column() %>% 
  mutate(pathString = paste("attrition", Department, JobRole, rowid, sep = "/")) %>%
  as.Node()
print(attrition_tree, "Age", "Education","EducationField", "DailyRate",  limit = 12)
```
#  compute torch sparse ancestor matrix for attrition_tree
```{r}
#|label: "hardhat.R L160-L170"
x <- attrition_tree$clone(deep = FALSE)

# ensure there is no level_* col in the Node object
check_compliant_node(x)
# get tree leaves and extract attributes into data.frames
xy_df <- node_to_df(x)
processed <- hardhat::mold(xy_df$x, xy_df$y)
# Given n classes, ancestor is an (n x n) matrix where ancestor_ij = 1 if class i is descendant of class j
# ancestor_tt is the torch_tensor of ancestor

ancestor_tt <- build_ancestor_matrix_from_outcomes(
  x = x,
  outcomes = processed$outcomes,
  device = "cpu"
)
 # [1 12 12]
```


```{r}
#| label: "check tree levels attrition"

levels <- sapply(data.tree::Traverse(x), `[[`, "level")
table(levels)
```

```{r}
#|label: "hardhat.R  L369-L370"
# tabnet_bridge(processed, config = config, tabnet_model, from_epoch, task = "supervised") -->
config <- tabnet_config(epochs = 2)
config$ancestor <- ancestor_tt # class "sparse_coo_tensor"
config$outcomes <- processed$outcomes
tabnet_model_lst <- tabnet:::tabnet_initialize(processed$predictors, processed$outcomes, config = config)
tabnet_model <-  tabnet:::new_tabnet_fit(tabnet_model_lst, blueprint = processed$blueprint)

fit_lst <- tabnet:::tabnet_train_supervised(tabnet_model, processed$predictors, processed$outcomes, config = config) # Fails with 
#> ! Erreur : mat1 and mat2 shapes cannot be multiplied (1470x2 and 1470x12)
#> Exception raised from meta at /pytorch/aten/src/ATen/native/LinearAlgebra.cpp:204 (most recent call first):

```

```{r}
#|label: "model_training.R  L460"

obj <- tabnet_model
x <- processed$predictors
y <- processed$outcomes
train_ds <-   torch::dataset(
  initialize = function() {},
  .getbatch = function(batch) {resolve_data(x[batch,], y[batch,])},
  .length = function() {nrow(x)}
)()
train_dl <- torch::dataloader(
  train_ds,
  batch_size = config$batch_size,
  drop_last = config$drop_last,
  shuffle = TRUE ,
  num_workers = config$num_workers
)
# ... to be continued ...
```

```{r}
#|label: "model_training.R train_batch L285"
epoch <- 1
iter <- dataloader_make_iter(train_dl)
batch <- dataloader_next(iter)

c(output, M_loss) %<-% network(batch$x, batch$x_na_mask)
outcome_nlevels <- as.numeric(batch$output_dim$to(device="cpu")) # c(3, 9)
  
config$loss_fn(output,  class_ids_to_binary(
      y = batch$y,
      outcomes = config$outcomes,
      device = out$device
    ))
)

```

```{r}
#|label: "loss.R nnf_mc_loss L140"

target <- batch$y
R <- config$ancestor


```



#  compute torch sparse ancestor matrix for starwars_tree
```{r}
#|label: "hardhat.R L160-L170"
x <- starwars_tree$clone(deep = FALSE)

# ensure there is no level_* col in the Node object
check_compliant_node(x)
# get tree leaves and extract attributes into data.frames
xy_df <- node_to_df(x)
processed <- hardhat::mold(xy_df$x, xy_df$y)
# Given n classes, ancestor is an (n x n) matrix where ancestor_ij = 1 if class i is descendant of class j
# ancestor_tt is the `R` torch_tensor of ancestor

ancestor_tt <- build_ancestor_matrix_from_outcomes(
  x = x,
  outcomes = processed$outcomes,
  device = "cpu"
) # [1 43 43]
```

```{r}
#| label: "check tree levels starwars_tree"

levels <- sapply(data.tree::Traverse(x), `[[`, "level")
table(levels)
```


---






# quand tabnet_fit() marche
```{r}
fit <- tabnet_fit(attrition_tree, cat_emb_dim = 2)
predict(fit, attrition_tree)
# Erreur dans switch(object$spec$mode, regression = "numeric", classification = "class",  : 
#   EXPR doit être un vecteur de longueur 1
```

# develop `tabnet:::tabnet_fit.Node()`
```{r}
#| eval: false
x <- starwars_tree
# TODO check there is no level_* col in the tree
  # extract attributes data.frame
  xy_df <- ToDataFrameTypeCol(x, x$attributesAll)
  x_df <- xy_df %>% select(-starts_with("level_"))
  y_df <- xy_df %>% select(starts_with("level_"))
  processed <- hardhat::mold(x_df, y_df)
  # embed the M matrix in Sextra
  ancestor <- ToDataFrameNetwork(datatree) %>%
    mutate_if(is.character, . %>% as.factor %>% as.numeric) 
  processed$extra$M <- Matrix::sparseMatrix(ancestor$from, ancestor$to, x = 1)
  check_type(processed$outcomes)
```

----------------------------------------------------------
# a yarr sourced file
```{r}
#| eval: false

library(yarr)

# Error   Invalid type specification.
# yeast_train <- foreign::read.arff("~/_Data.science/C-HMCNN-master/HMC_data/others/D0_yeast_GO.trainvalid.arff")
yeast_train_raw <- yarr::read.arff("~/_Data.science/C-HMCNN-master/HMC_data/others/D0_yeast_GO.trainvalid.arff")
yeast_train <- yeast_train_raw %>% mutate_all(as.numeric) %>% mutate_all(as.logical)
```

# another yarr sourced file
```{r}
#| eval: false

library(yarr)
enron_train <- yarr::read.arff("~/_Data.science/C-HMCNN-master/HMC_data/others/Enron_corr_trainvalid.arff")
enron_test <- yarr::read.arff("~/_Data.science/C-HMCNN-master/HMC_data/others/Enron_corr_test.arff")
```

# print module works even after a reload from disk

    An `nn_module` containing 10,742 parameters.
    
    -- Modules ---------------------------------------
    * embedder: <embedding_generator> #283 parameters
    * embedder_na: <na_embedding_generator> #0 parameters
    * tabnet: <tabnet_no_embedding> #10,458 parameters
    
    -- Parameters ------------------------------------
    * .check: Float [1:1]

---

    An `nn_module` containing 10,742 parameters.
    
    -- Modules ---------------------------------------
    * embedder: <embedding_generator> #283 parameters
    * embedder_na: <na_embedding_generator> #0 parameters
    * tabnet: <tabnet_no_embedding> #10,458 parameters
    
    -- Parameters ------------------------------------
    * .check: Float [1:1]

# print module works

    An `nn_module` containing 13,190 parameters.
    
    -- Modules ---------------------------------------
    * initial_bn: <nn_batch_norm1d> #146 parameters
    * embedder: <embedding_generator> #283 parameters
    * embedder_na: <na_embedding_generator> #0 parameters
    * masker: <random_obfuscator> #0 parameters
    * encoder: <tabnet_encoder> #10,304 parameters
    * decoder: <tabnet_decoder> #2,456 parameters
    
    -- Parameters ------------------------------------
    * .check: Float [1:1]

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-pipe.R
\name{\%>\%}
\alias{\%>\%}
\title{Pipe operator}
\usage{
lhs \%>\% rhs
}
\value{
Returns \code{rhs(lhs)}.
}
\description{
See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabnet_package.R
\docType{package}
\name{tabnet-package}
\alias{tabnet-package}
\title{tabnet: Fit 'TabNet' Models for Classification and Regression}
\description{
Implements the 'TabNet' model by Sercan O. Arik et al. (2019) \doi{10.48550/arXiv.1908.07442} with 'Coherent Hierarchical Multi-label Classification Networks' by Giunchiglia et al. \doi{10.48550/arXiv.2010.10151} and provides a consistent interface for fitting and creating predictions. It's also fully compatible with the 'tidymodels' ecosystem.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://mlverse.github.io/tabnet/}
  \item \url{https://github.com/mlverse/tabnet}
  \item Report bugs at \url{https://github.com/mlverse/tabnet/issues}
}

}
\author{
\strong{Maintainer}: Christophe Regouby \email{christophe.regouby@free.fr} [contributor]

Authors:
\itemize{
  \item Daniel Falbel \email{daniel@rstudio.com}
}

Other contributors:
\itemize{
  \item  RStudio [copyright holder]
  \item Egill Fridgeirsson [contributor]
  \item Philipp Haarmeyer [contributor]
  \item Sven Verweij (\href{https://orcid.org/0000-0002-5573-3952}{ORCID}) [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss.R
\name{get_constr_output}
\alias{get_constr_output}
\title{Apply hierarchy constraints via max-pooling over descendants (MCM)}
\usage{
get_constr_output(x, R)
}
\arguments{
\item{x}{A \code{torch_tensor} of shape \verb{(batch_size, n_classes)}.}

\item{R}{A \code{torch_tensor} of shape \verb{(1, n_classes, n_classes)} where
\code{R[1, i, j] = 1} iff class \code{i} is a descendant of class \code{j}.}
}
\value{
A \code{torch_tensor} of shape \verb{(batch_size, n_classes)} with constrained outputs.
}
\description{
Given neural network outputs x and ancestor matrix R, enforces that
if a class is predicted positive, all its ancestors must also be positive.
Implements: \verb{final_out[i] = max\{x[j] : R[i,j] = 1\}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{tabnet}
\alias{tabnet}
\title{Parsnip compatible tabnet model}
\usage{
tabnet(
  mode = "unknown",
  cat_emb_dim = NULL,
  decision_width = NULL,
  attention_width = NULL,
  num_steps = NULL,
  mask_type = NULL,
  mask_topk = NULL,
  num_independent = NULL,
  num_shared = NULL,
  num_independent_decoder = NULL,
  num_shared_decoder = NULL,
  penalty = NULL,
  feature_reusage = NULL,
  momentum = NULL,
  epochs = NULL,
  batch_size = NULL,
  virtual_batch_size = NULL,
  learn_rate = NULL,
  optimizer = NULL,
  loss = NULL,
  clip_value = NULL,
  drop_last = NULL,
  lr_scheduler = NULL,
  rate_decay = NULL,
  rate_step_size = NULL,
  checkpoint_epochs = NULL,
  verbose = NULL,
  importance_sample_size = NULL,
  early_stopping_monitor = NULL,
  early_stopping_tolerance = NULL,
  early_stopping_patience = NULL,
  skip_importance = NULL,
  tabnet_model = NULL,
  from_epoch = NULL
)
}
\arguments{
\item{mode}{A single character string for the type of model. Possible values
for this model are "unknown", "regression", or "classification".}

\item{cat_emb_dim}{Size of the embedding of categorical features. If int, all categorical
features will have same embedding size, if list of int, every corresponding feature will have
specific embedding size.}

\item{decision_width}{(int) Width of the decision prediction layer. Bigger values gives
more capacity to the model with the risk of overfitting. Values typically
range from 8 to 64.}

\item{attention_width}{(int) Width of the attention embedding for each mask. According to
the paper n_d = n_a is usually a good choice. (default=8)}

\item{num_steps}{(int) Number of steps in the architecture
(usually between 3 and 10)}

\item{mask_type}{(character) Final layer of feature selector in the attentive_transformer
block, either \code{"sparsemax"}, \code{"entmax"} or \code{"entmax15"}.Defaults to \code{"sparsemax"}.}

\item{mask_topk}{(int) mask sparsity top-k for \code{sparsemax15} and \code{entmax15.} See \code{\link[=entmax15]{entmax15()}} for detail.}

\item{num_independent}{Number of independent Gated Linear Units layers at each step of the encoder.
Usual values range from 1 to 5.}

\item{num_shared}{Number of shared Gated Linear Units at each step of the encoder. Usual values
at each step of the decoder. range from 1 to 5}

\item{num_independent_decoder}{For pretraining, number of independent Gated Linear Units layers
Usual values range from 1 to 5.}

\item{num_shared_decoder}{For pretraining, number of shared Gated Linear Units at each step of the
decoder. Usual values range from 1 to 5.}

\item{penalty}{This is the extra sparsity loss coefficient as proposed
in the original paper. The bigger this coefficient is, the sparser your model
will be in terms of feature selection. Depending on the difficulty of your
problem, reducing this value could help (default 1e-3).}

\item{feature_reusage}{(num) This is the coefficient for feature reusage in the masks.
A value close to 1 will make mask selection least correlated between layers.
Values range from 1 to 2.}

\item{momentum}{Momentum for batch normalization, typically ranges from 0.01
to 0.4 (default=0.02)}

\item{epochs}{(int) Number of training epochs.}

\item{batch_size}{(int) Number of examples per batch, large batch sizes are
recommended. (default: 1024^2)}

\item{virtual_batch_size}{(int) Size of the mini batches used for
"Ghost Batch Normalization" (default=256^2)}

\item{learn_rate}{initial learning rate for the optimizer.}

\item{optimizer}{the optimization method. currently only \code{"adam"} is supported,
you can also pass any torch optimizer function.}

\item{loss}{(character or function) Loss function for training (default to mse
for regression and cross entropy for classification)}

\item{clip_value}{If a num is given this will clip the gradient at
clip_value. Pass \code{NULL} to not clip.}

\item{drop_last}{(logical) Whether to drop last batch if not complete during
training}

\item{lr_scheduler}{if \code{NULL}, no learning rate decay is used. If "step"
decays the learning rate by \code{lr_decay} every \code{step_size} epochs. If "reduce_on_plateau"
decays the learning rate by \code{lr_decay} when no improvement after \code{step_size} epochs.
It can also be a \link[torch:lr_scheduler]{torch::lr_scheduler} function that only takes the optimizer
as parameter. The \code{step} method is called once per epoch.}

\item{rate_decay}{multiplies the initial learning rate by \code{rate_decay} every
\code{rate_step_size} epochs. Unused if \code{lr_scheduler} is a \code{torch::lr_scheduler}
or \code{NULL}.}

\item{rate_step_size}{the learning rate scheduler step size. Unused if
\code{lr_scheduler} is a \code{torch::lr_scheduler} or \code{NULL}.}

\item{checkpoint_epochs}{checkpoint model weights and architecture every
\code{checkpoint_epochs}. (default is 10). This may cause large memory usage.
Use \code{0} to disable checkpoints.}

\item{verbose}{(logical) Whether to print progress and loss values during
training.}

\item{importance_sample_size}{sample of the dataset to compute importance metrics.
If the dataset is larger than 1e5 obs we will use a sample of size 1e5 and
display a warning.}

\item{early_stopping_monitor}{Metric to monitor for early_stopping. One of "valid_loss", "train_loss" or "auto" (defaults to "auto").}

\item{early_stopping_tolerance}{Minimum relative improvement to reset the patience counter.
0.01 for 1\% tolerance (default 0)}

\item{early_stopping_patience}{Number of epochs without improving until stopping training. (default=5)}

\item{skip_importance}{if feature importance calculation should be skipped (default: \code{FALSE})}

\item{tabnet_model}{A previously fitted \code{tabnet_model} object to continue the fitting on.
if \code{NULL} (the default) a brand new model is initialized.}

\item{from_epoch}{When a \code{tabnet_model} is provided, restore the network weights from a specific epoch.
Default is last available checkpoint for restored model, or last epoch for in-memory model.}
}
\value{
A TabNet \code{parsnip} instance. It can be used to fit tabnet models using
\code{parsnip} machinery.
}
\description{
Parsnip compatible tabnet model
}
\section{Threading}{


TabNet uses \code{torch} as its backend for computation and \code{torch} uses all
available threads by default.

You can control the number of threads used by \code{torch} with:

\if{html}{\out{<div class="sourceCode">}}\preformatted{torch::torch_set_num_threads(1)
torch::torch_set_num_interop_threads(1)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
library(parsnip)
data("ames", package = "modeldata")
model <- tabnet() \%>\%
  set_mode("regression") \%>\%
  set_engine("torch")
model \%>\%
  fit(Sale_Price ~ ., data = ames)
\dontshow{\}) # examplesIf}
}
\seealso{
tabnet_fit
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss.R
\name{nnf_mc_loss}
\alias{nnf_mc_loss}
\title{Max-Constraint Margin Loss (functional)}
\usage{
nnf_mc_loss(
  output,
  target,
  R,
  to_eval = NULL,
  criterion = nnf_binary_cross_entropy_with_logits
)
}
\arguments{
\item{output}{A \code{torch_tensor} of raw network outputs (pre-sigmoid),
shape \verb{(batch_size, n_classes)}.}

\item{target}{Binary target labels, shape \verb{(batch_size, n_classes)}.}

\item{R}{Ancestor matrix tensor of shape \verb{(1, n_classes, n_classes)} where
\code{R[1, i, j] = 1} iff class \code{i} is a descendant of class \code{j}.}

\item{to_eval}{Optional logical tensor of shape \verb{(n_classes,)} indicating
which classes to include in the loss computation. If \code{NULL}, all classes
are evaluated.}

\item{criterion}{Loss function to apply after constraint propagation.
Default: \code{nnf_binary_cross_entropy_with_logits} (expects raw logits).}
}
\value{
A scalar \code{torch_tensor} containing the computed loss, or a tensor
of shape \verb{(batch_size, n_classes)} if \code{reduction = "none"}.
}
\description{
Computes the hierarchy-constrained loss for multi-label classification.
Enforces that if a class is predicted positive, all its ancestors must
also be positive, using the ancestor matrix R.
}
\details{
The loss combines constrained outputs differently for positive and negative
labels:
\itemize{
\item For positive labels: uses constrained output of label-weighted predictions
\item For negative labels: uses constrained raw predictions (penalizes ancestor violations)
}
}
\seealso{
\code{\link[=nn_mc_loss]{nn_mc_loss()}}, \code{\link[=get_constr_output]{get_constr_output()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss.R
\name{nn_mc_loss}
\alias{nn_mc_loss}
\title{Max-Constraint Margin Loss (module)}
\usage{
nn_mc_loss(
  R,
  to_eval = NULL,
  criterion = nnf_binary_cross_entropy_with_logits,
  reduction = "mean"
)
}
\arguments{
\item{R}{Ancestor matrix tensor of shape \verb{(1, n_classes, n_classes)}.}

\item{to_eval}{Optional logical tensor of shape \verb{(n_classes,)} indicating
which classes to include in loss computation.}

\item{criterion}{Loss function module or functional to apply after constraint
propagation. Default: \code{nn_binary_cross_entropy_with_logits()}.}

\item{reduction}{(string, optional): Reduction method: \code{'none'} | \code{'mean'} | \code{'sum'}.}
}
\description{
Module wrapper for \code{\link[=nnf_mc_loss]{nnf_mc_loss()}} with configurable parameters.
Stores the ancestor matrix R and evaluation mask for reuse across batches.
}
\section{Shape}{

\itemize{
\item Input \code{output}: \eqn{(N, C)} where N = batch size, C = number of classes
\item Input \code{target}: \eqn{(N, C)}, same shape as output, binary values
\item Output: scalar by default. If \code{reduction = "none"}, then \eqn{(N, C')}
where C' is the number of evaluated classes
}
}

\examples{
\dontrun{
# Build ancestor matrix from hierarchy
R <- build_ancestor_matrix_from_outcomes(my_tree, processed$outcomes, device = "cuda")

# Create loss module
loss_fn <- nn_mc_loss(R = R, reduction = "mean")

# Forward pass
output <- model(x)  # (batch, n_classes)
loss <- loss_fn(output, labels)
loss$backward()
}

}
\seealso{
\code{\link[=nnf_mc_loss]{nnf_mc_loss()}}, \code{\link[=build_ancestor_matrix_from_outcomes]{build_ancestor_matrix_from_outcomes()}}, \code{\link[=get_constr_output]{get_constr_output()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsnip.R
\name{min_grid.tabnet}
\alias{min_grid.tabnet}
\title{Determine the minimum set of model fits}
\usage{
\method{min_grid}{tabnet}(x, grid, ...)
}
\arguments{
\item{x}{A model specification.}

\item{grid}{A tibble with tuning parameter combinations.}

\item{...}{Not currently used.}
}
\value{
A tibble with the minimum tuning parameters to fit and an additional
list column with the parameter combinations used for prediction.
}
\description{
\code{min_grid()} determines exactly what models should be fit in order to
evaluate the entire set of tuning parameter combinations. This is for
internal use only and the API may change in the near future.
}
\details{
\code{fit_max_value()} can be used in other packages to implement a \code{min_grid()}
method.
}
\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
library(dials)
library(tune)
library(parsnip)

tabnet_spec <- tabnet(decision_width = tune(), attention_width = tune()) \%>\%
  set_mode("regression") \%>\%
  set_engine("torch")

tabnet_grid <-
  tabnet_spec \%>\%
  extract_parameter_set_dials() \%>\%
  grid_regular(levels = 3)

min_grid(tabnet_spec, tabnet_grid)
\dontshow{\}) # examplesIf}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hardhat.R
\name{nn_prune_head.tabnet_fit}
\alias{nn_prune_head.tabnet_fit}
\alias{nn_prune_head.tabnet_pretrain}
\title{Prune top layer(s) of a tabnet network}
\usage{
\method{nn_prune_head}{tabnet_fit}(x, head_size)

\method{nn_prune_head}{tabnet_pretrain}(x, head_size)
}
\arguments{
\item{x}{nn_network to prune}

\item{head_size}{number of nn_layers to prune, should be less than 2}
}
\value{
a tabnet network with the top nn_layer removed
}
\description{
Prune \code{head_size} last layers of a tabnet network in order to
use the pruned module as a sequential embedding module.
}
\examples{
\dontshow{if ((torch::torch_is_installed())) withAutoprint(\{ # examplesIf}
data("ames", package = "modeldata")
x <- ames[,-which(names(ames) == "Sale_Price")]
y <- ames$Sale_Price
# pretrain a tabnet model on ames dataset
ames_pretrain <- tabnet_pretrain(x, y, epoch = 2, checkpoint_epochs = 1)
# prune classification head to get an embedding model
pruned_pretrain <- torch::nn_prune_head(ames_pretrain, 1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.tabnet_explain}
\alias{autoplot.tabnet_explain}
\title{Plot tabnet_explain mask importance heatmap}
\usage{
autoplot.tabnet_explain(
  object,
  type = c("mask_agg", "steps"),
  quantile = 1,
  ...
)
}
\arguments{
\item{object}{A \code{tabnet_explain} object as a result of \code{\link[=tabnet_explain]{tabnet_explain()}}.}

\item{type}{a character value. Either \code{"mask_agg"} the default, for a single
heatmap of aggregated mask importance per predictor along the dataset,
or \code{"steps"} for one heatmap at each mask step.}

\item{quantile}{numerical value between 0 and 1. Provides quantile clipping of the
mask values}

\item{...}{not used.}
}
\value{
A \code{ggplot} object.
}
\description{
Plot tabnet_explain mask importance heatmap
}
\details{
Plot the \code{tabnet_explain} object mask importance per variable along the predicted dataset.
\code{type="mask_agg"} output a single heatmap of mask aggregated values,
\code{type="steps"} provides a plot faceted along the \code{n_steps} mask present in the model.
\code{quantile=.995} may be used for strong outlier clipping, in order to better highlight
low values. \code{quantile=1}, the default, do not clip any values.
}
\examples{
\dontshow{if ((torch::torch_is_installed() && require("modeldata"))) withAutoprint(\{ # examplesIf}
 \dontrun{
library(ggplot2)
data("attrition", package = "modeldata")

## Single-outcome binary classification of `Attrition` in `attrition` dataset
attrition_fit <- tabnet_fit(Attrition ~. , data=attrition, epoch=11)
attrition_explain <- tabnet_explain(attrition_fit, attrition)
# Plot the model aggregated mask interpretation heatmap
autoplot(attrition_explain)

## Multi-outcome regression on `Sale_Price` and `Pool_Area` in `ames` dataset,
data("ames", package = "modeldata")
x <- ames[,-which(names(ames) \%in\% c("Sale_Price", "Pool_Area"))]
y <- ames[, c("Sale_Price", "Pool_Area")]
ames_fit <- tabnet_fit(x, y, epochs = 1, verbose=TRUE)
ames_explain <- tabnet_explain(ames_fit, x)
autoplot(ames_explain, quantile = 0.99)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask-type.R
\name{get_tau}
\alias{get_tau}
\title{Optimal threshold (tau) computation for 1.5-entmax}
\usage{
get_tau(input, dim = -1L, k = NULL)
}
\arguments{
\item{input}{The input tensor to compute thresholds over.}

\item{dim}{The dimension along which to apply 1.5-entmax. Default is -1.}

\item{k}{The number of largest elements to partial-sort over. For optimal
performance, should be slightly bigger than the expected number of
non-zeros in the solution. If the solution is more than k-sparse,
this function is recursively called with a 2*k schedule. If \code{NULL},
full sorting is performed from the beginning. Default is NULL.}
}
\value{
The threshold value for each vector, with all but the \code{dim}
dimension intact.
}
\description{
Optimal threshold (tau) computation for 1.5-entmax
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hardhat.R
\name{tabnet_pretrain}
\alias{tabnet_pretrain}
\alias{tabnet_pretrain.default}
\alias{tabnet_pretrain.data.frame}
\alias{tabnet_pretrain.formula}
\alias{tabnet_pretrain.recipe}
\alias{tabnet_pretrain.Node}
\title{Tabnet model}
\usage{
tabnet_pretrain(x, ...)

\method{tabnet_pretrain}{default}(x, ...)

\method{tabnet_pretrain}{data.frame}(
  x,
  y = NULL,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL
)

\method{tabnet_pretrain}{formula}(
  formula,
  data,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL
)

\method{tabnet_pretrain}{recipe}(
  x,
  data,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL
)

\method{tabnet_pretrain}{Node}(
  x,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL
)
}
\arguments{
\item{x}{Depending on the context:
\itemize{
\item A \strong{data frame} of predictors.
\item A \strong{matrix} of predictors.
\item A \strong{recipe} specifying a set of preprocessing steps
created from \code{\link[recipes:recipe]{recipes::recipe()}}.
\item A \strong{Node} where tree leaves will be left out,
and attributes will be used as predictors.
}

The predictor data should be standardized (e.g. centered or scaled).
The model treats categorical predictors internally thus, you don't need to
make any treatment.
The model treats missing values internally thus, you don't need to make any
treatment.}

\item{...}{Model hyperparameters.
Any hyperparameters set here will update those set by the config argument.
See \code{\link[=tabnet_config]{tabnet_config()}} for a list of all possible hyperparameters.}

\item{y}{(optional) When \code{x} is a \strong{data frame} or \strong{matrix}, \code{y} is the outcome}

\item{tabnet_model}{A pretrained \code{tabnet_model} object to continue the fitting on.
if \code{NULL} (the default) a brand new model is initialized.}

\item{config}{A set of hyperparameters created using the \code{tabnet_config} function.
If no argument is supplied, this will use the default values in \code{\link[=tabnet_config]{tabnet_config()}}.}

\item{from_epoch}{When a \code{tabnet_model} is provided, restore the network weights from a specific epoch.
Default is last available checkpoint for restored model, or last epoch for in-memory model.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a \strong{recipe} or \strong{formula} is used, \code{data} is specified as:
\itemize{
\item A \strong{data frame} containing both the predictors and the outcome.
}}
}
\value{
A TabNet model object. It can be used for serialization, predictions, or further fitting.
}
\description{
Pretrain the \href{https://arxiv.org/abs/1908.07442}{TabNet: Attentive Interpretable Tabular Learning} model
on the predictor data exclusively (unsupervised training).
}
\section{outcome}{


Outcome value are accepted here only for consistent syntax with \code{tabnet_fit}, but
by design the outcome, if present, is ignored during pre-training.
}

\section{pre-training from a previous model}{


When providing a parent \code{tabnet_model} parameter, the model pretraining resumes from that model weights
at the following epoch:
\itemize{
\item last pretrained epoch for a model already in torch context
\item Last model checkpoint epoch for a model loaded from file
\item the epoch related to a checkpoint matching or preceding the \code{from_epoch} value if provided
The model pretraining metrics append on top of the parent metrics in the returned TabNet model.
}
}

\section{Threading}{


TabNet uses \code{torch} as its backend for computation and \code{torch} uses all
available threads by default.

You can control the number of threads used by \code{torch} with:

\if{html}{\out{<div class="sourceCode">}}\preformatted{torch::torch_set_num_threads(1)
torch::torch_set_num_interop_threads(1)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
data("ames", package = "modeldata")
pretrained <- tabnet_pretrain(Sale_Price ~ ., data = ames, epochs = 1)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.tabnet_fit}
\alias{autoplot.tabnet_fit}
\alias{autoplot.tabnet_pretrain}
\title{Plot tabnet_fit model loss along epochs}
\usage{
autoplot.tabnet_fit(object, ...)

autoplot.tabnet_pretrain(object, ...)
}
\arguments{
\item{object}{A \code{tabnet_fit} or \code{tabnet_pretrain} object as a result of
\code{\link[=tabnet_fit]{tabnet_fit()}} or \code{\link[=tabnet_pretrain]{tabnet_pretrain()}}.}

\item{...}{not used.}
}
\value{
A \code{ggplot} object.
}
\description{
Plot tabnet_fit model loss along epochs
}
\details{
Plot the training loss along epochs, and validation loss along epochs if any.
A dot is added on epochs where model snapshot is available, helping
the choice of \code{from_epoch} value for later model training resume.
}
\examples{
\dontshow{if ((torch::torch_is_installed() && require("modeldata"))) withAutoprint(\{ # examplesIf}
 \dontrun{
library(ggplot2)
data("attrition", package = "modeldata")
attrition_fit <- tabnet_fit(Attrition ~. , data=attrition, valid_split=0.2, epoch=11)

# Plot the model loss over epochs
autoplot(attrition_fit)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask-type.R
\name{entmax}
\alias{entmax}
\alias{entmax15}
\title{Alpha-entmax}
\usage{
entmax(dim = -1)

entmax15(dim = -1L, k = NULL)
}
\arguments{
\item{dim}{The dimension along which to apply 1.5-entmax.}

\item{k}{The number of largest elements to partial-sort input over. For optimal
performance, should be slightly bigger than the expected number of
non-zeros in the solution. If the solution is more than k-sparse,
this function is recursively called with a 2*k schedule. If \code{NULL}, full
sorting is performed from the beginning.}
}
\value{
The projection result P  of the same shape as input, such that
\eqn{\sum_{dim} P = 1 \forall dim} elementwise.
}
\description{
With alpha = 1.5 and normalizing sparse transform (a la softmax).
}
\details{
Solves the optimization problem:
\eqn{\max_p <input, P> - H_{1.5}(P) \text{ s.t. } P \geq 0, \sum(P) == 1}
where \eqn{H_{1.5}(P)} is the Tsallis alpha-entropy with \eqn{\alpha=1.5}.
}
\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
\dontrun{
input <- torch::torch_randn(10,5, requires_grad = TRUE)
# create a top3 alpha=1.5 entmax on last input dimension
nn_entmax <- entmax15(dim=-1L, k = 3)
result <- nn_entmax(input)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mask-type.R
\name{sparsemax}
\alias{sparsemax}
\alias{sparsemax15}
\title{Sparsemax}
\usage{
sparsemax(dim = -1L)

sparsemax15(dim = -1L, k = NULL)
}
\arguments{
\item{dim}{The dimension along which to apply sparsemax.}

\item{k}{The number of largest elements to partial-sort input over. For optimal
performance, \code{k} should be slightly bigger than the expected number of
non-zeros in the solution. If the solution is more than k-sparse,
this function is recursively called with a 2*k schedule. If \code{NULL}, full
sorting is performed from the beginning.}
}
\value{
The projection result, such that \eqn{\sum_{dim} P = 1 \forall dim} elementwise.
}
\description{
Normalizing sparse transform (a la softmax).
}
\details{
Solves the projection:

\eqn{\min_P ||input - P||_2 \text{ s.t. } P \geq0, \sum(P) ==1}
}
\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
\dontrun{
input <- torch::torch_randn(10, 5, requires_grad = TRUE)
# create a top3 alpha=1.5 sparsemax on last input dimension
nn_sparsemax <- sparsemax15(dim=1, k=3)
result <- nn_sparsemax(input)
print(result)
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hardhat.R
\name{build_ancestor_matrix_from_outcomes}
\alias{build_ancestor_matrix_from_outcomes}
\title{Build ancestor matrix aligned with observed outcome classes}
\usage{
build_ancestor_matrix_from_outcomes(x, outcomes, device = "cpu")
}
\arguments{
\item{x}{A \code{data.tree::Node} object.}

\item{outcomes}{A tibble with factor columns (one per hierarchy level),
as returned by \code{hardhat::mold()$outcomes}.}

\item{device}{Torch device ("cpu" or "cuda").}
}
\value{
A \code{torch_tensor} of shape \verb{(1, n_classes, n_classes)}.
}
\description{
Extracts class names from the outcome tibble (factor levels) and builds
the ancestor matrix only for classes that actually appear in the data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss.R
\name{nnf_multilabel_one_hot}
\alias{nnf_multilabel_one_hot}
\title{Convert class_id tensor to binary one-hot tensor}
\usage{
nnf_multilabel_one_hot(y, outcomes, device = "cpu")
}
\arguments{
\item{y}{A \code{torch_tensor} of shape \verb{(batch_size, n_levels)} containing
1-based class indices.}

\item{outcomes}{A tibble with factor columns (as from \code{hardhat::mold()$outcomes}).}

\item{device}{Torch device.}
}
\value{
A \code{torch_tensor} of shape \verb{(batch_size, n_classes)} with binary values.
}
\description{
Transforms a tensor of class indices (one column per hierarchy level)
into a binary tensor where each column corresponds to a class.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hardhat.R
\name{tabnet_fit}
\alias{tabnet_fit}
\alias{tabnet_fit.default}
\alias{tabnet_fit.data.frame}
\alias{tabnet_fit.formula}
\alias{tabnet_fit.recipe}
\alias{tabnet_fit.Node}
\title{Tabnet model}
\usage{
tabnet_fit(x, ...)

\method{tabnet_fit}{default}(x, ...)

\method{tabnet_fit}{data.frame}(
  x,
  y,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL,
  weights = NULL
)

\method{tabnet_fit}{formula}(
  formula,
  data,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL,
  weights = NULL
)

\method{tabnet_fit}{recipe}(
  x,
  data,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL,
  weights = NULL
)

\method{tabnet_fit}{Node}(
  x,
  tabnet_model = NULL,
  config = tabnet_config(),
  ...,
  from_epoch = NULL
)
}
\arguments{
\item{x}{Depending on the context:
\itemize{
\item A \strong{data frame} of predictors.
\item A \strong{matrix} of predictors.
\item A \strong{recipe} specifying a set of preprocessing steps
created from \code{\link[recipes:recipe]{recipes::recipe()}}.
\item A \strong{Node} where tree will be used as hierarchical outcome,
and attributes will be used as predictors.
}

The predictor data should be standardized (e.g. centered or scaled).
The model treats categorical predictors internally thus, you don't need to
make any treatment.
The model treats missing values internally thus, you don't need to make any
treatment.}

\item{...}{Model hyperparameters.
Any hyperparameters set here will update those set by the config argument.
See \code{\link[=tabnet_config]{tabnet_config()}} for a list of all possible hyperparameters.}

\item{y}{When \code{x} is a \strong{data frame} or \strong{matrix}, \code{y} is the outcome
specified as:
\itemize{
\item A \strong{data frame} with 1 or many numeric column (regression) or 1 or many categorical columns (classification) .
\item A \strong{matrix} with 1 column.
\item A \strong{vector}, either numeric or categorical.
}}

\item{tabnet_model}{A previously fitted \code{tabnet_model} object to continue the fitting on.
if \code{NULL} (the default) a brand new model is initialized.}

\item{config}{A set of hyperparameters created using the \code{tabnet_config} function.
If no argument is supplied, this will use the default values in \code{\link[=tabnet_config]{tabnet_config()}}.}

\item{from_epoch}{When a \code{tabnet_model} is provided, restore the network weights from a specific epoch.
Default is last available checkpoint for restored model, or last epoch for in-memory model.}

\item{weights}{Unused. Placeholder for hardhat::importance_weight() variables.}

\item{formula}{A formula specifying the outcome terms on the left-hand side,
and the predictor terms on the right-hand side.}

\item{data}{When a \strong{recipe} or \strong{formula} is used, \code{data} is specified as:
\itemize{
\item A \strong{data frame} containing both the predictors and the outcome.
}}
}
\value{
A TabNet model object. It can be used for serialization, predictions, or further fitting.
}
\description{
Fits the \href{https://arxiv.org/abs/1908.07442}{TabNet: Attentive Interpretable Tabular Learning} model
}
\section{Fitting a pre-trained model}{


When providing a parent \code{tabnet_model} parameter, the model fitting resumes from that model weights
at the following epoch:
\itemize{
\item last fitted epoch for a model already in torch context
\item Last model checkpoint epoch for a model loaded from file
\item the epoch related to a checkpoint matching or preceding the \code{from_epoch} value if provided
The model fitting metrics append on top of the parent metrics in the returned TabNet model.
}
}

\section{Multi-outcome}{


TabNet allows multi-outcome prediction, which is usually named \href{https://en.wikipedia.org/wiki/Multi-label_classification}{multi-label classification}
or multi-output regression when outcomes are numerical.
Multi-outcome currently expect outcomes to be either all numeric or all categorical.
}

\section{Threading}{


TabNet uses \code{torch} as its backend for computation and \code{torch} uses all
available threads by default.

You can control the number of threads used by \code{torch} with:

\if{html}{\out{<div class="sourceCode">}}\preformatted{torch::torch_set_num_threads(1)
torch::torch_set_num_interop_threads(1)
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if ((torch::torch_is_installed() && require("modeldata"))) withAutoprint(\{ # examplesIf}
\dontrun{
data("ames", package = "modeldata")
data("attrition", package = "modeldata")

## Single-outcome regression using formula specification
fit <- tabnet_fit(Sale_Price ~ ., data = ames, epochs = 4)

## Single-outcome classification using data-frame specification
attrition_x <- attrition[ids,-which(names(attrition) == "Attrition")]
fit <- tabnet_fit(attrition_x, attrition$Attrition, epochs = 4, verbose = TRUE)

## Multi-outcome regression on `Sale_Price` and `Pool_Area` in `ames` dataset using formula,
ames_fit <- tabnet_fit(Sale_Price + Pool_Area ~ ., data = ames, epochs = 4, valid_split = 0.2)

## Multi-label classification on `Attrition` and `JobSatisfaction` in
## `attrition` dataset using recipe
library(recipes)
rec <- recipe(Attrition + JobSatisfaction ~ ., data = attrition) \%>\%
  step_normalize(all_numeric(), -all_outcomes())

attrition_fit <- tabnet_fit(rec, data = attrition, epochs = 4, valid_split = 0.2)

## Hierarchical classification on  `acme`
data(acme, package = "data.tree")

acme_fit <- tabnet_fit(acme, epochs = 4, verbose = TRUE)

# Note: Model's number of epochs should be increased for publication-level results.
}
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dials.R
\name{attention_width}
\alias{attention_width}
\alias{decision_width}
\alias{feature_reusage}
\alias{momentum}
\alias{mask_type}
\alias{num_independent}
\alias{num_shared}
\alias{num_steps}
\title{Parameters for the tabnet model}
\usage{
attention_width(range = c(8L, 64L), trans = NULL)

decision_width(range = c(8L, 64L), trans = NULL)

feature_reusage(range = c(1, 2), trans = NULL)

momentum(range = c(0.01, 0.4), trans = NULL)

mask_type(values = c("sparsemax", "entmax"))

num_independent(range = c(1L, 5L), trans = NULL)

num_shared(range = c(1L, 5L), trans = NULL)

num_steps(range = c(3L, 10L), trans = NULL)
}
\arguments{
\item{range}{the default range for the parameter value}

\item{trans}{whether to apply a transformation to the parameter}

\item{values}{possible values for factor parameters

These functions are used with \code{tune} grid functions to generate
candidates.}
}
\value{
A \code{dials} parameter to be used when tuning TabNet models.
}
\description{
Parameters for the tabnet model
}
\examples{
\dontshow{if ((require("dials") && require("parsnip") && torch::torch_is_installed())) withAutoprint(\{ # examplesIf}
  model <- tabnet(attention_width = tune(), feature_reusage = tune(),
    momentum = tune(), penalty = tune(), rate_step_size = tune()) \%>\%
    parsnip::set_mode("regression") \%>\%
    parsnip::set_engine("torch")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{node_to_df}
\alias{node_to_df}
\title{Turn a Node object into predictor and outcome.}
\usage{
node_to_df(x, drop_last_level = TRUE)
}
\arguments{
\item{x}{Node object}

\item{drop_last_level}{TRUE unused}
}
\value{
a named list of x and y, being respectively the predictor data-frame and the outcomes data-frame,
as expected inputs for \code{hardhat::mold()} function.
}
\description{
Turn a Node object into predictor and outcome.
}
\examples{
\dontshow{if ((require("data.tree") || require("dplyr"))) withAutoprint(\{ # examplesIf}
library(dplyr)
library(data.tree)
data(starwars)
starwars_tree <- starwars \%>\%
  mutate(pathString = paste("tree", species, homeworld, `name`, sep = "/")) \%>\%
  as.Node()
node_to_df(starwars_tree)$x \%>\% head()
node_to_df(starwars_tree)$y \%>\% head()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tabnet_network.R
\name{tabnet_nn}
\alias{tabnet_nn}
\title{TabNet Model Architecture}
\usage{
tabnet_nn(
  input_dim,
  output_dim,
  n_d = 8,
  n_a = 8,
  n_steps = 3,
  gamma = 1.3,
  cat_idxs = c(),
  cat_dims = c(),
  cat_emb_dim = 1,
  n_independent = 2,
  n_shared = 2,
  epsilon = 1e-15,
  virtual_batch_size = 128,
  momentum = 0.02,
  mask_type = "sparsemax",
  mask_topk = NULL
)
}
\arguments{
\item{input_dim}{Initial number of features.}

\item{output_dim}{Dimension of network output. Examples : one for regression, 2 for
binary classification etc.. Vector of those dimensions in case of multi-output.}

\item{n_d}{Dimension of the prediction layer (usually between 4 and 64).}

\item{n_a}{Dimension of the attention layer (usually between 4 and 64).}

\item{n_steps}{Number of successive steps in the network (usually between 3 and 10).}

\item{gamma}{Scaling factor for attention updates (usually between 1 and 2).}

\item{cat_idxs}{Index of each categorical column in the dataset.}

\item{cat_dims}{Number of categories in each categorical column.}

\item{cat_emb_dim}{Size of the embedding of categorical features if int, all categorical
features will have same embedding size if list of int, every corresponding feature will have
specific size.}

\item{n_independent}{Number of independent GLU layer in each GLU block of the encoder.}

\item{n_shared}{Number of shared GLU layer in each GLU block of the encoder.}

\item{epsilon}{Avoid log(0), this should be kept very low.}

\item{virtual_batch_size}{Batch size for Ghost Batch Normalization.}

\item{momentum}{Numerical value between 0 and 1 which will be used for momentum in all batch norm.}

\item{mask_type}{Either "sparsemax", "entmax" or "entmax15": the sparse masking function to use.}

\item{mask_topk}{the mask top-k value for k-sparsity selection in the mask for \code{sparsemax} and \code{entmax15}.
defaults to 1/4 of last \code{input_dim} if \code{NULL}. See \link{entmax15} for details.}
}
\description{
This is a \code{nn_module} representing the TabNet architecture from
\href{https://arxiv.org/abs/1908.07442}{Attentive Interpretable Tabular Deep Learning}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/loss.R
\name{nn_aum_loss}
\alias{nn_aum_loss}
\title{AUM loss}
\usage{
nn_aum_loss()
}
\description{
Creates a criterion that measures the Area under the \eqn{Min(FPR, FNR)} (AUM) between each
element in the input \eqn{pred_tensor} and target \eqn{label_tensor}.
}
\details{
This is used for measuring the error of a binary reconstruction within highly unbalanced dataset,
where the goal is optimizing the ROC curve. Note that the targets \eqn{label_tensor} should be factor
level of the binary outcome, i.e. with values \code{1L} and \code{2L}.
}
\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}
loss <- nn_aum_loss()
input <- torch::torch_randn(4, 6, requires_grad = TRUE)
target <- input > 1.5
output <- loss(input, target)
output$backward()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hardhat.R
\name{predict.tabnet_fit}
\alias{predict.tabnet_fit}
\alias{augment.tabnet_fit}
\title{Predict using \code{tabnet}}
\usage{
\method{predict}{tabnet_fit}(object, new_data, type = NULL, ..., epoch = NULL)

\method{augment}{tabnet_fit}(x, new_data, ...)
}
\arguments{
\item{object, x}{A \code{tabnet_fit} object.}

\item{new_data}{A data frame or matrix of new predictors.}

\item{type}{expected outcome type within  \code{c("numeric", "prob", "class")}.}

\item{...}{Not used, but required for extensibility.}

\item{epoch}{the epoch of an existing checkpoint to infer from.}
}
\value{
\code{\link[=predict]{predict()}} returns a tibble of predictions and \code{\link[=augment]{augment()}} appends the
columns in \code{new_data}. In either case, the number of rows in the tibble is
guaranteed to be the same as the number of rows in \code{new_data}.

For regression data, the prediction is in the column \code{.pred}. For
classification, the class predictions are in \code{.pred_class} and the
probability estimates are in columns with the pattern \verb{.pred_\{level\}} where
\code{level} is the levels of the outcome factor vector.
}
\description{
Predict using \code{tabnet}
}
\examples{
# Minimal example for quick execution
car_split <- rsample::initial_split(mtcars[ 1:6,   ])

\dontrun{
# Fit
if (torch_is_installed() & interactive()) {
 mod <- tabnet_fit(mpg ~ cyl + log(drat), training(car_split))

 # Predict
 predict(mod, testing(car_split))
 augment(mod, testing(car_split))
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_explain.R
\name{tabnet_explain}
\alias{tabnet_explain}
\alias{tabnet_explain.default}
\alias{tabnet_explain.tabnet_fit}
\alias{tabnet_explain.tabnet_pretrain}
\alias{tabnet_explain.model_fit}
\title{Interpretation metrics from a TabNet model}
\usage{
tabnet_explain(object, new_data)

\method{tabnet_explain}{default}(object, new_data)

\method{tabnet_explain}{tabnet_fit}(object, new_data)

\method{tabnet_explain}{tabnet_pretrain}(object, new_data)

\method{tabnet_explain}{model_fit}(object, new_data)
}
\arguments{
\item{object}{a TabNet fit object}

\item{new_data}{a data.frame to obtain interpretation metrics.}
}
\value{
Returns a list with
\itemize{
\item \code{M_explain}: the aggregated feature importance masks as detailed in
TabNet's paper.
\item \code{masks} a list containing the masks for each step.
}
}
\description{
Interpretation metrics from a TabNet model
}
\examples{
\dontshow{if (torch::torch_is_installed()) withAutoprint(\{ # examplesIf}

set.seed(2021)

n <- 256
x <- data.frame(
  x = rnorm(n),
  y = rnorm(n),
  z = rnorm(n)
)

y <- x$x

fit <- tabnet_fit(x, y, epochs = 10,
                  num_steps = 1,
                  batch_size = 512,
                  attention_width = 1,
                  num_shared = 1,
                  num_independent = 1)


 ex <- tabnet_explain(fit, x)

\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{check_compliant_node}
\alias{check_compliant_node}
\title{Check that Node object names are compliant}
\usage{
check_compliant_node(node)
}
\arguments{
\item{node}{the Node object, or a dataframe ready to be parsed by \code{data.tree::as.Node()}}
}
\value{
node if it is compliant, else an Error with the column names to fix
}
\description{
Check that Node object names are compliant
}
\examples{
\dontshow{if ((require("data.tree") || require("dplyr"))) withAutoprint(\{ # examplesIf}
library(dplyr)
library(data.tree)
data(starwars)
starwars_tree <- starwars \%>\%
  mutate(pathString = paste("tree", species, homeworld, `name`, sep = "/"))

# pre as.Node() check
try(check_compliant_node(starwars_tree))

# post as.Node() check
check_compliant_node(as.Node(starwars_tree))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dials.R
\name{cat_emb_dim}
\alias{cat_emb_dim}
\alias{checkpoint_epochs}
\alias{drop_last}
\alias{encoder_activation}
\alias{lr_scheduler}
\alias{mlp_activation}
\alias{mlp_hidden_multiplier}
\alias{num_independent_decoder}
\alias{num_shared_decoder}
\alias{optimizer}
\alias{penalty}
\alias{verbose}
\alias{virtual_batch_size}
\title{Non-tunable parameters for the tabnet model}
\usage{
cat_emb_dim(range = NULL, trans = NULL)

checkpoint_epochs(range = NULL, trans = NULL)

drop_last(range = NULL, trans = NULL)

encoder_activation(range = NULL, trans = NULL)

lr_scheduler(range = NULL, trans = NULL)

mlp_activation(range = NULL, trans = NULL)

mlp_hidden_multiplier(range = NULL, trans = NULL)

num_independent_decoder(range = NULL, trans = NULL)

num_shared_decoder(range = NULL, trans = NULL)

optimizer(range = NULL, trans = NULL)

penalty(range = NULL, trans = NULL)

verbose(range = NULL, trans = NULL)

virtual_batch_size(range = NULL, trans = NULL)
}
\arguments{
\item{range}{unused}

\item{trans}{unused}
}
\description{
Non-tunable parameters for the tabnet model
}
